SystemOrganization addCategory: #'Magma server-API'!
SystemOrganization addCategory: #'Magma server-Exceptions'!
SystemOrganization addCategory: #'Magma server-File'!
SystemOrganization addCategory: #'Magma server-Info'!
SystemOrganization addCategory: #'Magma server-Preferences'!
SystemOrganization addCategory: #'Magma server-Recovery'!
SystemOrganization addCategory: #'Magma server-Utilities'!
SystemOrganization addCategory: #'Magma server-private'!

!MagmaLargeCollectionChanges methodsFor: '*magma server' stamp: 'cmm 10/26/2006 17:02'!
requiresServerRefresh
	^ false! !

!MaSegmentByIndexRequest methodsFor: '*magma server' stamp: 'brp 6/14/2006 10:50'!
process

	^ self repository
		upTo: quantity
		oidsFromIndex: lowIndex
		inLargeCollection: collectionOid
		using: readStrategy
		forSession: sessionId! !

!MagmaRequestSwapPrimaryDutyRequest methodsFor: '*magma server' stamp: 'cmm 3/3/2009 20:56'!
nodeCheck
	console ifNil: 
		[ MagmaUserError signal: self class name , ' only applies when the repository is deployed on-line, with at least two servers forming a HA node.' ].
	console restoreMode ifFalse: [ MagmaUserError signal: 'Repository is not a warm-backup.' ].
	(self node includesServerAt: self thisLocation) ifFalse: 
		[ MagmaEnvironmentError signal: 'Node should include the secondary requesting primary duty.' ].
	self shouldBeSecondary! !

!MagmaRequestSwapPrimaryDutyRequest methodsFor: '*magma server' stamp: 'cmm 3/8/2009 17:58'!
process
	| lastCommits sess |
	sess := self node primarySession disconnect.
	lastCommits := sess 
		relinquishPrimaryDutyTo: self thisLocation
		atCommitNumber: repositoryController commitNumber + 1
		branchCode: repositoryController branchCode.
	"Not using #commitRestore because I don't see that we need a new branchCode."
	self repository
		applyCommitRecords: lastCommits ;
		restoreMode: false.
	^ true! !

!MagmaRequestSwapPrimaryDutyRequest methodsFor: '*magma server' stamp: 'cmm 3/3/2009 20:32'!
shouldBeSecondary
	"Never ping anything else when receiving a swap request, if the administrator has the wrong server, the jig is up."
	self isSecondary not ifTrue: 
		[ MagmaWrongServerError new
			correctServer: self node primaryLocation ;
			signal: self printString , ' should be sent to a secondary server of multi-server nodes.' ]! !

!MaStorageObjectBuffer methodsFor: '*magma server' stamp: 'cmm 3/8/2007 13:00'!
graphDo: oneArgBlock using: aMaReadStrategy for: aMagmaFileTraverser
	"Though I am an object stored in a different representation, that representation is fully part of the repository, so enumerate the buffers my MaSerializedGraphBuffers buffers."
	oneArgBlock value: self.
	self buffersDo:
		[ : eachBuffer |
		"The buffers of a storage-object are never assigned permanent oids.  But if already-persistent objects are referenced from the storage-object, they will be and should be traversed normally."
		(MaOidCalculator isOidForUserObject: eachBuffer oid)
			ifTrue:
				[ eachBuffer
					graphDo: oneArgBlock
					using: aMaReadStrategy
					for: aMagmaFileTraverser ]
			ifFalse:
				[ oneArgBlock value: eachBuffer ] ]! !

!MaRemoveSecondaryLocationRequest methodsFor: '*magma server' stamp: 'cmm 3/4/2009 15:49'!
nodeCheck
	self isPrimary ifFalse: 
		[ MagmaUserError signal: 'This request may only be sent to the primary server.' ]! !

!MaRemoveSecondaryLocationRequest methodsFor: '*magma server' stamp: 'cmm 3/4/2009 15:48'!
process
	console ifNil: 
		[ MagmaUserError signal: 'This operation is only applicable to secondary servers in a multi-server HA Node configuration.' ].
	console removeSecondaryLocation: locationToRemove.
	^ true! !

!MaWarmBackupUpdateRequest methodsFor: '*magma server' stamp: 'cmm 12/29/2008 13:50'!
ensureNoWarmBackupNode
	^ (repositoryController node locations includes: requestingLocation) ifTrue: 
		[ repositoryController session commit: [ repositoryController node removeLocation: requestingLocation ] ]! !

!MaWarmBackupUpdateRequest methodsFor: '*magma server' stamp: 'cmm 12/29/2008 13:50'!
ensureWarmBackupNode
	^ (repositoryController node locations includes: requestingLocation) ifFalse: 
		[ repositoryController session commit: [ repositoryController node addLocation: requestingLocation ] ]! !

!MaWarmBackupUpdateRequest methodsFor: '*magma server' stamp: 'cmm 12/29/2008 13:54'!
process
	| answer doneIndicator |
	doneIndicator := true.
	answer := super process.
	answer 
		ifEmpty: [ self ensureWarmBackupNode ]
		ifNotEmpty: 
			[ "Sending back some commit-records, but did we make it to the end?  If so, we can add as a warm-backup right now."
			answer last commitNumber = repositoryController commitNumber 
				ifTrue: [ self ensureWarmBackupNode ]
				ifFalse: 
					[ self ensureNoWarmBackupNode.
					doneIndicator := false ] ].
	^ answer -> doneIndicator! !

!MaRepositoryDisconnectRequest methodsFor: '*magma server' stamp: 'cmm 5/14/2005 18:15'!
process

	self repository removeSessionFor: sessionId.
	^ nil! !

!MagmaArray methodsFor: '*magma server' stamp: 'cmm 7/12/2005 12:52'!
implementationClass

	^ MaLargeArrayOfNumbers! !

!MagmaArray methodsFor: '*magma server' stamp: 'cmm 3/17/2005 21:33'!
newManagerUsing: aMaObjectRepository

	^ MagmaArrayManager
		collection: self
		repository: aMaObjectRepository! !

!MagmaIdRequest methodsFor: '*magma server' stamp: 'cmm 1/14/2005 16:05'!
process

	^ repositoryController magmaId! !

!MaByteObjectBuffer methodsFor: '*magma server' stamp: 'cmm 1/25/2005 14:07'!
appendGraphUsing: aMaReadStrategy into: aMaSerializedGraphBuffer currentDepth: currentDepth minDepth: minDepth with: alreadyAppendedSet filer: aMaObjectFiler

	"Nothing to do for a byte-object."
! !

!MaByteObjectBuffer methodsFor: '*magma server' stamp: 'cmm 3/8/2007 12:58'!
graphDo: oneArgBlock using: aMaReadStrategy for: aMagmaFileTraverser

	"nothing to do for a Byte-object buffer."! !

!MagmaSwapPrimaryDutyRequest methodsFor: '*magma server' stamp: 'cmm 3/3/2009 21:38'!
nodeCheck
	self shouldBePrimary! !

!MagmaSwapPrimaryDutyRequest methodsFor: '*magma server' stamp: 'cmm 12/29/2008 16:41'!
process
	^ repositoryController 
		executeSwapWith: newPrimary
		commitNumber: commitNumber
		branchCode: branchCode! !

!MagmaSwapPrimaryDutyRequest methodsFor: '*magma server' stamp: 'cmm 3/3/2009 21:38'!
requiresConnection
	^ false! !

MaObject subclass: #MaAtomicFileStream
	instanceVariableNames: 'filename fileStreams sizes guard uncommittedSize position writers pool'
	classVariableNames: 'MaxPhysicalFileSize'
	poolDictionaries: ''
	category: 'Magma server-File'!

!MaAtomicFileStream commentStamp: 'cmm 3/5/2009 16:44' prior: 0!
This class emulates a StandardFileStream except doesn't actually write contents to disk until you send #maCommit.  Instead, it caches the written contents to a collection of MaAtomicFileWriters.

Also, the latest additional feature is that it can be of unlimited size, it reads and writes its multiple physical 'fileStreams' depending on what the file-position requested is.!

!MaAtomicFileStream class methodsFor: 'create' stamp: 'cmm 4/23/2007 00:24'!
fileNamed: filenameString 
	^ self new
		filename: filenameString ;
		initializeStreamsAndSizes ;
		yourself! !

!MaAtomicFileStream class methodsFor: 'class initialization' stamp: 'cmm 5/23/2007 23:18'!
initialize
	super initialize.
	self maxPhysicalFileSize: 1800000000 "1.8 gigabytes"! !

!MaAtomicFileStream class methodsFor: 'class initialization' stamp: 'cmm 5/23/2007 23:18'!
maxPhysicalFileSize: anInteger
	"Do not change this unless you know what you are doing."
	MaxPhysicalFileSize _ anInteger! !

!MaAtomicFileStream methodsFor: 'private' stamp: 'cmm 8/12/2007 15:08'!
addPhysicalFile
	sizes add: 0.
	^ (StandardFileStream newFileNamed: self nextName) in: 
		[ : stream | 
		stream closed ifTrue: [ MagmaEnvironmentError signal: 'file did not open' ].
		fileStreams add: stream ]! !

!MaAtomicFileStream methodsFor: 'accessing' stamp: 'cmm 4/23/2007 00:39'!
baseName
	^ FileDirectory baseNameFor: self localName! !

!MaAtomicFileStream methodsFor: 'filestream api' stamp: 'cmm 4/26/2007 22:59'!
close
	self isPhysical ifTrue: 
		[ self maCommit.
		fileStreams copy withIndexDo: 
			[ : each : x | 
			pool close: each.
			"We're going in order, so removeFirst works."
			fileStreams removeFirst.
			sizes removeFirst ] ]! !

!MaAtomicFileStream methodsFor: 'filestream api' stamp: 'cmm 4/28/2007 19:44'!
closed
	"Answer whether I have a file-handle allocated from the operating system."
	^ self isPhysical not or: 
		[ self isPhysical and: [ fileStreams allSatisfy: [ : each | each closed ] ] ]! !

!MaAtomicFileStream methodsFor: 'accessing' stamp: 'cmm 4/11/2007 23:05'!
committedSize
	^ sizes 
		inject: 0
		into: [ : sum : each | sum + each ]! !

!MaAtomicFileStream methodsFor: '*magma server-file' stamp: 'cmm 6/23/2007 16:01'!
currentRecordFor: aMaAtomicFileWriter 
	| answer |
	answer _ ByteArray new: aMaAtomicFileWriter byteArray size.
	guard critical: 
		[ self 
			streamAt: aMaAtomicFileWriter filePos
			withRelativePositionDo: 
				[ : physicalStream : relativePos | 
				self renewAccess: physicalStream.
				physicalStream 
					maRead: answer size
					bytesFromPosition: 1
					of: answer
					atFilePosition: relativePos ]
			ifNone: 
				[ MagmaSoftwareError signal: 'expected physical data at ' , aMaAtomicFileWriter filePos asString ] ].
	^ answer! !

!MaAtomicFileStream methodsFor: 'private' stamp: 'cmm 4/10/2007 21:48'!
delete: fileIndex 
	"Delete the particular index of my file numbered fileIndex.  This is used when truncating."
	FileDirectory deleteFilePath: (self nameFor: fileIndex)! !

!MaAtomicFileStream methodsFor: 'accessing' stamp: 'cmm 6/4/2007 08:53'!
deletePhysicalFiles
	self directory in: [ : dir | self physicalFilenamesDo: [ : each | dir deleteFileNamed: each ] ]! !

!MaAtomicFileStream methodsFor: 'filestream api' stamp: 'cmm 7/11/2005 22:22'!
directory

	^ FileDirectory forFileName: filename! !

!MaAtomicFileStream methodsFor: 'testing' stamp: 'cmm 9/28/2008 23:03'!
exists
	| fd |
	fd := FileDirectory on: (FileDirectory dirPathFor: filename).
	^ fd fileExists: self baseName , '.1.' , self extension! !

!MaAtomicFileStream methodsFor: 'accessing' stamp: 'cmm 4/23/2007 00:39'!
extension
	^ FileDirectory extensionFor: self localName! !

!MaAtomicFileStream methodsFor: 'filestream api' stamp: 'cmm 5/24/2006 21:50'!
filename

	^ filename! !

!MaAtomicFileStream methodsFor: 'initialize-release' stamp: 'cmm 7/11/2005 22:28'!
filename: aString

	filename _ aString! !

!MaAtomicFileStream methodsFor: 'filestream api' stamp: 'cmm 4/11/2007 22:44'!
flush
	"API compatibility with StandardFileStream.  Write all of my writers and force an unconditional flush of each file."
	self maCommit.
	guard critical: [ self physicalFlush ]! !

!MaAtomicFileStream methodsFor: 'initialize-release' stamp: 'cmm 9/23/2008 14:26'!
initialize
	super initialize.
	self initializeWriters.
	position := 0.
	guard := Mutex new.
	"To make code simple, everyone gets a default pool, which is just a pool of one, me."
	pool := MaFileStreamPool maxSize: 1.
	fileStreams := OrderedCollection new.
	sizes := OrderedCollection new! !

!MaAtomicFileStream methodsFor: 'initialize-release' stamp: 'cmm 8/12/2007 15:10'!
initializeStreamsAndSizes
	"Determine which streams are available in the current directory and their sizes."
	| fd baseName extension matchingEntries |
	fd _ FileDirectory on: (FileDirectory dirPathFor: filename).
	baseName _ self baseName.
	extension _ self extension.
	"Gather up the entries for my filename and extension."
	matchingEntries _ fd entries select: 
		[ : eachEntry | 
		| eachNameVersionAndExtension |
		eachNameVersionAndExtension _ fd splitNameVersionExtensionFor: eachEntry name.
		eachNameVersionAndExtension first = baseName and: [ eachNameVersionAndExtension last = extension ] ].
	"Then sort them by the file-number (second dot-delimited field)"
	matchingEntries _ matchingEntries asSortedCollection: 
		[ : a : b | 
		| nameVersionAndExtensionA nameVersionAndExtensionB |
		nameVersionAndExtensionA _ fd splitNameVersionExtensionFor: a name.
		nameVersionAndExtensionB _ fd splitNameVersionExtensionFor: b name.
		(nameVersionAndExtensionA at: 2) < (nameVersionAndExtensionB at: 2) ].
	"Now load my fileStreams and sizes with these correctly-ordered entries."
	matchingEntries do: 
		[ : eachEntry | 
		| physicalStream |
		physicalStream _ StandardFileStream readOnlyFileNamed: (fd fullNameFor: eachEntry name).
		physicalStream closed ifTrue: [ MagmaEnvironmentError signal: 'file did not open for read-only' ].
		fileStreams add: physicalStream.
		physicalStream
			close ;
			readWrite.
		sizes add: eachEntry fileSize ]! !

!MaAtomicFileStream methodsFor: 'initialize-release' stamp: 'cmm 7/13/2005 15:04'!
initializeWriters

	writers _ BTree new.
	uncommittedSize _ 0! !

!MaAtomicFileStream methodsFor: 'testing' stamp: 'cmm 7/11/2005 22:04'!
isNew

	^ self isPhysical not! !

!MaAtomicFileStream methodsFor: 'testing' stamp: 'cmm 4/9/2007 23:59'!
isPhysical
	^ fileStreams notEmpty! !

!MaAtomicFileStream methodsFor: 'accessing' stamp: 'cmm 4/23/2007 00:30'!
localName
	^ (FileDirectory on: (FileDirectory dirPathFor: filename)) localNameFor: filename! !

!MaAtomicFileStream methodsFor: 'commit' stamp: 'cmm 7/4/2005 17:47'!
maCommit

	self maCommit: false! !

!MaAtomicFileStream methodsFor: 'commit' stamp: 'cmm 4/11/2007 22:46'!
maCommit: shouldFlush 
	"writers, write!!  If shouldFlush is true, only flush if I actually wrote something (if an unconditional flush is needed, use #flush)."
	| needsFlush |
	guard critical: 
		[ needsFlush _ false.
		self writersDo: 
			[ : each | 
			each write.
			needsFlush _ true ].
		self initializeWriters.
		(shouldFlush and: [ needsFlush ]) ifTrue: [ self physicalFlush ] ]! !

!MaAtomicFileStream methodsFor: 'testing' stamp: 'cmm 10/22/2006 15:01'!
maIsTransactional
	^ true! !

!MaAtomicFileStream methodsFor: 'printing' stamp: 'cmm 4/23/2007 17:37'!
maPrintAbbreviatedOn: aStream
	super maPrintAbbreviatedOn: aStream.
	aStream maPrint: filename! !

!MaAtomicFileStream methodsFor: 'filestream api' stamp: 'cmm 4/28/2007 19:08'!
maRead: someNumber bytesFromPosition: bufferPosition of: aByteArray atFilePosition: filePosition 
	| writer numberOfBytesRead |
	^ guard critical: 
		[ (writer _ self writerAt: filePosition) 
			ifNil: 
				[ numberOfBytesRead _ self 
					streamAt: filePosition
					withRelativePositionDo: 
						[ : physicalStream : relativePosition | 
						self renewAccess: physicalStream.
						physicalStream 
							maRead: someNumber
							bytesFromPosition: bufferPosition
							of: aByteArray
							atFilePosition: relativePosition ]
					ifNone: [ 0 ].
				position _ filePosition + numberOfBytesRead.
				numberOfBytesRead ]
			ifNotNil: 
				[ numberOfBytesRead _ someNumber min: writer byteArray size.
				aByteArray 
					replaceFrom: bufferPosition
					to: bufferPosition + numberOfBytesRead - 1
					with: writer byteArray
					startingAt: 1.
				position _ filePosition + numberOfBytesRead.
				numberOfBytesRead ] ]! !

!MaAtomicFileStream methodsFor: 'filestream api' stamp: 'cmm 4/29/2007 21:48'!
maWrite: numberOfBytes bytesFromPosition: bufferPosition of: byteArray atFilePosition: filePosition 
	| writer |
	writer _ MaAtomicFileWriter 
		file: self
		filePos: filePosition
		byteArray: (byteArray 
				copyFrom: bufferPosition
				to: bufferPosition + numberOfBytes - 1).
	guard critical: 
		[ writers 
			at: writer endFilePos
			put: writer.
		uncommittedSize _ uncommittedSize max: filePosition + numberOfBytes.
		position _ filePosition + numberOfBytes ]! !

!MaAtomicFileStream methodsFor: 'private' stamp: 'cmm 5/23/2007 23:16'!
maxPhysicalFileSize
	^ MaxPhysicalFileSize! !

!MaAtomicFileStream methodsFor: 'private' stamp: 'cmm 4/20/2007 16:13'!
nameFor: fileIndex 
	"The name of the file numbered fileIndex."
	^ (FileDirectory default splitNameVersionExtensionFor: filename) in: 
		[ : nameNumberAndExtension | 
		String streamContents: 
			[ : stream | 
			stream
				maPrint: nameNumberAndExtension first ;
				nextPut: $. ;
				maPrint: fileIndex.
			nameNumberAndExtension last ifNotEmpty: 
				[ stream
					nextPut: $. ;
					maPrint: nameNumberAndExtension last ] ] ]! !

!MaAtomicFileStream methodsFor: 'filestream api' stamp: 'cmm 2/13/2008 17:22'!
next: n into: aByteArray startingAt: startIndex 
	"This method is only called when the initial read of an object did not fit within the #trackSize."
	| writer |
	guard critical: 
		[ (writer := self writerAt: position) 
			ifNil: 
				[ self 
					streamAt: position
					withRelativePositionDo: 
						[ : physicalStream : relativePosition | 
						self renewAccess: physicalStream.
						physicalStream 
							next: n
							into: aByteArray
							startingAt: startIndex ]
					ifNone: [ MaSoftwareError signal: 'unexpected EOF - ' , position printString ] ]
			ifNotNil: 
				[ aByteArray 
					replaceFrom: startIndex
					to: startIndex + (n min: writer byteArray size) - 1
					with: writer byteArray
					startingAt: position - writer filePos + 1 ].
		position := position + n ]! !

!MaAtomicFileStream methodsFor: 'filestream api' stamp: 'cmm 7/7/2005 22:17'!
next: numberOfBytes putAll: aByteArray startingAt: startPos

	"This isn't needed for Magma, and I don't feel like implementing it right now.  Better to just give an explicit error than get it wrong."

	self error: 'not supported'! !

!MaAtomicFileStream methodsFor: 'private' stamp: 'cmm 4/11/2007 23:03'!
nextName
	^ self nameFor: fileStreams size+1! !

!MaAtomicFileStream methodsFor: 'accessing' stamp: 'cmm 4/28/2007 18:23'!
numberOfPhysicalFiles
	^ sizes size! !

!MaAtomicFileStream methodsFor: 'accessing' stamp: 'cmm 6/4/2007 08:41'!
physicalFilenamesDo: oneArgBlock 
	"Value oneArgBlock with the names of my physical files."
	self directory in: 
		[ : dir | 
		dir entries do: 
			[ : each | 
			| eachParts |
			eachParts _ dir splitNameVersionExtensionFor: each name.
			each name = (dir localNameFor: (self nameFor: eachParts second)) ifTrue: 
				[ "version"
				oneArgBlock value: each name ] ] ]! !

!MaAtomicFileStream methodsFor: 'private' stamp: 'cmm 4/11/2007 22:47'!
physicalFlush
	^ fileStreams do: [ : each | each closed ifFalse: [ each flush ] ]! !

!MaAtomicFileStream methodsFor: 'private' stamp: 'cmm 6/23/2007 15:48'!
physicallyWrite: numberOfBytes bytesFromPosition: bufferPosition of: aByteArray atFilePosition: filePosition 
	guard critical: 
		[ self 
			streamAt: filePosition
			withRelativePositionDo: 
				[ : fileStream : relativePos | 
				self renewAccess: fileStream.
				fileStream 
					maWrite: aByteArray size
					bytesFromPosition: 1
					of: aByteArray
					atFilePosition: relativePos ]
			ifNone: 
				[ "filePosition is beyond the end of the last file, so we know we will write either the last file or a new file,"
				| streamToWrite positionOfLastFileStream |
				streamToWrite _ (sizes isEmpty or: [ aByteArray size + sizes last > self maxPhysicalFileSize ]) 
					ifTrue: [ self addPhysicalFile ]
					ifFalse: [ fileStreams last ].
				self renewAccess: streamToWrite.
				positionOfLastFileStream _ self committedSize - sizes last.
				streamToWrite 
					maWrite: aByteArray size
					bytesFromPosition: 1
					of: aByteArray
					atFilePosition: filePosition - positionOfLastFileStream.
				sizes 
					at: sizes size
					put: fileStreams last size ] ]! !

!MaAtomicFileStream methodsFor: 'initialize-release' stamp: 'cmm 10/22/2006 15:00'!
pool: aMaFileStreamPool
	pool _ aMaFileStreamPool! !

!MaAtomicFileStream methodsFor: 'private' stamp: 'cmm 4/28/2007 19:08'!
renewAccess: aPhysicalStream
	pool renewAccess: aPhysicalStream! !

!MaAtomicFileStream methodsFor: 'filestream api' stamp: 'cmm 7/7/2005 12:51'!
size

	^ self committedSize max: uncommittedSize! !

!MaAtomicFileStream methodsFor: 'private' stamp: 'cmm 4/20/2007 14:10'!
streamAt: filePosition withRelativePositionDo: twoArgBlock ifNone: aBlock 
	"Value twoArgBlock with the fileStream that contains filePosition, and the relative position within that fileStream that matches filePosition."
	^ self 
		streamIndexAt: filePosition
		withRelativePositionDo: 
			[ : index : relativePosition | 
			twoArgBlock 
				value: (fileStreams at: index)
				value: relativePosition ]
		ifNone: aBlock! !

!MaAtomicFileStream methodsFor: 'private' stamp: 'cmm 4/20/2007 16:07'!
streamIndexAt: filePosition withRelativePositionDo: twoArgBlock ifNone: aBlock
	"Value twoArgBlock with the index of the fileStream that contains filePosition, and the relative position within that fileStream that matches filePosition."
	| runningSize |
	runningSize _ 0.
	sizes withIndexDo: 
		[ : eachSize : index | 
		filePosition < (runningSize + eachSize) ifTrue: 
			[ | relativePosition |
			relativePosition _ filePosition - runningSize.
			^ twoArgBlock 
				value: index
				value: relativePosition ].
		runningSize _ runningSize + eachSize ].
	"filePosition is Beyond end of last fileStream."
	^ aBlock value! !

!MaAtomicFileStream methodsFor: 'filestream api' stamp: 'cmm 4/10/2007 21:27'!
truncate: filePosition 
	self
		truncateStreams: filePosition ;
		truncateWriters: filePosition! !

!MaAtomicFileStream methodsFor: 'private' stamp: 'cmm 8/14/2007 22:50'!
truncateStreams: filePosition 
	| toBeDeleted fileStreamToTruncate |
	guard critical: 
		[ self 
			streamIndexAt: filePosition
			withRelativePositionDo: 
				[ : index : relativePosition | 
				toBeDeleted _ index + 1 to: fileStreams size.
				fileStreamToTruncate _ fileStreams at: index.
				self renewAccess: fileStreamToTruncate.
				fileStreamToTruncate
					truncate: relativePosition ;
					close ;
					open
				"close and open to get around bug; see http://lists.squeakfoundation.org/pipermail/squeak-dev/2005-June/092041.html" ]
			ifNone: 
				[ "do nothing"
				 ].
		toBeDeleted ifNotNil: [ toBeDeleted do: [ : each | self delete: each ] ] ]! !

!MaAtomicFileStream methodsFor: 'private' stamp: 'cmm 6/23/2007 15:57'!
truncateWriters: filePosition 
	| toBeRemoved |
	guard critical: 
		[ uncommittedSize _ uncommittedSize min: filePosition.
		toBeRemoved _ OrderedCollection new.	"Not sure if I can remove from a BTree while enumerating it.."
		writers 
			from: filePosition
			to: 999999999999999999999
			do: 
				[ : each | 
				(each coversFilePosition: filePosition) ifTrue: [ toBeRemoved add: each ] ].
		toBeRemoved do: [ : each | writers removeKey: each endFilePos ] ]! !

!MaAtomicFileStream methodsFor: 'private' stamp: 'cmm 6/16/2005 09:23'!
writerAt: filePosition

	writers
		from: filePosition
		to: 999999999999999999999
		do: [ : each | ^ (each coversFilePosition: filePosition) ifTrue: [ each ] ].
	^ nil! !

!MaAtomicFileStream methodsFor: 'accessing' stamp: 'cmm 7/13/2005 15:02'!
writersDo: aBlock

	writers do: aBlock! !

MaObject subclass: #MaAtomicFileWriter
	instanceVariableNames: 'file filePos byteArray'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Magma server-File'!

!MaAtomicFileWriter class methodsFor: 'as yet unclassified' stamp: 'cmm 10/21/2006 20:05'!
file: aMaTransactionalFileStream filePos: anInteger byteArray: aByteArray
	^ self new
		setFile: aMaTransactionalFileStream
		filePos: anInteger
		byteArray: aByteArray! !

!MaAtomicFileWriter methodsFor: 'accessing' stamp: 'cmm 6/15/2005 10:27'!
byteArray
	^ byteArray! !

!MaAtomicFileWriter methodsFor: 'testing' stamp: 'cmm 6/15/2005 17:11'!
coversFilePosition: filePosition

	^ filePosition
		between: filePos
		and: filePos + byteArray size - 1! !

!MaAtomicFileWriter methodsFor: 'accessing' stamp: 'cmm 8/13/2007 22:39'!
currentRecord
	^ file currentRecordFor: self! !

!MaAtomicFileWriter methodsFor: 'accessing' stamp: 'cmm 6/15/2005 22:18'!
endFilePos

	^ filePos + byteArray size - 1! !

!MaAtomicFileWriter methodsFor: 'accessing' stamp: 'cmm 6/15/2005 17:27'!
filePos

	^ filePos! !

!MaAtomicFileWriter methodsFor: 'initialize-release' stamp: 'cmm 10/21/2006 20:06'!
setFile: aMaTransactionalFileStream filePos: anInteger byteArray: aByteArray
	file _ aMaTransactionalFileStream.
	filePos _ anInteger.
	byteArray _ aByteArray! !

!MaAtomicFileWriter methodsFor: 'testing' stamp: 'cmm 7/12/2005 22:20'!
willCauseGrowth

	^ filePos >= file committedSize! !

!MaAtomicFileWriter methodsFor: 'actions' stamp: 'cmm 4/20/2007 15:28'!
write
	file 
		physicallyWrite: byteArray size
		bytesFromPosition: 1
		of: byteArray
		atFilePosition: filePos! !

MaObject subclass: #MaFileStreamPool
	instanceVariableNames: 'guard openStreams maxSize'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Magma server-File'!

!MaFileStreamPool class methodsFor: 'create' stamp: 'cmm 10/21/2006 17:42'!
maxSize: anInteger 
	^ (self new)
		setMaxPoolSize: anInteger ;
		yourself! !

!MaFileStreamPool methodsFor: 'api' stamp: 'cmm 4/20/2007 15:01'!
close: aStandardFileStream 
	| stream |
	guard critical: 
		[ stream _ openStreams 
			remove: aStandardFileStream
			ifAbsent: [ nil ].
		stream ifNotNil: [ stream close ] ]! !

!MaFileStreamPool methodsFor: 'initialize' stamp: 'cmm 9/23/2008 14:25'!
initialize
	super initialize.
	openStreams := OrderedCollection new.
	guard := Mutex new! !

!MaFileStreamPool methodsFor: 'api' stamp: 'cmm 10/22/2006 15:18'!
maxSize
	^maxSize! !

!MaFileStreamPool methodsFor: 'api' stamp: 'cmm 10/22/2006 15:19'!
maxSize: anInteger
	maxSize _ anInteger! !

!MaFileStreamPool methodsFor: 'api' stamp: 'cmm 6/2/2008 10:51'!
renewAccess: aStandardFileStream 
	"Add or move aStandardFileStream to the bottom (most recently accessed), ensuring I do not exceed my limit of simultaneously-allowed open FileStreams."
	guard critical: 
		[ "Performance enhancement:  since many consecutive writes to the same fileStream is common, check if already renewed."
		| existing |
		(openStreams notEmpty and: [ openStreams last == aStandardFileStream ]) ifFalse: 
			[ existing := openStreams 
				remove: aStandardFileStream
				ifAbsent: [ nil ].
			"If already existing, no need to pop anything off, otherwise make sure we don't exceed.."
			existing ifNil: 
				[ [ openStreams size >= maxSize ] whileTrue: [ self close: openStreams first ].
				aStandardFileStream open.
				aStandardFileStream closed ifTrue: 
					[ MagmaEnvironmentError signal: 'Could not open ' , aStandardFileStream name , '.  File in use?' ] ].
			openStreams addLast: aStandardFileStream ] ]! !

!MaFileStreamPool methodsFor: 'initialize' stamp: 'cmm 10/22/2006 15:18'!
setMaxPoolSize: anInteger 
	maxSize _ anInteger! !

MaObject subclass: #MaLargeCollectionManager
	instanceVariableNames: 'collection memberIndex repository'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Magma server-private'!

!MaLargeCollectionManager commentStamp: 'cmm 3/16/2005 12:48' prior: 0!
I perform server-side operations for MagmaLargeCollections.  Adding, removing and accessing.!

!MaLargeCollectionManager class methodsFor: 'creation' stamp: 'cmm 7/18/2005 13:22'!
collection: aMagmaCollection repository: aMaObjectRepository

	^self new
		collection: aMagmaCollection ;
		repository: aMaObjectRepository ;
		yourself! !

!MaLargeCollectionManager methodsFor: 'initializing' stamp: 'cmm 3/16/2005 14:24'!
close

	memberIndex ifNotNil: [ memberIndex close.  memberIndex _ nil ]! !

!MaLargeCollectionManager methodsFor: 'accessing' stamp: 'cmm 6/11/2006 21:57'!
collection
	^ collection! !

!MaLargeCollectionManager methodsFor: 'private' stamp: 'cmm 3/17/2005 22:23'!
collection: aMagmaLargeCollection

	collection := aMagmaLargeCollection
! !

!MaLargeCollectionManager methodsFor: 'private' stamp: 'cmm 12/2/2004 13:22'!
collectionOid

	^ repository session oidFor: collection! !

!MaLargeCollectionManager methodsFor: 'override' stamp: 'cmm 7/12/2005 11:02'!
createNewLargeCollection
	self subclassResponsibility! !

!MaLargeCollectionManager methodsFor: 'actions' stamp: 'cmm 7/12/2005 11:18'!
createSupplementaryIndexes
	"Nothing to do by default."! !

!MaLargeCollectionManager methodsFor: 'initializing' stamp: 'cmm 10/22/2006 15:56'!
ensureMemberIndexOpen
	memberIndex ifNil:
		[ memberIndex _ self openIndex: (repository collectionFilenameFor: collection) asString.
		memberIndex filePool: repository filePool ]! !

!MaLargeCollectionManager methodsFor: 'actions' stamp: 'cmm 3/17/2005 17:22'!
ensureOpen

	^ self 
		ensureMemberIndexOpen ;
		yourself! !

!MaLargeCollectionManager methodsFor: 'recovery' stamp: 'cmm 7/9/2006 17:25'!
fileIdForIndex: anInteger
	"To support recovery, we must populate MaBeforeImageRecords with a fileId, which identifies which file it is a before-image of.  We have eight bytes available for this.  The high-order 6 bytes are my collectionOid, the low-order two bytes indicate which index of the collection it is.  0 is the memberIndex."
	^ (self collectionOid bitShift: 16) + anInteger! !

!MaLargeCollectionManager methodsFor: 'initializing' stamp: 'cmm 10/22/2006 15:09'!
filePool: aMaFileStreamPool
	"Set my pool so that Magma can limit the number simultaneously open files, something required by some ISP's."
	memberIndex filePool: aMaFileStreamPool! !

!MaLargeCollectionManager methodsFor: 'accessing' stamp: 'cmm 8/22/2002 19:25'!
filename

	^memberIndex filename! !

!MaLargeCollectionManager methodsFor: 'accessing' stamp: 'cmm 6/8/2005 15:13'!
filenames

	^ { repository directory localNameFor: self filename }! !

!MaLargeCollectionManager methodsFor: 'recovery' stamp: 'cmm 7/5/2005 23:43'!
filesDo: twoArgBlock

	memberIndex filesDo: 
		[ : each |
		twoArgBlock
			value: each
			value: (self fileIdForIndex: 0) ]! !

!MaLargeCollectionManager methodsFor: 'recovery' stamp: 'cmm 5/30/2005 21:25'!
indexForFileId: anInteger
	"To support recovery of LargeCollections, we must replace the applied records to their prior-state, which is contained inside MaBeforeImageRecords."

	^ memberIndex "the only possibility"! !

!MaLargeCollectionManager methodsFor: 'testing' stamp: 'cmm 7/9/2006 17:42'!
isLoading
	"Answer whether this collection is being loaded from another collection.  This happens only for MagmaCollections when the client requests they be sorted, it never happens for MagmaArrays."
	^ false! !

!MaLargeCollectionManager methodsFor: 'testing' stamp: 'cmm 2/25/2003 10:19'!
isOpen

	^memberIndex ~= nil! !

!MaLargeCollectionManager methodsFor: 'accessing' stamp: ''!
memberIndex

	^memberIndex
! !

!MaLargeCollectionManager methodsFor: 'override' stamp: 'cmm 3/17/2005 23:14'!
openIndex: filenameString

	self subclassResponsibility! !

!MaLargeCollectionManager methodsFor: 'override' stamp: 'cmm 3/16/2005 22:57'!
process: aMaCollectionChanges recordResultsIn: aMaCommitResult

	self subclassResponsibility! !

!MaLargeCollectionManager methodsFor: 'override' stamp: 'cmm 3/16/2005 14:17'!
processAddsIn: aMaCollectionChanges using: aMaCommitResult

	self subclassResponsibility! !

!MaLargeCollectionManager methodsFor: 'override' stamp: 'cmm 3/16/2005 21:24'!
recordConflictsBetween: aMagmaCollectionChanges and: challengingCommitLogEntry in: myCommitLogEntry
	"If there are any conflicts, record them in myCommitLogEntry."

	self subclassResponsibility! !

!MaLargeCollectionManager methodsFor: 'override' stamp: 'cmm 3/16/2005 21:24'!
recordGeneralConflictsWith: aMagmaLargeCollectionChanges in: aMaCommitLogEntry

	self subclassResponsibility! !

!MaLargeCollectionManager methodsFor: 'private' stamp: ''!
repository: aMaObjectRepository

	repository := aMaObjectRepository
! !

!MaLargeCollectionManager methodsFor: 'accessing' stamp: ''!
size

	^memberIndex size
! !

!MaLargeCollectionManager methodsFor: 'override' stamp: 'cmm 4/30/2005 11:16'!
upTo: anInteger inSegmentFromIndex: lowIndex

	self subclassResponsibility! !

MaLargeCollectionManager subclass: #MagmaArrayManager
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Magma server-private'!

!MagmaArrayManager methodsFor: 'overriding' stamp: 'cmm 6/6/2007 00:05'!
createNewLargeCollection
	| fullFilename |
	fullFilename _ repository collectionFilenameFor: collection.
	memberIndex _ MaLargeArrayOfNumbers 
		create: (
			(MaAtomicFileStream fileNamed: fullFilename)
				pool: repository filePool ;
				yourself)
		bitSize: MaObjectFiler maxDbSize! !

!MagmaArrayManager methodsFor: 'initializing' stamp: 'cmm 4/29/2007 21:48'!
openIndex: filenameString 
	^ MaLargeArrayOfNumbers open: ((MaAtomicFileStream fileNamed: filenameString)
			pool: repository filePool ;
			yourself)! !

!MagmaArrayManager methodsFor: 'actions' stamp: 'cmm 3/18/2005 10:55'!
process: aMaCollectionChanges recordResultsIn: aMaCommitResult

	self
		processAddsIn: aMaCollectionChanges
		using: aMaCommitResult.! !

!MagmaArrayManager methodsFor: 'private' stamp: 'cmm 2/13/2008 11:58'!
processAddsIn: aMagmaArrayChanges using: aMaCommitResult 
	aMagmaArrayChanges additionsDo: 
		[ : eachIndex : eachOid | 
		memberIndex 
			at: eachIndex
			put: eachOid ]! !

!MagmaArrayManager methodsFor: 'overriding' stamp: 'cmm 2/13/2008 13:35'!
recordConflictsBetween: aMagmaArrayChanges and: challengingCommitLogEntry in: myCommitLogEntry 
	"For MagmaArray's, different sessions may update the Array as long as they update different indexes."
	(challengingCommitLogEntry hasChangesToMagmaArray: aMagmaArrayChanges collectionOid) ifTrue: 
		[ | foreignChanges |
		foreignChanges := challengingCommitLogEntry magmaArrayChangesAt: aMagmaArrayChanges collectionOid.
		aMagmaArrayChanges additionsDo: 
			[ : eachIndex : eachOid | 
			((foreignChanges includesObject: eachIndex) and: [ (foreignChanges added at: eachIndex) ~= eachOid ]) ifTrue: 
				[ "this user tried to put a different oid at the same index, conflict!!"
				myCommitLogEntry 
					failResultWith: challengingCommitLogEntry connection
					on: aMagmaArrayChanges collectionOid ] ] ]! !

!MagmaArrayManager methodsFor: 'overriding' stamp: 'cmm 3/18/2005 13:33'!
recordGeneralConflictsWith: aMagmaLargeCollectionChanges in: aMaCommitLogEntry

	"no general conflicts for MagmaArray's."! !

!MagmaArrayManager methodsFor: 'overriding' stamp: 'cmm 6/5/2006 23:35'!
upTo: anInteger inSegmentFromIndex: lowIndex

	| oids |
	oids _ 
		memberIndex
			upTo: anInteger
			startingAt: lowIndex.
	^ MagmaCollectionSegment new
		lastKnownSize: memberIndex size ;
		startIndex: lowIndex ;
		objects: oids ;
		yourself! !

MaLargeCollectionManager subclass: #MagmaCollectionManager
	instanceVariableNames: 'oidOfIndexesCollection lockedBy keyIndexesByAttribute oidIndexesByAttribute keyIndexes oidIndexes loadProcess loadExecutor'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Magma server-private'!

!MagmaCollectionManager methodsFor: 'private' stamp: 'cmm 7/4/2006 15:43'!
addMember: addedObjectOid atHashes: eachHashIndexArrays 
	memberIndex 
		add: self dummyValue
		at: addedObjectOid.
	eachHashIndexArrays withIndexDo: 
		[ : eachHashValues : x | 
		eachHashValues do: 
			[ : eachHashValue | 
			(oidIndexes at: x) 
				add: addedObjectOid
				at: eachHashValue.
			(keyIndexes at: x) 
				add: eachHashValue
				at: addedObjectOid ] ]! !

!MagmaCollectionManager methodsFor: 'initialize-release' stamp: 'cmm 5/28/2007 10:26'!
addNewIndex: aMagmaIndexDescription 
	oidIndexes add: (oidIndexesByAttribute 
			at: aMagmaIndexDescription attribute
			put: (MaHashIndex 
					create: (
						(MaAtomicFileStream fileNamed: (self 
								fullFileNameStringFor: 'oids'
								of: aMagmaIndexDescription))
							pool: repository filePool ;
							yourself)
					keySize: aMagmaIndexDescription keySize
					valueSize: MaObjectBuffer oidSize
					recordSize: aMagmaIndexDescription recordSize)).
	keyIndexes add: (keyIndexesByAttribute 
			at: aMagmaIndexDescription attribute
			put: (MaHashIndex 
					create: (
						(MaAtomicFileStream fileNamed: (self 
								fullFileNameStringFor: 'keys'
								of: aMagmaIndexDescription))
							pool: repository filePool ;
							yourself)
					keySize: MaObjectBuffer oidSize
					valueSize: aMagmaIndexDescription keySize
					recordSize: aMagmaIndexDescription recordSize))! !

!MagmaCollectionManager methodsFor: 'locking' stamp: 'cmm 3/16/2005 14:13'!
beLockedBy: aMagmaClientConnection

	lockedBy _ aMagmaClientConnection! !

!MagmaCollectionManager methodsFor: 'initialize-release' stamp: 'cmm 4/13/2006 14:30'!
close
	super close.
	oidIndexes do: [ : each | each close ].
	keyIndexes do: [ : each | each close ]! !

!MagmaCollectionManager methodsFor: 'overriding' stamp: 'cmm 6/6/2007 00:03'!
createNewLargeCollection
	| fullFilename |
	fullFilename _ repository collectionFilenameFor: collection.
	memberIndex _ MaHashIndex 
		create: (
			(MaAtomicFileStream fileNamed: fullFilename)
				pool: repository filePool ;
				yourself)
		keySize: MaObjectBuffer oidSize
		valueSize: 8
		recordSize: 8! !

!MagmaCollectionManager methodsFor: 'actions' stamp: 'cmm 5/28/2007 21:21'!
createSupplementaryIndexes
	| position |
	position _ 0.
	collection descriptionsDo:
		[ : each |
		self
			ensureIndex: each
			at: (position _ position + 1) ]! !

!MagmaCollectionManager methodsFor: 'actions' stamp: 'cmm 6/13/2006 00:26'!
delete
	self filesDo: 
		[ : eachFile : eachFileId | 
		eachFile close.
		repository directory deleteFileNamed: eachFile filename ]! !

!MagmaCollectionManager methodsFor: 'private' stamp: 'cmm 7/2/2006 21:22'!
dummyValue
	"For the memberIndex, we only use the keys, answer a dummy value for the values."
	^ 0! !

!MagmaCollectionManager methodsFor: 'initialize-release' stamp: 'cmm 4/13/2006 10:05'!
ensureIndex: aMaIndexDefinition at: indexPosition 
	oidIndexes size < indexPosition ifTrue: [ self addNewIndex: aMaIndexDefinition ]! !

!MagmaCollectionManager methodsFor: 'initialize-release' stamp: 'cmm 5/28/2007 21:21'!
ensureIndexesOpen
	collection descriptionsDo: 
		[ : each | 
		(self hasIndexFor: each attribute) ifFalse: 
			[ oidIndexes add: 
				(oidIndexesByAttribute 
					at: each attribute
					put: 
						(self openIndex: 
							(self 
								fullFileNameStringFor: 'oids'
								of: each))).
			keyIndexes add: 
				(keyIndexesByAttribute 
					at: each attribute
					put: 
						(self openIndex: 
							(self 
								fullFileNameStringFor: 'keys'
								of: each))) ] ]! !

!MagmaCollectionManager methodsFor: 'actions' stamp: 'cmm 3/17/2005 17:22'!
ensureOpen

	super ensureOpen.
	self ensureIndexesOpen! !

!MagmaCollectionManager methodsFor: 'private' stamp: 'cmm 7/11/2006 21:38'!
fileIdForOids: aBoolean index: anInteger
	"To support recovery, we must populate MaBeforeImageRecords with a fileId identifying which file it is a before-image of.  We have eight bytes available for this.  The high-order 48-bits are my collectionOid, the low-order 15 bits indicate which index of the collection it is.  0 is the memberIndex.  For the other indexes (where lower-order 15 bits > 0), bit 16 is 1 for keys, 0 for oids."
	| isKeysBit indexId |
	isKeysBit _ 16.
	(anInteger maIsBitSet: isKeysBit) ifTrue: 
		[ MagmaSoftwareError signal: 'Cannot have more than ' , ((0 maBitSet: isKeysBit) - 1) printString , ' indexes for a single collection.' ].
	indexId _ aBoolean 
		ifTrue: [ anInteger ]
		ifFalse: [ anInteger maBitSet: isKeysBit ].
	^ (self collectionOid bitShift: 16) + indexId! !

!MagmaCollectionManager methodsFor: 'initialize-release' stamp: 'cmm 10/22/2006 15:10'!
filePool: aMaFileStreamPool
	super filePool: aMaFileStreamPool.
	oidIndexes do: [ : each | each filePool: aMaFileStreamPool ].
	keyIndexes do: [ : each | each filePool: aMaFileStreamPool ]! !

!MagmaCollectionManager methodsFor: 'private' stamp: 'cmm 5/28/2007 10:28'!
filenameComponentFor: aMagmaIndexDescription 
	^ self collectionOid printString , aMagmaIndexDescription attribute! !

!MagmaCollectionManager methodsFor: 'private' stamp: 'cmm 5/28/2007 10:27'!
filenameFor: keysOrOids of: aMagmaIndexDescription
	^ (String streamContents:
		[ : stream |
		stream
			nextPutAll: (self filenameComponentFor: aMagmaIndexDescription) ;
			nextPut: $- ;
			nextPutAll: keysOrOids ;
			nextPut: $. ;
			nextPutAll: MaHashIndex suggestedFileExtension ])! !

!MagmaCollectionManager methodsFor: 'accessing' stamp: 'cmm 4/13/2006 12:48'!
filenames
	^ super filenames , 
	(oidIndexes, keyIndexes collect: [ : each | repository directory localNameFor: each filename ])! !

!MagmaCollectionManager methodsFor: 'recovery' stamp: 'cmm 4/14/2006 12:41'!
filesDo: twoArgBlock 
	super filesDo: twoArgBlock.
	oidIndexes withIndexDo: 
		[ : each : x | 
		each filesDo: 
			[ : eachIndexFile | 
			twoArgBlock 
				value: eachIndexFile
				value: 
					(self 
						fileIdForOids: true
						index: x) ] ].
	keyIndexes withIndexDo: 
		[ : each : x | 
		each filesDo: 
			[ : eachIndexFile | 
			twoArgBlock 
				value: eachIndexFile
				value: 
					(self 
						fileIdForOids: false
						index: x) ] ]! !

!MagmaCollectionManager methodsFor: 'private' stamp: 'cmm 5/28/2007 10:27'!
fullFileNameStringFor: keysOrOids of: aMagmaIndexDescription 
	^ repository directory fullNameFor: 
		(self 
			filenameFor: keysOrOids
			of: aMagmaIndexDescription)! !

!MagmaCollectionManager methodsFor: 'testing' stamp: 'cmm 4/13/2006 10:05'!
hasIndexFor: attributeSymbol 
	^ oidIndexesByAttribute includesKey: attributeSymbol! !

!MagmaCollectionManager methodsFor: 'recovery' stamp: 'cmm 7/16/2006 22:41'!
indexForFileId: anInteger 
	"To support recovery of LargeCollections, we must replace the applied records to their prior-state, which is contained inside MaBeforeImageRecords."
	| indexId |
	indexId _ 2r111111111111111 bitAnd: anInteger.	"15-bit index #"
	^ indexId = 0 
		ifTrue: [ memberIndex ]
		ifFalse: 
			[ (anInteger maIsBitSet: 16) 
				ifTrue: [ keyIndexes at: indexId ]
				ifFalse: [ oidIndexes at: indexId ] ]! !

!MagmaCollectionManager methodsFor: 'initialize-release' stamp: 'cmm 8/12/2007 20:22'!
initialize
	super initialize.
	"we need them mapped to the correct attribute."
	oidIndexesByAttribute _ Dictionary new.
	keyIndexesByAttribute _ Dictionary new.
	"and we must know the correct order, as well"
	oidIndexes _ OrderedCollection new.
	keyIndexes _ OrderedCollection new! !

!MagmaCollectionManager methodsFor: 'testing' stamp: 'cmm 7/4/2006 18:34'!
isLoading
	^ loadProcess notNil and: [ loadProcess isTerminated not ]! !

!MagmaCollectionManager methodsFor: 'testing' stamp: 'cmm 3/16/2005 14:22'!
isLocked

	^lockedBy ~= nil! !

!MagmaCollectionManager methodsFor: 'accessing' stamp: 'cmm 7/19/2007 22:26'!
keysIndexForAttribute: aSymbol
	aSymbol isNil ifTrue: [ ^ self memberIndex ].
	^ keyIndexesByAttribute 
		at: aSymbol
		ifAbsent: 
			[ MagmaUserError signal: 'The keys index for ' , aSymbol , ' was not found. ' ]! !

!MagmaCollectionManager methodsFor: 'private' stamp: 'cmm 8/6/2006 21:21'!
load: aMagmaCollectionReaderSegment upTo: anInteger where: aMaExpression withObjects: aBoolean except: removedOids
	(MaQueryExecutor 
		collectionManager: self
		expression: aMaExpression) 
		load: aMagmaCollectionReaderSegment
		upTo: anInteger
		withObjects: aBoolean
		except: removedOids.
	^ aMagmaCollectionReaderSegment! !

!MagmaCollectionManager methodsFor: 'accessing' stamp: 'cmm 7/26/2006 17:39'!
loadExecutor: aMaQueryExecutor
	loadExecutor := aMaQueryExecutor! !

!MagmaCollectionManager methodsFor: 'accessing' stamp: 'cmm 7/4/2006 18:33'!
loadProcess
	^loadProcess! !

!MagmaCollectionManager methodsFor: 'accessing' stamp: 'cmm 7/4/2006 18:33'!
loadProcess: aProcess
	loadProcess := aProcess! !

!MagmaCollectionManager methodsFor: 'accessing' stamp: 'cmm 7/26/2006 20:26'!
loadProgress
	^ self isLoading 
		ifTrue: 
			[ loadExecutor 
				ifNil: [ 0 ]
				ifNotNil: [ loadExecutor fractionComplete ] ]
		ifFalse: [ 1 ]! !

!MagmaCollectionManager methodsFor: 'accessing' stamp: 'cmm 3/16/2005 14:22'!
lockedBy
	"Answers a MagmaClientConnection."

	^lockedBy! !

!MagmaCollectionManager methodsFor: 'accessing' stamp: 'cmm 3/16/2005 14:13'!
oidOfIndexesCollection: anInteger

	oidOfIndexesCollection := anInteger
! !

!MagmaCollectionManager methodsFor: 'accessing' stamp: 'cmm 7/19/2007 22:25'!
oidsIndexForAttribute: aSymbol
	aSymbol isNil ifTrue: [ ^ self memberIndex ].
	^ oidIndexesByAttribute 
		at: aSymbol
		ifAbsent: 
			[ MagmaUserError signal: 'The oids index for ' , aSymbol , ' was not found. ' ]! !

!MagmaCollectionManager methodsFor: 'overriding' stamp: 'cmm 4/29/2007 21:48'!
openIndex: aString 
	^ MaHashIndex open: ((MaAtomicFileStream fileNamed: aString)
			pool: repository filePool ;
			yourself)! !

!MagmaCollectionManager methodsFor: 'overriding' stamp: 'cmm 3/16/2005 22:57'!
process: aMaCollectionChanges recordResultsIn: aMaCommitResult

	self
		processLockIn: aMaCollectionChanges
		using: aMaCommitResult connection.
	self processRemovedIndexesIn: aMaCollectionChanges.
	self
		processAddsIn: aMaCollectionChanges
		using: aMaCommitResult.
	self
		processRemovesIn: aMaCollectionChanges;
		processChangedIndexValuesIn: aMaCollectionChanges;
		processAddNewIndexValuesIn: aMaCollectionChanges! !

!MagmaCollectionManager methodsFor: 'private' stamp: 'cmm 4/13/2006 17:34'!
processAddNewIndexValuesIn: aMagmaCollectionChanges 
	"Serialize the new values for an index that is being built."
	aMagmaCollectionChanges hasNewIndexValues ifFalse: [ ^ self ].
	aMagmaCollectionChanges newIndexValues keysAndValuesDo: 
		[ : eachAttribute : eachValues | 
		| oidsIndex keysIndex |
		oidsIndex _ self oidsIndexForAttribute: eachAttribute.
		keysIndex _ self keysIndexForAttribute: eachAttribute.
		eachValues do: 
			[ : eachAssociation | 
			eachAssociation key do: 
				[ : eachKey | 
				oidsIndex 
					add: eachAssociation value
					at: eachKey.
				keysIndex
					add: eachKey
					at: eachAssociation value ] ] ]! !

!MagmaCollectionManager methodsFor: 'private' stamp: 'cmm 6/11/2006 20:35'!
processAddsIn: aMaCollectionChanges using: aMaCommitResult 
	aMaCollectionChanges additionsDo: 
		[ : eachOid : eachHashIndexArrays | 
		| addedObjectOid |
		addedObjectOid _ (MaOidCalculator isOidForNewObject: eachOid) 
			ifTrue: [ aMaCommitResult permanentOids at: eachOid ]
			ifFalse: [ eachOid ].
		self 
			addMember: addedObjectOid
			atHashes: eachHashIndexArrays ]! !

!MagmaCollectionManager methodsFor: 'private' stamp: 'cmm 8/17/2007 21:59'!
processChangedIndexValuesIn: aMaCollectionChanges 
	aMaCollectionChanges changedKeySpecificationsDo: 
		[ : eachChangedKeySpecification | 
		| oidsIndex keysIndex |
		oidsIndex _ self oidsIndexForAttribute: eachChangedKeySpecification attribute.
		keysIndex _ self keysIndexForAttribute: eachChangedKeySpecification attribute.
		(memberIndex includesKey: eachChangedKeySpecification object) ifTrue: 
			[ eachChangedKeySpecification oldHashValues do: 
				[ : eachChangedKeySpecificationHashKeyValue | 
				oidsIndex 
					remove: eachChangedKeySpecification oid
					at: eachChangedKeySpecificationHashKeyValue.
				keysIndex
					remove: eachChangedKeySpecificationHashKeyValue 
					at: eachChangedKeySpecification oid ].
			eachChangedKeySpecification newHashValues do: 
				[ : eachChangedKeySpecificationHashKeyValue | 
				oidsIndex 
					add: eachChangedKeySpecification oid
					at: eachChangedKeySpecificationHashKeyValue.
				keysIndex
					add: eachChangedKeySpecificationHashKeyValue 
					at: eachChangedKeySpecification oid ] ] ]! !

!MagmaCollectionManager methodsFor: 'private' stamp: 'cmm 3/16/2005 22:58'!
processLockIn: aMaCollectionChanges using: aMagmaConnection

	aMaCollectionChanges hasLockSpecified
		ifTrue:
			[ aMaCollectionChanges isLocked
				ifTrue: [ self beLockedBy: aMagmaConnection ]
				ifFalse: [ self beLockedBy: nil ] ]! !

!MagmaCollectionManager methodsFor: 'private' stamp: 'cmm 4/13/2006 17:19'!
processRemovedIndexesIn: aMaCollectionChanges 
	aMaCollectionChanges removedIndexesDo: 
		[ : eachAttribute | 
		{ (self oidsIndexForAttribute: eachAttribute).
		(self keysIndexForAttribute: eachAttribute) } 
			do: [ : eachHashIndex | self removeIndex: eachHashIndex ] ]! !

!MagmaCollectionManager methodsFor: 'private' stamp: 'cmm 7/12/2006 23:29'!
processRemovesIn: aMaLargeCollectionChanges 
	aMaLargeCollectionChanges removedDo: 
		[ : eachOid : eachHashIndexValuesArray | 
		| didRemove |
		didRemove _ memberIndex 
			remove: self dummyValue
			at: eachOid.
		didRemove ifTrue: 
			[ eachHashIndexValuesArray withIndexDo: 
				[ : eachArray : x | 
				eachArray do: 
					[ : eachKeyValue | 
					(oidIndexes at: x) 
						remove: eachOid
						at: eachKeyValue.
					(keyIndexes at: x) 
						remove: eachKeyValue
						at: eachOid ] ] ] ]! !

!MagmaCollectionManager methodsFor: 'private' stamp: 'cmm 8/19/2007 21:00'!
recordConflictsBetween: aMagmaCollectionChanges and: challengingCommitLogEntry in: myCommitLogEntry 
	"Fail my result if any indexes have been added or removed since my last transaction boundary, because any adds or removes will have the wrong number of index values."
	(challengingCommitLogEntry hasCommitted: oidOfIndexesCollection) ifTrue: 
		[ myCommitLogEntry 
			failResultWith: challengingCommitLogEntry connection
			on: oidOfIndexesCollection ].
	"Fail my result if any objects being added or removed have been changed by any other session, because otherwise their index hash calculations may be incorrect."
	aMagmaCollectionChanges additionsDo: 
		[ : eachOid : eachHashIndexValues | 
		(challengingCommitLogEntry hasCommitted: eachOid) ifTrue: 
			[ myCommitLogEntry 
				failResultWith: challengingCommitLogEntry connection
				on: eachOid ] ].
	aMagmaCollectionChanges removedDo: 
		[ : eachOid : eachHashIndexValues | 
		(challengingCommitLogEntry hasCommitted: eachOid) ifTrue: 
			[ myCommitLogEntry 
				failResultWith: challengingCommitLogEntry connection
				on: eachOid ] ].
	"Fail my result if any objects being indexed have been changed by any other session, because otherwise their index hash calculations may be incorrect."
	aMagmaCollectionChanges newIndexValuesDo: 
		[ : eachOrderedCollection | 
		eachOrderedCollection do: 
			[ : eachAssocation | 
			(challengingCommitLogEntry hasCommitted: eachAssocation value) ifTrue: 
				[ myCommitLogEntry 
					failResultWith: challengingCommitLogEntry connection
					on: eachAssocation value ] ] ].
	"Fail my result if object keys being changed have been added or removed by any other session."
	challengingCommitLogEntry commitPackage allLargeCollectionChanges do: 
		[ : eachChallengingChanges | 
		eachChallengingChanges collectionOid = aMagmaCollectionChanges collectionOid ifTrue: 
			[ aMagmaCollectionChanges changedKeySpecificationsDo: 
				[ : eachChangedKeySpecification | 
				{ 
					(eachChallengingChanges added).
					(eachChallengingChanges removed)
				 } do: 
					[ : addedOrRemoved | 
					(addedOrRemoved includesKey: eachChangedKeySpecification oid) ifTrue: 
						[ myCommitLogEntry 
							failResultWith: challengingCommitLogEntry connection
							on: eachChangedKeySpecification oid ] ] ] ] ]! !

!MagmaCollectionManager methodsFor: 'private' stamp: 'cmm 9/28/2008 15:11'!
recordGeneralConflictsWith: aMagmaLargeCollectionChanges in: aMaCommitLogEntry

	"If its locked by someone else, no changes are allowed."
	(self isLocked
	and: [ aMaCommitLogEntry connection sessionId ~= lockedBy sessionId ])
		ifTrue:
			[ aMaCommitLogEntry
				failResultWith: self lockedBy
				on: aMagmaLargeCollectionChanges collectionOid ]! !

!MagmaCollectionManager methodsFor: 'private' stamp: 'cmm 6/1/2007 14:13'!
removeIndex: aMaHashIndex 
	aMaHashIndex
		close ;
		filesDo: [ : each | repository queueForDelete: each ]! !

!MagmaCollectionManager methodsFor: 'testing' stamp: 'cmm 8/12/2007 20:22'!
shouldStopLoading
	^ repository wantsToClose! !

!MagmaCollectionManager methodsFor: 'private' stamp: 'cmm 6/5/2006 23:34'!
upTo: anInteger inSegmentFromIndex: lowIndex

	| keysAndValues |
	keysAndValues _  
		memberIndex
			upTo: anInteger
			keysAndValuesFromIndex: lowIndex.
	^ MagmaCollectionSegment new
		lastKnownSize: memberIndex size ;
		startIndex: lowIndex ;
		objects: keysAndValues first  "there are no 'values' in the member-index, the keys are the objects." ;
		yourself! !

MaObject subclass: #MaObjectFiler
	instanceVariableNames: 'directory file preMadeObjectBuffer filePositionIndex usedByteArrays primitiveAttributeAddressesMap'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Magma server-private'!

!MaObjectFiler commentStamp: 'cmm 5/25/2005 14:24' prior: 0!
I am used to cache values committed in the server.  When a value is cached, we must remember from which commitPackage it came from in case the same object is updated again by a subsequent commitPackage, in which case we would not want to uncache the value when the first commitPackage is applied.!

!MaObjectFiler class methodsFor: 'creation' stamp: 'cmm 9/24/2004 12:52'!
create: aFileDirectory

	^ self new create: aFileDirectory! !

!MaObjectFiler class methodsFor: 'accessing' stamp: 'cmm 4/28/2005 11:36'!
maxDbSize
	"Answers the maximum size of a Magma database as a power of 2.  Hmm, wow, 281 Terabytes huh..  This seems to be beyond any practical limits for a single repository; at least in year 2005, so we'll go with this for now."

	^ 48! !

!MaObjectFiler class methodsFor: 'accessing' stamp: 'cmm 5/10/2005 19:04'!
objectPointersFilename

	^ 'objects.idx'! !

!MaObjectFiler class methodsFor: 'accessing' stamp: 'cmm 4/23/2007 00:49'!
objectsFilename
	^ 'objects.dat'! !

!MaObjectFiler class methodsFor: 'accessing' stamp: 'cmm 4/23/2007 15:08'!
objectsPhysicalFilename
	"Since we're using MaTransactionalFileStreams, they also save multiple numbered files, so that unlimited repository sizes are possible."
	^ 'objects.1.dat'! !

!MaObjectFiler class methodsFor: 'creation' stamp: 'cmm 7/18/2005 13:23'!
open: aFileDirectory

	^ self new
		directory: aFileDirectory ;
		open ;
		yourself! !

!MaObjectFiler class methodsFor: 'accessing' stamp: ''!
sizeOfEmptyRepository
	"Answer in bytes."

	^1024
! !

!MaObjectFiler methodsFor: 'accessing' stamp: 'kph 6/18/2007 12:14'!
anchorOid

	^ self primitiveAttributeAt: #anchorOid! !

!MaObjectFiler methodsFor: 'building' stamp: 'kph 6/18/2007 12:15'!
anchorOid: anInteger

	^ self
		primitiveAttributeAt: #anchorOid
		put: anInteger! !

!MaObjectFiler methodsFor: 'private' stamp: 'cmm 11/24/2004 15:33'!
append: aMaObjectBuffer to: aMaSerializedGraphBuffer withDataAt: filePositionInteger

	| startPos |

	startPos _ aMaSerializedGraphBuffer bufferSize.
	self
		ensureByteArrayIn: aMaObjectBuffer
		and: aMaSerializedGraphBuffer
		canAccomodate: self trackSize + aMaObjectBuffer bufferPositionOfBytesToStore
		bytesFrom: startPos + 1.
	aMaSerializedGraphBuffer increaseBufferSizeBy:
		(self
			read: self trackSize
			bytesInto: aMaObjectBuffer
			and: aMaSerializedGraphBuffer
			startingAt: startPos
			filePosition: filePositionInteger)! !

!MaObjectFiler methodsFor: 'actions' stamp: 'cmm 1/25/2005 14:10'!
appendObject: oidInteger into: aMaSerializedGraphBuffer
	"Read a single object into aMaSerializedGraphBuffer and adjust its bufferSize accordingly."

	| buffer |
	preMadeObjectBuffer
		byteArray: aMaSerializedGraphBuffer byteArray;
		startPos: aMaSerializedGraphBuffer bufferSize.
	self
		append: preMadeObjectBuffer
		to: aMaSerializedGraphBuffer
		withDataAt: (self filePointerForOid: oidInteger).
	buffer _ preMadeObjectBuffer asSentientBuffer.
	"Set oid only after append because that is where the need for growth is checked."
	buffer oid: oidInteger.
	^buffer
! !

!MaObjectFiler methodsFor: 'private' stamp: 'cmm 10/24/2008 00:06'!
booleanFlags
	^ self primitiveAttributeAt: #booleanFlags! !

!MaObjectFiler methodsFor: 'private' stamp: 'cmm 10/24/2008 00:06'!
booleanFlags: anInteger 
	^ self 
		primitiveAttributeAt: #booleanFlags
		put: anInteger! !

!MaObjectFiler methodsFor: 'accessing' stamp: 'cmm 10/23/2008 17:58'!
branchCode
	"The branchCode is a randomly-generated number used to prevent accidentally applying the wrong commit-logs to a repository.  When performing a forward recovery, each CommitLogRecord's #branchCode must match mine.  Every time a commitRestore is performed, this branchCode is updated to a new number."
	^ self primitiveAttributeAt: #branchCode
! !

!MaObjectFiler methodsFor: 'private' stamp: 'cmm 10/23/2008 17:53'!
branchCode: anInteger 
	^ self 
		primitiveAttributeAt: #branchCode
		put: anInteger! !

!MaObjectFiler methodsFor: 'debugging' stamp: 'cmm 3/26/2007 23:09'!
bufferFor: oidInteger 
	| buffer |
	buffer _ MaSerializedGraphBuffer new: 400.
	self 
		appendObject: oidInteger
		into: buffer.
	^ buffer rootBuffer trimByteArray! !

!MaObjectFiler methodsFor: 'accessing' stamp: 'kph 6/18/2007 12:14'!
classDefinitionsOid

	^self primitiveAttributeAt: #classDefinitionsOid
! !

!MaObjectFiler methodsFor: 'building' stamp: 'kph 6/18/2007 12:15'!
classDefinitionsOid: anInteger

	self
		primitiveAttributeAt: #classDefinitionsOid
		put: anInteger
! !

!MaObjectFiler methodsFor: 'actions' stamp: 'cmm 9/24/2004 13:22'!
close

	file ifNotNil:
		[ file closed ifFalse: [ file close ].
		file _ nil ].
	filePositionIndex ifNotNil:
		[ filePositionIndex close.
		filePositionIndex _ nil ]! !

!MaObjectFiler methodsFor: 'accessing' stamp: 'kph 6/18/2007 12:15'!
commitNumber
	^ self primitiveAttributeAt: #commitNumber! !

!MaObjectFiler methodsFor: 'building' stamp: 'kph 6/18/2007 12:15'!
commitNumber: anInteger 
	self 
		primitiveAttributeAt: #commitNumber
		put: anInteger! !

!MaObjectFiler methodsFor: 'private' stamp: 'cmm 9/24/2004 16:42'!
create: aFileDirectory

	self
		directory: aFileDirectory assureExistence ;
		createDataFile ;
		createFilePositionIndex.
	^ self! !

!MaObjectFiler methodsFor: 'private' stamp: 'cmm 8/15/2008 12:23'!
createDataFile
	(directory fileExists: self class objectsPhysicalFilename) ifTrue: 
		[ MagmaEnvironmentError signal: directory pathName , ' already exists.' ].
	self openObjectsFile.
	"Write some extra blank space to allow for possible future fields."
	file 
		maWrite: self class sizeOfEmptyRepository
		bytesFromPosition: 1
		of: (ByteArray new: self class sizeOfEmptyRepository)
		atFilePosition: 0; 
		maCommit.
	self
		restoreMode: false ;
		classDefinitionsOid: MaOidCalculator oidForNil ;
		definitionOid: MaOidCalculator oidForNil ;
		signature: MaObjectRepository recognitionSignature ;
		version: MagmaSession magmaVersion ;
		commitNumber: 0 ;
		initializeBranchCode ! !

!MaObjectFiler methodsFor: 'private' stamp: 'cmm 4/29/2007 21:48'!
createFilePositionIndex
	filePositionIndex _ MaLargeArrayOfNumbers 
		create: (MaAtomicFileStream fileNamed: self filePositionIndexFileName)
		bitSize: self class maxDbSize! !

!MaObjectFiler methodsFor: 'accessing' stamp: 'kph 6/18/2007 12:15'!
definitionOid

	^ self primitiveAttributeAt: #definitionOid! !

!MaObjectFiler methodsFor: 'building' stamp: 'kph 6/18/2007 12:15'!
definitionOid: anInteger

	^self
		primitiveAttributeAt: #definitionOid
		put: anInteger
! !

!MaObjectFiler methodsFor: 'accessing' stamp: 'cmm 9/26/2004 20:25'!
directory
	^ directory! !

!MaObjectFiler methodsFor: 'private' stamp: 'cmm 9/24/2004 12:55'!
directory: aFileDirectory

	directory _ aFileDirectory! !

!MaObjectFiler methodsFor: 'private' stamp: 'cmm 12/30/2002 17:28'!
ensureByteArrayIn: aMaObjectBuffer and: aMaSerializedGraphBuffer canAccomodate: anInteger bytesFrom: startPos

	(startPos + anInteger) > aMaObjectBuffer byteArray size
		ifTrue:
			[ | bigByteArray beforeSize afterSize |
			beforeSize _ aMaObjectBuffer byteArray size.
			afterSize _ (aMaObjectBuffer byteArray size * 2) max: startPos + anInteger.
			bigByteArray _ ByteArray new: afterSize.
			bigByteArray
				replaceFrom: 1
				to: beforeSize
				with: aMaSerializedGraphBuffer byteArray
				startingAt: 1.
			aMaObjectBuffer byteArray: bigByteArray.
			aMaSerializedGraphBuffer byteArray: bigByteArray ]! !

!MaObjectFiler methodsFor: 'private' stamp: 'cmm 6/1/2005 23:57'!
ensureFilesOpen

	file ifNotNil: [ file closed ifTrue: [ self openObjectsFile ] ].
	filePositionIndex ifNil: [ self openFilePositionIndex ]! !

!MaObjectFiler methodsFor: 'accessing' stamp: 'cmm 3/9/2009 22:28'!
filePointerForOid: oidInteger 
	| answer |
	answer := filePositionIndex 
		at: (MaOidCalculator indexPositionForOid: oidInteger)
		ifAbsent: 
			[ MagmaInvalidOid signal: oidInteger printString , ' is not a valid oid.' ].
	answer = 0 ifTrue: 
		[ MagmaInvalidOid signal: oidInteger printString , ' has been reclaimed.' ].
	^ answer! !

!MaObjectFiler methodsFor: 'accessing' stamp: 'cmm 9/24/2004 13:07'!
filePositionIndexFileName

	^ directory fullNameFor: self class objectPointersFilename! !

!MaObjectFiler methodsFor: 'recovery' stamp: 'cmm 7/5/2005 23:39'!
filesDo: twoArgBlock

	twoArgBlock
		value: file 
		value: MaApplyRecord objectsFileId.
	filePositionIndex filesDo:
		[ : each |
		twoArgBlock
			value: each
			value: MaApplyRecord filePositionsFileId ]! !

!MaObjectFiler methodsFor: 'accessing' stamp: 'cmm 5/27/2007 15:08'!
getNextOid
	^ MaOidCalculator oidForFilePositionIndexPosition: filePositionIndex size! !

!MaObjectFiler methodsFor: 'actions' stamp: 'cmm 12/2/2004 13:45'!
index: oidInteger at: filePositionInteger

	^filePositionIndex
		at: (MaOidCalculator indexPositionForOid: oidInteger)
		put: filePositionInteger! !

!MaObjectFiler methodsFor: 'private' stamp: 'cmm 10/21/2004 12:53'!
initialize

	super initialize.
	preMadeObjectBuffer _ MaObjectBuffer new.
	usedByteArrays _ (1 to: 8) collect: [ :each | ByteArray new: each ].
	self initializePrimitiveAttributeAddressesMap! !

!MaObjectFiler methodsFor: 'accessing' stamp: 'cmm 2/24/2009 14:37'!
initializeBranchCode
	| oldBranchCode newBranchCode |
	oldBranchCode := self branchCode.
	[ (newBranchCode := (1 to: (1 bitShift: 32) - 1) atRandom) = oldBranchCode ] whileTrue.
	self branchCode: newBranchCode! !

!MaObjectFiler methodsFor: 'private' stamp: 'cmm 10/23/2008 17:57'!
initializePrimitiveAttributeAddressesMap

	primitiveAttributeAddressesMap _ Dictionary new
		at: #signature put: (Association key: 0 value: 64);
		at: 0 put: #signature ;

		at: #version put: (Association key: 8 value: 16);
		at: 2 put: #version ;

		at: #booleanFlags put: (Association key: 10 value: 8);
		at: 10 put: #booleanFlags ;

		at: #definitionOid put: (Association key: 11 value: 64);
		at: 11 put: #definitionOid ;

		"gap 19-26"
		at: #classDefinitionsOid put: (Association key: 27 value: 64) ;
		at: 27 put: #classDefinitionsOid ;

		at: #anchorOid put: (Association key: 43 value: 64);
		at: 43 put: #anchorOid ;

		at: #commitNumber put: (51 -> 48) ;
		at: 51 put: #commitNumber ;

		at: #branchCode put: (57->32) ;
		at: 57 put: #branchCode ;

		yourself! !

!MaObjectFiler methodsFor: 'accessing' stamp: 'cmm 5/1/2003 17:37'!
isOpen

	^file notNil and: [ file closed not ]! !

!MaObjectFiler methodsFor: 'utilities' stamp: 'cmm 3/26/2007 23:10'!
objectBuffersDo: oneArgBlock 
	MaOidCalculator firstUserObjectOid 
		to: self getNextOid
		do: 
			[ : eachOid | 
			| buffer |
			[ buffer _ self bufferFor: eachOid.
			oneArgBlock value: buffer trimByteArray ] 
				on: MagmaInvalidOid
				do: [ : err |  ] ]! !

!MaObjectFiler methodsFor: 'utilities' stamp: 'cmm 3/26/2007 22:10'!
objectFileOnlyBuffersDo: oneArgBlock 
	| graphBuffer nextStart |
	nextStart _ self class sizeOfEmptyRepository.
	[ nextStart < file size ] whileTrue: 
		[ graphBuffer _ MaSerializedGraphBuffer new: 400.
		preMadeObjectBuffer
			byteArray: graphBuffer byteArray ;
			startPos: graphBuffer headerSize.
		self 
			append: preMadeObjectBuffer
			to: graphBuffer
			withDataAt: nextStart.
		oneArgBlock value: graphBuffer.
		nextStart _ nextStart + preMadeObjectBuffer physicalSize - (MaObjectBuffer oidSize / 8) ]! !

!MaObjectFiler methodsFor: 'accessing' stamp: 'cmm 4/21/2003 16:29'!
objectFileSize

	^file size! !

!MaObjectFiler methodsFor: 'accessing' stamp: 'cmm 2/21/2008 12:02'!
open
	directory exists ifFalse: [ MagmaEnvironmentError signal: 'Database Directory Not Found' ].
	(directory fileExists: self class objectsPhysicalFilename) ifFalse: 
		[ "This might be a version 8, check if we have the old filename, 'objects'"
		(directory fileExists: 'objects') 
			ifFalse: 
				[ MagmaEnvironmentError signal: (directory fullNameFor: self class objectsPhysicalFilename) , ' not found.' ] ].
	self openObjectsFile.
	
	[ self
		validateSignature ;
		openFilePositionIndex ] ifCurtailed: [ self close ]! !

!MaObjectFiler methodsFor: 'private' stamp: 'cmm 4/29/2007 21:48'!
openFile: unqualifiedName 
	^ MaAtomicFileStream fileNamed: (directory fullNameFor: unqualifiedName)! !

!MaObjectFiler methodsFor: 'private' stamp: 'cmm 4/29/2007 21:48'!
openFilePositionIndex
	filePositionIndex _ 
	[ MaLargeArrayOfNumbers open: (MaAtomicFileStream fileNamed: self filePositionIndexFileName) ] ifCurtailed: [ self close ]! !

!MaObjectFiler methodsFor: 'private' stamp: 'cmm 5/10/2005 19:08'!
openObjectsFile

	file _ self openFile: self class objectsFilename! !

!MaObjectFiler methodsFor: 'private' stamp: 'cmm 12/30/2002 21:15'!
primitiveAttributeAddressesMap

	^primitiveAttributeAddressesMap! !

!MaObjectFiler methodsFor: 'accessing' stamp: 'kph 6/18/2007 12:14'!
primitiveAttributeAt: aSymbol

	| assoc |

	^self
		uint: (assoc _ self primitiveAttributeAddressesMap at: aSymbol) value
		at: assoc key
! !

!MaObjectFiler methodsFor: 'private' stamp: 'kph 6/18/2007 12:13'!
primitiveAttributeAt: aSymbol put: anInteger

	| assoc |

	self
		uint: (assoc _ self primitiveAttributeAddressesMap at: aSymbol) value
		at: assoc key
		put: anInteger
! !

!MaObjectFiler methodsFor: 'private' stamp: 'cmm 6/16/2005 10:23'!
read: numberOfBytes bytesInto: aMaObjectBuffer and: aMaSerializedGraphBuffer startingAt: startPos filePosition: anInteger
	"startPos is 0-based, answer the physicalSize of the buffer read."

	| physicalSize numberOfBytesRead numberOfBytesStored |
	numberOfBytesRead _
		file
			maRead: numberOfBytes
			bytesFromPosition: startPos + aMaObjectBuffer bufferPositionOfBytesToStore
			of: aMaObjectBuffer byteArray
			atFilePosition: anInteger.
	"See if we have read enough, if not read the rest of the buffer."
	numberOfBytesStored _ aMaObjectBuffer numberOfBytesToStore.
	physicalSize _ aMaObjectBuffer physicalSize.
	numberOfBytesStored > numberOfBytesRead
		ifTrue:
			[ | resumePosition |
			resumePosition _ startPos + numberOfBytesRead + aMaObjectBuffer bufferPositionOfBytesToStore.
			self
				ensureByteArrayIn: aMaObjectBuffer
				and: aMaSerializedGraphBuffer
				canAccomodate: physicalSize
				bytesFrom: resumePosition.
			file
				next: numberOfBytesStored - numberOfBytesRead
				into: aMaSerializedGraphBuffer byteArray
				startingAt: resumePosition ].
	^ physicalSize! !

!MaObjectFiler methodsFor: 'accessing' stamp: 'cmm 8/27/2008 23:13'!
recordSampleAt: filePointer do: oneArgBlock 
	"Value oneArgBlock with the header and enough bytes for the first two oid pointers of MaObjectBuffer currently residing at filePointer."
	preMadeObjectBuffer startPos: 0.
	file 
		maRead: 32 "(MagmaBufferPositionMap headerSize - preMadeObjectBuffer bufferPositionOfBytesToStore + 1) + (2*(MaObjectBuffer oidSize/8))"
		bytesFromPosition: preMadeObjectBuffer bufferPositionOfBytesToStore
		of: preMadeObjectBuffer byteArray
		atFilePosition: filePointer.
	^ oneArgBlock value: preMadeObjectBuffer! !

!MaObjectFiler methodsFor: 'accessing' stamp: 'cmm 6/18/2007 23:17'!
restoreMode
	^ self booleanFlags maIsBitSet: 1! !

!MaObjectFiler methodsFor: 'accessing' stamp: 'cmm 6/19/2007 21:53'!
restoreMode: aBoolean 
	^ self booleanFlags: 
		(aBoolean 
			ifTrue: [ self booleanFlags maBitSet: 1 ]
			ifFalse: [ self booleanFlags maBitClear: 1 ])! !

!MaObjectFiler methodsFor: 'accessing' stamp: 'kph 6/18/2007 12:16'!
signature

	^self primitiveAttributeAt: #signature
! !

!MaObjectFiler methodsFor: 'private' stamp: 'kph 6/18/2007 12:17'!
signature: anInteger

	self
		primitiveAttributeAt: #signature
		put: anInteger
! !

!MaObjectFiler methodsFor: 'private' stamp: 'cmm 8/12/2002 23:48'!
trackSize
	"This is the minimum number of bytes to read from disk for any object.
	This particular value allows up to 32 instVars to be read at once."

	"^MaAbstractObjectBuffer addressSize * 32 + MaNormalObjectBuffer headerSize"
	^280 "same as above, but faster.."! !

!MaObjectFiler methodsFor: 'private' stamp: 'cmm 9/24/2004 13:22'!
uint: bits at: anInteger

	| address bytes |

	address _ self usedByteArray: (bytes _ bits / 8).
	file
		maRead: bytes
		bytesFromPosition: 1
		of: address
		atFilePosition: anInteger.
	^address
		maUint: bits
		at: 0
! !

!MaObjectFiler methodsFor: 'private' stamp: 'cmm 5/16/2007 10:19'!
uint: bits at: filePointer put: anInteger 
	| ba bytes |
	ba _ self usedByteArray: (bytes _ bits / 8).
	ba 
		maUint: bits
		at: 0
		put: anInteger.
	file 
		maWrite: bytes
		bytesFromPosition: 1
		of: ba
		atFilePosition: filePointer! !

!MaObjectFiler methodsFor: 'recovery' stamp: 'cmm 6/5/2005 18:24'!
unapplyFilePositionsFileUsing: aMaApplyRecord

	filePositionIndex unapplyUsing: aMaApplyRecord! !

!MaObjectFiler methodsFor: 'recovery' stamp: 'cmm 4/10/2007 22:35'!
unapplyObjectsFileUsing: aMaApplyRecord 
	aMaApplyRecord unapplySelf: file! !

!MaObjectFiler methodsFor: 'recovery' stamp: 'cmm 6/13/2005 21:50'!
unapplyUsing: aMaApplyRecord
	"Reapply the contents of aMaBeforeImageRecord to the file it came from."

	aMaApplyRecord fileId
		caseOf:
			{ [ MaApplyRecord objectsFileId ] -> [ self unapplyObjectsFileUsing: aMaApplyRecord ].
			[ MaApplyRecord filePositionsFileId ] -> [ self unapplyFilePositionsFileUsing: aMaApplyRecord ] }
		otherwise:
			[ MagmaSoftwareError signal: 'unknown fileId' ]! !

!MaObjectFiler methodsFor: 'private' stamp: 'cmm 7/16/2002 00:08'!
usedByteArray: size

	^usedByteArrays at: size
! !

!MaObjectFiler methodsFor: 'actions' stamp: 'cmm 2/21/2008 12:04'!
validateSignature
	self signature = MaObjectRepository recognitionSignature ifFalse: 
		[ MagmaUserError signal: 'This does not appear to be a Magma database.' ]! !

!MaObjectFiler methodsFor: 'debugging' stamp: 'cmm 3/25/2007 14:39'!
validateSpecialOid: selectorSymbol 
	((self perform: selectorSymbol) 
		between: MaOidCalculator firstUserObjectOid
		and: MaOidCalculator firstUserObjectOid + 10000) ifFalse: 
		[ MagmaCorruptionError signal: 'The ' , selectorSymbol , ' does not appear valid' ]! !

!MaObjectFiler methodsFor: 'actions' stamp: 'cmm 8/29/2008 11:47'!
validateVersion
	"10 can open 9.  All others must match exactly."
	self version = 9 
		ifTrue: 
			[ MagmaWarning signal: 'This repository is version 9 but you are running version 10.  The repository will be upgraded to version 10, after which this version 10 code base must be used to access it.'.
			self version: 10 ]
		ifFalse: 
			[ ({  (MagmaSession magmaVersion)  } includes: self version) ifFalse: 
				[ MagmaUserError signal: 'This database requires code version ' , self version printString , '.  You are running version ' , MagmaSession magmaVersion printString ] ]! !

!MaObjectFiler methodsFor: 'accessing' stamp: 'kph 6/18/2007 12:17'!
version

	^self primitiveAttributeAt: #version! !

!MaObjectFiler methodsFor: 'private' stamp: 'kph 6/18/2007 12:17'!
version: anInteger

	self
		primitiveAttributeAt: #version
		put: anInteger
! !

!MaObjectFiler methodsFor: 'private' stamp: 'cmm 1/19/2005 15:40'!
write: aMaObjectBuffer at: filePosition
	"Note that the oid is stored in the filePositionsIndex, so there is no need to store it in the data file too.  The oid is the first 8-bytes of the buffer."

	file
		maWrite: aMaObjectBuffer numberOfBytesToStore
		bytesFromPosition:
			aMaObjectBuffer startPos + aMaObjectBuffer bufferPositionOfBytesToStore
		of: aMaObjectBuffer byteArray
		atFilePosition: filePosition! !

!MaObjectFiler methodsFor: 'actions' stamp: 'cmm 7/24/2007 17:33'!
write: aMaObjectBuffer index: aBoolean 
	"write a single object encapsulated by aMaObjectBuffer."
	| filePosition |
	filePosition _ aBoolean 
		ifTrue: 
			[ aMaObjectBuffer preWriteAdjustmentFrom: nil.
			self 
				index: aMaObjectBuffer oid
				at: file size ]
		ifFalse: 
			[ | pos |
			pos _ self filePointerForOid: aMaObjectBuffer oid.
			self 
				recordSampleAt: pos
				do: 
					[ : existingBuffer | 
					aMaObjectBuffer preWriteAdjustmentFrom: existingBuffer.
					existingBuffer physicalSize = aMaObjectBuffer physicalSize 
						ifTrue: [ pos ]
						ifFalse: 
							[ self 
								index: aMaObjectBuffer oid
								at: file size ] ] ].
	self 
		write: aMaObjectBuffer
		at: filePosition! !

MaObject subclass: #MaObjectRepository
	instanceVariableNames: 'transactionLog sessions filer repositoryController largeCollectionManagers systemReadStrategy nextOid recoveryManager commitGuard applyProcess wantsToClose filePool flushGuard restoreMode'
	classVariableNames: 'SimulateOutage'
	poolDictionaries: ''
	category: 'Magma server-private'!

!MaObjectRepository class methodsFor: 'creation' stamp: 'cmm 9/24/2004 12:50'!
create: aFileDirectory controller: aMaRepositoryController

	^ (self new repositoryController: aMaRepositoryController) create: aFileDirectory! !

!MaObjectRepository class methodsFor: 'class initialization' stamp: 'cmm 11/7/2008 16:01'!
initialize
	super initialize.
	SimulateOutage := false! !

!MaObjectRepository class methodsFor: 'accessing' stamp: 'cmm 5/29/2007 22:35'!
oldRecognitionSignature
	^ 5191969! !

!MaObjectRepository class methodsFor: 'creation' stamp: 'cmm 7/18/2005 13:23'!
open: aFileDirectory controller: aMaRepositoryController

	^self new
		repositoryController: aMaRepositoryController ;
		open: aFileDirectory ;
		yourself! !

!MaObjectRepository class methodsFor: 'accessing' stamp: 'cmm 4/2/2007 17:38'!
recognitionSignature
	"('cmmuller' asArray collect: [ : e | e asciiValue * 2 ]) asByteArray maUint: 64 at: 0"
	^ 16486227812107868870! !

!MaObjectRepository methodsFor: 'client-requests' stamp: 'cmm 1/18/2005 22:39'!
abortTransactionFor: sessionId

	^ transactionLog 
		abortTransactionEntryFor: sessionId
		repository: self! !

!MaObjectRepository methodsFor: 'high-availability' stamp: 'cmm 3/3/2009 16:53'!
applyCommitRecords: commitRecords 
	"Private - apply commitRecords to this repository."
	| branchCode |
	branchCode := self branchCode.
	self flushCritical: 
		[ commitRecords do: 
			[ : each | 
			| commitLogEntry |
			commitLogEntry := self recoveryManager materializeCommitLogEntryIn: each.
			"Note, we've already validated our branchCode (the sender of this method), so if we notice a change in the branchCode in a commit-log, we can safely assume the repository branched at that point, therefore we can update our branchCode."
			branchCode = each branchCode ifFalse: 
				[ branchCode := each branchCode.
				self branchCode: branchCode ].
			self 
				write: commitLogEntry
				logged: true
				commitLogRecord: each.
			self session advanceWith: commitLogEntry ] ].
	^ branchCode! !

!MaObjectRepository methodsFor: 'actions' stamp: 'cmm 5/27/2007 15:20'!
applyToCache: aMaCommitLogEntry 
	| newObjects |
	newObjects _ aMaCommitLogEntry result newObjectBuffers.
	aMaCommitLogEntry commitPackage objectsDo: 
		[ : eachBuffer | 
		filer 
			write: eachBuffer
			index: (newObjects includesKey: eachBuffer oid) ].
	"If writing regular objects caused any new LargeCollections we need to refresh my sessions view of the repository so I know about those collections because I'm about to update them, except of course unless aMaCommitLogEntry is for the servers session."
	(aMaCommitLogEntry isSystem not and: 
		[ aMaCommitLogEntry commitPackage shouldServerRefresh or: [ aMaCommitLogEntry shouldServerRefresh ] ]) ifTrue: [ repositoryController refreshView ].
	self 
		updateLargeCollectionsIn: aMaCommitLogEntry commitPackage
		recordResultsIn: aMaCommitLogEntry result.
	filer commitNumber: recoveryManager commitNumber! !

!MaObjectRepository methodsFor: 'backup / restore' stamp: 'cmm 8/1/2007 22:16'!
archiveCommitLog
	recoveryManager archiveCommitLog! !

!MaObjectRepository methodsFor: 'high-availability' stamp: 'cmm 11/15/2008 11:34'!
beginRestore
	"Set my restoreMode bit, indicating I can only receive commitLogRecords in order from the primary.  This should only be done in conjunction with a corresponding change in the Node."
	self restoreMode 
		ifTrue: [ MagmaUserError signal: 'Already in restoreMode' ]
		ifFalse: [ self restoreMode: true ]! !

!MaObjectRepository methodsFor: 'accessing' stamp: 'cmm 11/18/2008 12:54'!
branchCode
	"The branchCode is a randomly-generated number used to prevent accidentally applying the wrong commit-logs to a repository.  When performing a forward recovery, each CommitLogRecord's #branchCode must match mine.  Every time a commitRestore is performed, this branchCode is updated to a new number."
	^ filer branchCode! !

!MaObjectRepository methodsFor: 'accessing' stamp: 'cmm 2/27/2009 15:31'!
branchCode: anInteger 
	"Private - Do not update the branchCode from outside!!"
	filer branchCode: anInteger.
	recoveryManager ifNotNil: [ recoveryManager branchCode: anInteger ]! !

!MaObjectRepository methodsFor: 'accessing' stamp: 'cmm 3/19/2005 10:59'!
byteArrayAt: oidInteger using: aMaReadStrategy

	^ (MaOidCalculator isOidForUserObject: oidInteger)
		ifTrue:
			[ (self
				graphBufferAt: oidInteger
				using: aMaReadStrategy) trimmedByteArray ]
		ifFalse:
			[ "When could this happen?  (Sigh), when they put an atomic in a MagmaLargeCollection."
			(self session serializer serializeGraph:
				(MaOidCalculator
					objectWithOid: oidInteger
					ifNone: [ MagmaSoftwareError signal: 'unknown oid' ])) trimmedByteArray ]! !

!MaObjectRepository methodsFor: 'recovery' stamp: 'cmm 8/13/2007 22:22'!
captureBeforeImages
	"Capture a current description of the actual files in the places where we intend to change them.  We capture whether a file is new, whether it will be extended in length, and what file positions will be overwritten and what contents are there presently."
	self openDbFilesDo: 
		[ : eachAtomicFileStream : eachFileId | 
		recoveryManager 
			writeRollbackRecordsFor: eachAtomicFileStream
			fileId: eachFileId ].
	recoveryManager flush! !

!MaObjectRepository methodsFor: 'high-availability' stamp: 'cmm 3/8/2009 12:33'!
catch: thisBackupLocation upTo: primaryLocation 
	"Request batches of commit-log records and apply them to my repository.  When this method finishes successfully I will be caught up and the primary will be sending me the updates automatically."
	| session |
	session := primaryLocation newAdminSession.
	
	[ 
	[ 
	[ | serverResponse commitRecords done |
	repositoryController requestCritical: 
		[ serverResponse := session 
			warmBackupUpdatesFrom: self commitNumber + 1
			for: thisBackupLocation.
		done := serverResponse value.
		commitRecords := serverResponse key.
		self applyCommitRecords: commitRecords.
		done ] ] whileFalse.
	"I have now been made a secondary in the Node 'locations', refresh so I know about it."
	repositoryController requestCritical: 
		[ self session abort.
		"Make absolutely sure the node locations are materialized."
		self session node printString ] ] 
		on: MagmaUnavailableCommitRecord
		do: 
			[ : err | 
			"The warm backup must be re-established, page the DBA?"
			"pass the buck for now"
			err pass ] ] ensure: [ session ifNotNil: [ session disconnect ] ]! !

!MaObjectRepository methodsFor: 'actions' stamp: 'cmm 9/20/2008 14:11'!
challengeOthersWith: aMaCommitLogEntry 
	| entriesToRemove |
	entriesToRemove := transactionLog challengeEntriesWith: aMaCommitLogEntry.
	entriesToRemove ifNotNil: 
		[ entriesToRemove do: [ : each | self removeSessionFor: each sessionId ] ]! !

!MaObjectRepository methodsFor: 'accessing' stamp: 'cmm 12/2/2004 13:41'!
classDefinitionsByteArray
	"When initially creating the repository, before even the system root is
	written, the classDefinitionsOid will be a new-object oid."

	| loid |

	^MaOidCalculator oidForNil = (loid _ filer classDefinitionsOid)
		ifTrue:
			[ (self serverSerializer serializeGraph: repositoryController definition classDefinitions) trimmedByteArray ]
		ifFalse:
			[ self
				byteArrayAt: loid
				using: MaBasicReadStrategy deep "we need it all" ]! !

!MaObjectRepository methodsFor: 'actions' stamp: 'cmm 6/12/2007 21:52'!
close
	self isOpen ifFalse: [ ^ self ].
	wantsToClose _ true.
	[ applyProcess notNil and: [ applyProcess isTerminated not ] ] whileTrue: [ (Delay forSeconds: 1) wait ].
	[ largeCollectionManagers anySatisfy: [ : each | each isLoading ] ] whileTrue: [ (Delay forMilliseconds: 500) wait ].
	self commitCritical: 
		[ self isOpen ifTrue: [ self flushCritical: [ self flushCache	"just to be sure" ] ] ].
	filer ifNotNil: [ filer close ].
	largeCollectionManagers do: [ : each | each close ].
	recoveryManager ifNotNil: [ recoveryManager close ]! !

!MaObjectRepository methodsFor: 'accessing' stamp: 'cmm 4/26/2007 23:38'!
collectionFilenameFor: aMaLargeCollection
	^ self directory fullNameFor:
		(String streamContents:
			[ : stream |
			stream
				print: (self session oidFor: aMaLargeCollection) ;
				nextPutAll: '-members.' ;
				nextPutAll: aMaLargeCollection implementationClass suggestedFileExtension ])! !

!MaObjectRepository methodsFor: 'magma-collections' stamp: 'cmm 6/16/2005 10:53'!
collectionManagerFor: oidInteger

	^ self
		collectionManagerFor: oidInteger
		ifAbsent:
			[ MagmaSoftwareError signal:
				'MagmaCollection ' , oidInteger printString ,
					' not found.  MaObjectRepository>>collectionManagerFor:' ]! !

!MaObjectRepository methodsFor: 'magma-collections' stamp: 'cmm 10/25/2006 23:29'!
collectionManagerFor: oidInteger ifAbsent: aBlock 
	^ largeCollectionManagers 
		at: oidInteger
		ifAbsent: 
			[ | mc |
			mc _ self session 
				objectWithOid: oidInteger
				ifAbsent: [ ^ aBlock value ].
			self openLargeCollectionIfNecessary: mc ]! !

!MaObjectRepository methodsFor: 'actions' stamp: 'cmm 7/5/2005 12:36'!
commitCritical: aBlock

	^ commitGuard critical: aBlock! !

!MaObjectRepository methodsFor: 'accessing' stamp: 'cmm 7/15/2008 14:52'!
commitNumber
	^ filer commitNumber! !

!MaObjectRepository methodsFor: 'client-requests' stamp: 'cmm 3/14/2009 17:17'!
commitRecordAt: startNumber 
	"Answer the CommitLogRecord for startNumber."
	^ (self 
		commitRecordsFrom: startNumber
		to: startNumber) ifNotEmptyDo: [ : arr | arr first ]! !

!MaObjectRepository methodsFor: 'client-requests' stamp: 'cmm 3/14/2009 16:41'!
commitRecordsFrom: startNumber 
	"Answer a collection of MaCommitLogEntry which clients use to advance their image."
	^ self 
		commitRecordsFrom: startNumber
		to: 281474976710655	"The maximum possible commitNumber."! !

!MaObjectRepository methodsFor: 'client-requests' stamp: 'cmm 3/14/2009 17:14'!
commitRecordsFrom: startNumber to: endNumber 
	"Answer a collection of MaCommitLogEntry which clients use to advance their image."
	| startTime size answer |
	answer := OrderedCollection new.
	startTime := Time millisecondClockValue.
	size := 0.
	recoveryManager 
		commitLogRecordsFrom: startNumber
		do: 
			[ : each | 
			size := size + each record size.
			answer add: each independentCopy ]
		until: 
			[ : each | 
			answer notEmpty and: 
				[ (Time millisecondsSince: startTime) > 1000 or: 
					[ size > 2000000 or: 
						[ "2 million bytes should be plenty at a time"
						each commitNumber >= endNumber ] ] ] ].
	^ answer asArray! !

!MaObjectRepository methodsFor: 'high-availability' stamp: 'cmm 3/3/2009 21:12'!
commitRestore
	"Do NOT use this unless you know what you are doing, use the public API only.  Set whether the receiver is operating in restoreMode.  Warm backups operate in restoreMode, taking commits broadcasted to them from the 'primary' and applying them."
	self restoreMode 
		ifTrue: 
			[ self
				restoreMode: false ;
				initializeBranchCode ]
		ifFalse: [ MagmaUserError signal: 'Not in restore mode!!' ]! !

!MaObjectRepository methodsFor: 'actions' stamp: 'cmm 5/21/2007 00:08'!
create: aFileDirectory 
	filer _ MaObjectFiler create: aFileDirectory.
	self postOpenInitialize! !

!MaObjectRepository methodsFor: 'accessing' stamp: 'cmm 12/2/2004 13:43'!
definitionByteArray

	| oid |
	^MaOidCalculator oidForNil = (oid _ filer definitionOid)
		ifTrue:
			[ (self serverSerializer serializeGraph: repositoryController definition) trimmedByteArray ]
		ifFalse:
			[ self
				byteArrayAt: oid
				using: MaBasicReadStrategy deep "we need it all" ]! !

!MaObjectRepository methodsFor: 'backup / restore' stamp: 'cmm 8/3/2007 11:02'!
deleteUnnecessaryCommitLogs
	recoveryManager deleteUnnecessaryCommitLogFiles ! !

!MaObjectRepository methodsFor: 'accessing' stamp: 'cmm 9/24/2004 15:42'!
directory

	^ filer directory! !

!MaObjectRepository methodsFor: 'building' stamp: 'cmm 6/11/2006 21:30'!
ensureManagerCreatedFor: aMagmaLargeCollection 
	^ largeCollectionManagers 
		at: (self session oidFor: aMagmaLargeCollection)
		ifAbsentPut: [ aMagmaLargeCollection newManagerUsing: self ]! !

!MaObjectRepository methodsFor: 'building' stamp: 'cmm 5/14/2005 18:08'!
ensureReadStrategy: aMaReadStrategy forSession: sessionId

	| strategy |
	^(strategy _
		self
			renewCurrentReadStrategy: aMaReadStrategy
			for: sessionId)

				ifNil: [ self systemReadStrategy ] 
				ifNotNil: [ strategy ]! !

!MaObjectRepository methodsFor: 'recovery' stamp: 'cmm 6/16/2005 11:04'!
fileManagerFor: aMaApplyRecord ifPresent: oneArgBlock ifAbsent: aBlock

	| fileManager |
	fileManager _ aMaApplyRecord fileId
		caseOf:
			{ [ MaApplyRecord objectsFileId ] -> [ filer ].
			[ MaApplyRecord filePositionsFileId ] -> [ filer ] }
		otherwise:
			[ (self 
				collectionManagerFor: aMaApplyRecord collectionOid
				ifAbsent: [ ^ aBlock value ]) indexForFileId: aMaApplyRecord fileId ].
	^ oneArgBlock value: fileManager! !

!MaObjectRepository methodsFor: 'accessing' stamp: 'cmm 10/22/2006 15:39'!
filePool
	^ filePool! !

!MaObjectRepository methodsFor: 'accessing' stamp: 'cmm 10/22/2006 21:39'!
filePoolSize: anInteger 
	"Set the maximum number of simultaneously-open files that Magma will have."
	anInteger < 4 ifTrue: [ MagmaUserError signal: 'Magma requires at least 4 open files.' ].
	filePool maxSize: anInteger - 4	"-4 : 1 for objects file, 1 for file-positions index, 2 for recovery files."! !

!MaObjectRepository methodsFor: 'accessing' stamp: ''!
filer

	^filer
! !

!MaObjectRepository methodsFor: 'recovery' stamp: 'cmm 6/5/2008 21:08'!
flushCache
	"Flush all caches as an uncorruptible unit-of-work.  If this forked code completes all files are flushed.  Should a power outage occur somewhere in the middle, the recovery will reverse what had been written so far."
	"Define what we're going to do to (flush) into a block, since we have to determine whether to time it."
	| stats |
	stats := repositoryController serverStatistics.
	stats flushTimeHistory addValue: (Time millisecondsToRun: 
			[ recoveryManager beginApply.
			self
				captureBeforeImages ;
				flushIndividualCaches.
			SimulateOutage ifTrue: [ self simulateOutage ].
			recoveryManager endApply ])! !

!MaObjectRepository methodsFor: 'recovery' stamp: 'cmm 6/12/2007 21:18'!
flushCacheSoon
	"Flushing is horribly expensive, but essential to ensure we can survive a power-outage without corruption.  To balance these opposing forces, we apply and flush only at certain interval, only in the background, only the files that needed flushed, and each file only flushed once."
	applyProcess ifNil: 
		[ applyProcess _ 
		[ | targetTime |
		targetTime _ Time millisecondClockValue + (repositoryController cacheFlushFrequency * 1000).
		[ Time millisecondClockValue > targetTime or: [ wantsToClose ] ] whileFalse: [ (Delay forSeconds: 1) wait ].
		self flushCritical: 
			[ self commitCritical: 
				[ self flushCache.
				applyProcess _ nil ] ] ] fork.
		applyProcess name: 'flush Magma cache' ]! !

!MaObjectRepository methodsFor: 'backup / restore' stamp: 'cmm 6/6/2007 22:46'!
flushCritical: aBlock
	^ flushGuard critical: aBlock! !

!MaObjectRepository methodsFor: 'recovery' stamp: 'cmm 1/2/2009 20:50'!
flushIndividualCaches
	self openDbFilesDo: [ : eachFile : eachFileId | eachFile maCommit ]! !

!MaObjectRepository methodsFor: 'accessing' stamp: 'cmm 12/2/2004 13:25'!
getNextOid

	^nextOid _ nextOid + 1! !

!MaObjectRepository methodsFor: 'client-requests' stamp: 'cmm 8/6/2006 18:32'!
getTrunkFor: collectionOid expression: aMaTerm 
	| executor |
	executor _ MaQueryExecutor 
		collectionManager: (self collectionManagerFor: collectionOid)
		expression: aMaTerm.
	^ executor trunk! !

!MaObjectRepository methodsFor: 'accessing' stamp: 'cmm 1/25/2005 14:04'!
graphBufferAt: oidInteger using: aMaReadStrategy
	"Answers a MaSerializedGraphBuffer."

	| answer rootBuffer |
	answer _ MaSerializedGraphBuffer new: 400.
	rootBuffer _ 
		filer
			appendObject: oidInteger
			into: answer.
	rootBuffer
		appendGraphUsing: aMaReadStrategy
		into: answer
		currentDepth: 1
		minDepth: aMaReadStrategy minimumDepth
		with: Set new
		filer: filer.
	^ answer! !

!MaObjectRepository methodsFor: 'magma-collections' stamp: 'cmm 5/28/2007 21:21'!
hashKeysReferencing: eachOid inIndexesOf: sourceManager using: aMagmaCollectionManager 
	| hashesByIndex |
	hashesByIndex _ OrderedCollection new.
	"Normally there will be just one index at this time (building a sort result), but be generic anyway.."
	aMagmaCollectionManager collection descriptionsDo:  
		[ : each | 
		| eachKeysHdx eachHashes |
		eachHashes _ OrderedCollection new.
		eachKeysHdx _ sourceManager keysIndexForAttribute: each attribute.
		eachKeysHdx 
			keysAndValuesFrom: eachOid
			startingAt: 1
			do: [ : eachOidAgain : eachKeyLocation | eachHashes add: eachKeyLocation ]
			until: [ : eachOidAgain : eachKeyLocation | eachOidAgain > eachOid ].
		hashesByIndex add: eachHashes ].
	^ hashesByIndex! !

!MaObjectRepository methodsFor: 'notifications' stamp: 'cmm 11/2/2008 09:39'!
informDba: aString 
	"Signal DBA notification.."
	self halt: 'For informational purposes.  Does not page.'! !

!MaObjectRepository methodsFor: 'initialize-release' stamp: 'cmm 11/2/2008 09:49'!
initialize
	super initialize.
	transactionLog := MaTransactionLog new.
	largeCollectionManagers := Dictionary new.
	sessions := Dictionary new.
	commitGuard := Mutex new.
	flushGuard := Mutex new.
	filePool := MaFileStreamPool maxSize: 60.
	wantsToClose := false! !

!MaObjectRepository methodsFor: 'high-availability' stamp: 'cmm 11/23/2008 12:38'!
initializeBranchCode
	filer initializeBranchCode.
	recoveryManager ifNotNil: [ recoveryManager branchCode: filer branchCode ]! !

!MaObjectRepository methodsFor: 'initialize-release' stamp: 'cmm 2/1/2005 22:45'!
initializeClassDefinitionsSpecialOids
	"This collection of special oids can grow.  When it does, clients need to know about it."

	repositoryController definition classDefinitions do:
		[ : each |
		(repositoryController session isPersistent: each)
			ifTrue:
				[ transactionLog
					onChangeOf: (repositoryController session oidFor: each "OrderedCollection of definitions")
					clientsShould: #refreshClassVersions ] ]! !

!MaObjectRepository methodsFor: 'actions' stamp: 'cmm 5/27/2007 15:19'!
initializeNextOid
	"Initialize it to the current-highest oid, because we always incrmement it when getting it."
	nextOid _ filer getNextOid - 1! !

!MaObjectRepository methodsFor: 'initialize-release' stamp: 'cmm 11/18/2008 15:43'!
initializeSpecialOidsList
	"See MagmaSession>>#handleSpecialOidNotifications:"
	| classDefinitions session |
	classDefinitions := repositoryController definition classDefinitions.
	session := self session.
	(session isPersistent: classDefinitions) ifTrue: 
		[ transactionLog
			onChangeOf: (self session oidFor: repositoryController definition classDefinitions)
				clientsShould: #refreshClassDefinitions ;
			onChangeOf: (self session oidFor: self node)
				clientsShould: #refreshNode ].
	self initializeClassDefinitionsSpecialOids! !

!MaObjectRepository methodsFor: 'accessing' stamp: 'cmm 5/30/2007 21:38'!
initializeSystemReadStrategyUsing: aMaClassIdManager 
	"The object and its immediate variables"
	systemReadStrategy _ MaReadStrategy minimumDepth: 1.
	systemReadStrategy makeReadyForUseUsing: aMaClassIdManager! !

!MaObjectRepository methodsFor: 'accessing' stamp: 'cmm 11/1/2004 10:23'!
isOpen

	^ filer notNil and: [ filer isOpen ]! !

!MaObjectRepository methodsFor: 'magma-collections' stamp: 'cmm 8/19/2007 20:41'!
loadMemberIndexOf: aMagmaCollectionManager using: aMaTerm from: oidInteger distinct: wantsDistinct 
	"Load memberIndex from the results expression aMaTerm."
	| executor sourceManager x mcChanges stillDoing nextWriteStop |
	sourceManager _ self collectionManagerFor: oidInteger.
	[ sourceManager isLoading ] whileTrue: [ (Delay forMilliseconds: 500) wait ].
	executor _ MaQueryExecutor 
		collectionManager: sourceManager
		expression: aMaTerm.
	aMagmaCollectionManager loadExecutor: executor.
	x _ 1.
	nextWriteStop _ 150.
	[ stillDoing _ false.
	mcChanges _ (MagmaCollectionChanges collection: aMagmaCollectionManager collection)
		collectionOid: aMagmaCollectionManager collectionOid ;
		yourself.
	executor 
		trunkFrom: x
		do: 
			[ : eachKey : eachOid | 
			| canPossiblyAdd |
			stillDoing _ true.
			canPossiblyAdd _ (wantsDistinct and: 
				[ (mcChanges added includesKey: eachOid) or: [ aMagmaCollectionManager memberIndex includesKey: eachOid ] ]) not.
			(canPossiblyAdd and: 
				[ aMaTerm 
					shouldInclude: eachOid
					using: sourceManager ]) ifTrue: 
				[ | hashesByIndex |
				hashesByIndex _ self 
					hashKeysReferencing: eachOid
					inIndexesOf: sourceManager
					using: aMagmaCollectionManager.
				mcChanges added 
					at: eachOid
					put: hashesByIndex ].
			x _ x + 1 ]
		until: [ aMagmaCollectionManager shouldStopLoading or: [ x = nextWriteStop ] ].
	stillDoing ] whileTrue: 
		[ | commitPackage |
		commitPackage _ MaCommitPackage new
			objects: (MaSerializedGraphBuffer new: 4) ;
			yourself.
		commitPackage addLargeCollectionChanges: mcChanges.
		"Spare a potentially skewed read with requestCritical, not just commitCritical"
		repositoryController requestCritical: 
			[ self session begin.
			repositoryController forceWritePackage: commitPackage ].
		self flushCacheSoon.
		nextWriteStop _ nextWriteStop + 150 ]! !

!MaObjectRepository methodsFor: 'backup / restore' stamp: 'cmm 8/2/2007 00:16'!
logArchiveFrequency
	"Answer the frequency which commit.log files are closed and a new one started.  Commit log files are used to apply to a backup to bring it up to date."
	^ recoveryManager logArchiveFrequency! !

!MaObjectRepository methodsFor: 'backup / restore' stamp: 'cmm 8/2/2007 00:16'!
logArchiveFrequency: aDuration 
	"Override the default frequency which commit.log files are closed and a new one started.  Commit log files are used to apply to a backup to bring it up to date with the latest commits since that backup."
	recoveryManager logArchiveFrequency: aDuration! !

!MaObjectRepository methodsFor: 'building' stamp: 'cmm 3/6/2009 16:21'!
newSessionFor: userIdString sessionId: aUuid numberOfChallengers: anInteger clientConnection: aMaClientConnection 
	"Answer an Array whose first element is the sessionId, second is the classDefinitionsByteArray, third is the RepositoryDefinition byteArray, fourth is a boolean, true if single-user, false if listening on a port."
	| connection |
	aUuid class = UUID ifFalse: 
		[ MagmaSoftwareError signal: 'sessionId is ' , aUuid printString , ' but must be a UUID.' ].
	userIdString isString ifFalse: 
		[ MagmaSoftwareError signal: 'userId must be a String but is ' , userIdString printString ].
	anInteger isInteger ifFalse: 
		[ MagmaSoftwareError signal: 'numberOfChallengers must be an Integer but is ' , anInteger printString ].
	aMaClientConnection ifNotNil: 
		[ aMaClientConnection class = MaClientConnection ifFalse: 
			[ MagmaSoftwareError signal: 'clientConnection must be a MaClientConnection but is ' , aMaClientConnection printString ] ].
	(transactionLog hasEntryFor: aUuid) 
		ifTrue: 
			[ "Be extra safe for now, only a notification."
			MagmaNotification signal: 'Potential software error:  Client ' , aUuid asString , ' attempted to connect to the same server even though it was already connected.' ]
		ifFalse: 
			[ connection := transactionLog 
				newSessionFor: userIdString
				sessionId: aUuid
				numberOfChallengers: anInteger
				clientConnection: aMaClientConnection.
			sessions 
				at: aUuid
				put: connection ].
	^ (Array new: 4)
		at: 1
			put: self classDefinitionsByteArray ;
		at: 2
			put: self definitionByteArray ;
		at: 3
			put: self commitNumber ;
		at: 4
			put: self restoreMode not ;
		yourself! !

!MaObjectRepository methodsFor: 'building' stamp: 'cmm 12/18/2008 19:32'!
newTransactionFor: sessionUuid 
	(transactionLog entryFor: sessionUuid) isCommitEntry ifTrue: [ MagmaAlreadyInTransaction signal: 'Already in a transaction.' ].
	^ transactionLog 
		newTransactionFor: sessionUuid
		repository: self! !

!MaObjectRepository methodsFor: 'accessing' stamp: 'cmm 11/13/2008 10:33'!
node
	^ repositoryController node! !

!MaObjectRepository methodsFor: 'client-requests' stamp: 'cmm 4/13/2006 17:11'!
numberOfEntriesFrom: lowKeyInteger to: highKeyInteger inLargeCollection: collectionOid on: attributeSymbol 
	| index |
	index _ (self collectionManagerFor: collectionOid) 
		oidsIndexForAttribute: attributeSymbol.
	^ index 
		numberOfEntriesFrom: lowKeyInteger
		to: highKeyInteger! !

!MaObjectRepository methodsFor: 'accessing' stamp: 'cmm 4/21/2003 16:29'!
objectFileSize

	^filer objectFileSize! !

!MaObjectRepository methodsFor: 'accessing' stamp: 'cmm 12/22/2008 13:16'!
oidCount
	"The number of oids allocated by this repository.  The total number of objects ever committed, even if some have since been garbage-collected."
	^ nextOid - MaOidCalculator firstUserObjectOid! !

!MaObjectRepository methodsFor: 'actions' stamp: 'cmm 6/12/2007 22:18'!
open: aFileDirectory 
	filer _ MaObjectFiler open: aFileDirectory.
	[ filer validateVersion ] ifCurtailed: [ filer close ].
	self postOpenInitialize! !

!MaObjectRepository methodsFor: 'recovery' stamp: 'cmm 10/26/2006 18:43'!
openDbFilesDo: twoArgBlock 
	"Value twoArgBlock with each database file and its fileId."
	filer filesDo: twoArgBlock.
	largeCollectionManagers do: [ : each | each filesDo: twoArgBlock ]! !

!MaObjectRepository methodsFor: 'building' stamp: 'cmm 10/25/2006 23:48'!
openLargeCollectionIfNecessary: aMagmaLargeCollection
	^ (self ensureManagerCreatedFor: aMagmaLargeCollection) ensureOpen! !

!MaObjectRepository methodsFor: 'notifications' stamp: 'cmm 11/2/2008 09:39'!
pageDba: aString 
	"Signal DBA notification.."
	self halt: 'System needs help from the DBA'! !

!MaObjectRepository methodsFor: 'actions' stamp: 'cmm 10/22/2008 15:40'!
postOpenInitialize
	self initializeNextOid.
	recoveryManager := 
		(MaRecoveryManager repository: self)
		open: self repositoryController localLocation ;
		yourself.
	self postOpenInitializeRecoveryManager! !

!MaObjectRepository methodsFor: 'actions' stamp: 'cmm 8/3/2007 13:00'!
postOpenInitializeRecoveryManager
	recoveryManager
		commitNumber: filer commitNumber ;
		branchCode: filer branchCode! !

!MaObjectRepository methodsFor: 'accessing' stamp: 'cmm 11/6/2008 16:37'!
primitiveAttributes
	"Answer a Dictionary of the primitiveAttributes of my filer (with special logic for exploding the single #booleanFlags attribute to all of its logical boolean attributes we really care about)."
	| answer |
	answer := Dictionary new.
	filer primitiveAttributeAddressesMap keys do: 
		[ : key | 
		key isSymbol ifTrue: 
			[ | desiredKeys |
			desiredKeys := key = #booleanFlags 
				ifTrue: 
					[ {  #restoreMode  } ]
				ifFalse: 
					[ {  key  } ].
			desiredKeys do: 
				[ : eachDesiredKey | 
				answer 
					at: eachDesiredKey
					put: (filer perform: eachDesiredKey) ] ] ].
	^ answer! !

!MaObjectRepository methodsFor: 'recovery' stamp: 'cmm 6/1/2005 23:56'!
primitiveRecover
	"This is needed in very rare, dire circumstances where the server was actually writing its own system-definition when an outage occurred.  In that case the repository will not even be able to open normally to get to the normal recovery.  That condition is trapped for any Error and sent here to try opening all files in the directory and then doing a normal unapply (rollback).  Theoretically, that should work."

	filer ensureFilesOpen.
	recoveryManager primitiveRecoverUsing: self! !

!MaObjectRepository methodsFor: 'actions' stamp: 'cmm 6/1/2007 14:13'!
queueForDelete: aMaAtomicFileStream 
	"When removing indexees, we actually remove the supporting file.  However, we don't want to do that until the commit actually succeeds.  Therefore, we a queue the files to deleted until the commit succeeds, then we delete it."
	recoveryManager queueForDelete: aMaAtomicFileStream! !

!MaObjectRepository methodsFor: 'client-requests' stamp: 'cmm 7/6/2007 01:28'!
read: oidInteger for: sessionId using: aMaReadStrategy 
	| result |
	self commitCritical: 
		[ result _ MaReadResult byteArray: 
			(self 
				byteArrayAt: oidInteger
				using: (self 
						ensureReadStrategy: aMaReadStrategy
						forSession: sessionId)).
		self 
			renewServerNotificationsIn: result
			for: sessionId ].
	^ result! !

!MaObjectRepository methodsFor: 'client-requests' stamp: 'cmm 5/14/2005 18:27'!
readAnchorFor: sessionId using: aMaReadStrategy

	| anchorOid |
	^ MaOidCalculator oidForNil = (anchorOid _ filer anchorOid)
		ifTrue:
			[ (self serverSerializer serializeGraph: MaRootAnchor new) trimmedByteArray ]
		ifFalse:
			[ self
				read: anchorOid
				for: sessionId
				using: aMaReadStrategy ]! !

!MaObjectRepository methodsFor: 'recovery' stamp: 'cmm 10/28/2008 11:27'!
recoveryManager
	^ recoveryManager ! !

!MaObjectRepository methodsFor: 'client-requests' stamp: 'cmm 2/11/2008 17:25'!
refreshAll: oidsCollection for: sessionId 
	"We will use a minimal strategy because we really just want exactly the buffers requested.  So save whatever the sessions current strategy."
	| answer sessionReadStrategy minimalStrategy |
	sessionReadStrategy := (self sessionWithId: sessionId) currentReadStrategy.
	minimalStrategy := MaBasicReadStrategy minimumDepth: 0.
	answer := Array new: oidsCollection size.
	1 
		to: oidsCollection size
		do: 
			[ : n | 
			answer 
				at: n
				put: (MaSerializedGraphBuffer byteArray: (self 
						read: (oidsCollection at: n)
						for: sessionId
						using: minimalStrategy) byteArray) ].
	"Now restore the readStrategy to whatever it was."
	(self sessionWithId: sessionId) currentReadStrategy: sessionReadStrategy.
	^ answer! !

!MaObjectRepository methodsFor: 'client-requests' stamp: 'cmm 11/12/2007 22:28'!
refreshSegmentIn: aMaRefreshSegmentRequest 
	(self collectionManagerFor: aMaRefreshSegmentRequest collectionOid) 
		load: aMaRefreshSegmentRequest segment
		upTo: (aMaRefreshSegmentRequest quantity min: 10000)
		where: aMaRefreshSegmentRequest expression
		withObjects: aMaRefreshSegmentRequest loadObjects
		except: aMaRefreshSegmentRequest exceptions.
	aMaRefreshSegmentRequest segment 
		collectBuffers: aMaRefreshSegmentRequest sessionId
		from: self
		using: aMaRefreshSegmentRequest readStrategy.
	self 
		renewServerNotificationsIn: aMaRefreshSegmentRequest segment
		for: aMaRefreshSegmentRequest sessionId.
	^ aMaRefreshSegmentRequest segment! !

!MaObjectRepository methodsFor: 'magma-collections' stamp: 'cmm 10/26/2006 15:09'!
registerAndLoad: aMaCommitPackage using: aMaTerm from: oidInteger forSession: sessionId distinct: aBoolean 
	| newManager result |
	result _ self 
		registerResultSet: aMaCommitPackage
		forSession: sessionId.
	newManager _ self collectionManagerFor: aMaCommitPackage allLargeCollectionChanges anyOne collectionOid.
	newManager loadProcess: (
		[ self 
			loadMemberIndexOf: newManager
			using: aMaTerm
			from: oidInteger
			distinct: aBoolean ] forkAt: Processor userBackgroundPriority).
	newManager loadProcess name: 'loading ' , newManager collectionOid printString.
	^ result! !

!MaObjectRepository methodsFor: 'magma-collections' stamp: 'cmm 10/30/2006 11:58'!
registerResultSet: aMaCommitPackage forSession: sessionId 
	"Write buffers for aMaCommitPackage and its indexes collection so that, in case it becomes persistently referenced."
	self session begin.	"Begin a tran to allow the submitAll:for:.."
	^ repositoryController forceWritePackage: aMaCommitPackage! !

!MaObjectRepository methodsFor: 'client-requests' stamp: 'cmm 12/18/2008 20:50'!
removeSessionFor: sessionId 
	| clientConnection |
	clientConnection := sessions 
		removeKey: sessionId
		ifAbsent: 
			[ "Perhaps the other session suddenly disconnected..?"
			 ].
	clientConnection ifNotNil: 
		[ clientConnection maClientConnection ifNotNilDo: [ : cc | cc shouldContinue: false ] ].
	transactionLog 
		removeEntryForSessionId: sessionId
		ifAbsent: 
			[ "Perhaps the other session suddenly disconnected..?"
			 ]! !

!MaObjectRepository methodsFor: 'actions' stamp: 'cmm 4/28/2003 23:51'!
renewCurrentReadStrategy: aMaReadStrategy for: sessionId

	| session |
	session _ self sessionWithId: sessionId.
	aMaReadStrategy ifNotNil: [ session currentReadStrategy: aMaReadStrategy ].
	^session currentReadStrategy! !

!MaObjectRepository methodsFor: 'actions' stamp: 'cmm 11/5/2008 15:51'!
renewServerNotificationsIn: aMagmaServerResult for: sessionId 
	self 
		renewServerNotificationsIn: aMagmaServerResult
		from: (transactionLog entryFor: sessionId)! !

!MaObjectRepository methodsFor: 'actions' stamp: 'cmm 11/5/2008 15:50'!
renewServerNotificationsIn: aMagmaServerResult from: aMaTransactionLogEntry 
	aMagmaServerResult serverNotifications: aMaTransactionLogEntry renewServerNotifications.
	aMagmaServerResult serverNotifications keysAndValuesDo: 
		[ : eachKey : eachSet | 
		aMagmaServerResult serverNotifications 
			at: eachKey
			put: (eachSet collect: 
				[ : eachOid | 
				self 
					byteArrayAt: eachOid
					using: MaReadStrategy deep ]) ]! !

!MaObjectRepository methodsFor: 'client-requests' stamp: 'cmm 1/8/2009 19:02'!
renewTransactionEntryFor: sessionId 
	^ transactionLog 
		renewTransactionEntryFor: sessionId
		repository: self! !

!MaObjectRepository methodsFor: 'accessing' stamp: 'cmm 7/8/2005 17:16'!
repositoryController
	^ repositoryController! !

!MaObjectRepository methodsFor: 'building' stamp: ''!
repositoryController: aMaRepositoryController

	repositoryController _ aMaRepositoryController.
	aMaRepositoryController repository == self ifFalse: [ aMaRepositoryController repository: self ]
! !

!MaObjectRepository methodsFor: 'testing' stamp: 'cmm 11/15/2008 11:03'!
restoreMode
	^ restoreMode ifNil: [ restoreMode := filer restoreMode ]! !

!MaObjectRepository methodsFor: 'high-availability' stamp: 'cmm 11/14/2008 13:18'!
restoreMode: aBoolean 
	filer restoreMode: aBoolean.
	restoreMode := aBoolean! !

!MaObjectRepository methodsFor: 'recovery' stamp: 'cmm 3/2/2009 15:44'!
rollbackIfNecessary
	recoveryManager isRollbackNeeded ifTrue: 
		[ MagmaRepositoryDefinition runningTestCases ifFalse: 
			[ MagmaRollbackRecoveryNeeded signal: 'A partially flushed cache has been detected!!  This repository will now be recovered.' ].
		recoveryManager rollbackUsing: self	"NOTE:  This replaces me in my repositoryController." ]! !

!MaObjectRepository methodsFor: 'recovery' stamp: 'cmm 3/6/2009 15:52'!
rollforwardIfPossible
	^ recoveryManager isRollbackNeeded 
		ifTrue: 
			[ MagmaUserError signal: 'A rollback of the last incomplete commit is still needed.  Cannot roll forward until this is done.' ]
		ifFalse: 
			[ recoveryManager
				reapplyCommitPackagesUsing: self ;
				resetApplyFile ]! !

!MaObjectRepository methodsFor: 'high-availability' stamp: 'cmm 11/13/2008 20:57'!
sendToWarmBackups: aMaCommitLogRecord 
	"Update my warmBackups with aMaCommitLogEntry."
	repositoryController node secondaryLocationsDo: 
		[ : each | 
		"Asynchronous request!!  Only a send error will detect a problem."
		[ (self node sessionFor: each) write: aMaCommitLogRecord ] 
			on: NetworkError "this is a send-only request, there will be no response and no server-generated error (i.e., we can't handle MagmaWrongCommitNumber)."
			do: [ : err | MagmaNodeUpdate signalRemove: each ] ]! !

!MaObjectRepository methodsFor: 'accessing' stamp: 'cmm 12/31/2002 12:56'!
serverSerializer

	^repositoryController serverSerializer! !

!MaObjectRepository methodsFor: 'accessing' stamp: 'cmm 12/31/2002 12:55'!
session

	^repositoryController session! !

!MaObjectRepository methodsFor: 'accessing' stamp: 'cmm 2/8/2009 19:22'!
sessionWithId: aUuid 
	^ self 
		sessionWithId: aUuid
		ifAbsent: [ MagmaSessionLost signal ]! !

!MaObjectRepository methodsFor: 'accessing' stamp: 'cmm 2/8/2009 19:22'!
sessionWithId: aUuid ifAbsent: aBlock 
	^ sessions 
		at: aUuid
		ifAbsent: aBlock! !

!MaObjectRepository methodsFor: 'accessing' stamp: 'cmm 5/2/2003 15:05'!
sessions

	^sessions! !

!MaObjectRepository methodsFor: 'client-requests' stamp: 'cmm 6/14/2005 17:30'!
sizeOfLargeCollection: oidInteger

	^ (self collectionManagerFor: oidInteger) size! !

!MaObjectRepository methodsFor: 'client-requests' stamp: 'cmm 1/8/2009 18:52'!
submitAll: aMaCommitPackage for: sessionId beginAnother: aBoolean 
	"Save all of the serialized objects in aMaCommitPackage to the repository.
	If any of the objects in aMaCommitPackage have been changed by other transactions, signal a commit failure."
	| entry |
	self restoreMode ifTrue: 
		[ MagmaUserError signal: 'Commits are not allowed because this repository is in restoreMode.' ].
	entry := transactionLog 
		renewEntryFor: sessionId
		withNew: (aBoolean 
				ifTrue: [ MaCommitLogEntry ]
				ifFalse: [ MaTransactionLogEntry ])
		for: self.
	entry isCommitEntry ifFalse: 
		[ (MagmaNoTransactionError result: entry result) signal: 'No transaction to commit.' ].
	entry
		determineResultFor: aMaCommitPackage
			using: self ;
		initializeChallengers.
	"No need for these challengers anymore, delete them, save memory."
	entry result hasConflicts 
		ifTrue: 
			[ MagmaCommitConflictError new
				result: entry result ;
				signal ]
		ifFalse: 
			[ entry commitPackage: aMaCommitPackage.
			self write: entry ].
	(sessionId ~= repositoryController sessionId and: [ transactionLog isFirstEntryFor: repositoryController sessionId ]) ifTrue: [ repositoryController refreshView ].
	^ entry result! !

!MaObjectRepository methodsFor: 'read strategies' stamp: 'cmm 9/1/2003 19:11'!
systemReadStrategy

	^systemReadStrategy! !

!MaObjectRepository methodsFor: 'accessing' stamp: 'cmm 2/8/2009 19:26'!
systemReadStrategy: aMaReadStrategy
	systemReadStrategy := aMaReadStrategy! !

!MaObjectRepository methodsFor: 'recovery' stamp: 'cmm 4/11/2007 23:15'!
unapply: aMaApplyRecord 
	"Reapply the contents of aMaApplyRecord to the file it came from."
	self 
		fileManagerFor: aMaApplyRecord
		ifPresent: [ : fileManager | fileManager unapplyUsing: aMaApplyRecord ]
		ifAbsent: 
			[ "This could happen if a new file was created (i.e., a new large-collection) but the system crashed before the writes of the definition occurred; in which case the largeCollectionManager would not be able to be opened."
			"should be a MaNewFileRecord, so we don't even need to pass a fileStream of any kind.."
			aMaApplyRecord unapplySelf: nil ]! !

!MaObjectRepository methodsFor: 'client-requests' stamp: 'cmm 4/12/2006 21:14'!
upTo: anInteger oidsFromIndex: startIndex inLargeCollection: collectionOid using: aMaReadStrategy forSession: sessionId 
	| segmentResult |
	segmentResult _ (self collectionManagerFor: collectionOid) 
		upTo: anInteger
		inSegmentFromIndex: (startIndex max: 1).
	segmentResult 
		collectBuffers: sessionId
		from: self
		using: aMaReadStrategy.
	self 
		renewServerNotificationsIn: segmentResult
		for: sessionId.
	^ segmentResult! !

!MaObjectRepository methodsFor: 'magma-collections' stamp: 'cmm 7/12/2005 11:25'!
updateLargeCollectionsIn: aMaCommitPackage recordResultsIn: aMaCommitResult
	"The repositoryControllers sessions view of the repository must be completely refreshed, it must know about all new collections."

	aMaCommitPackage allLargeCollectionChangesDo:
		[ : each |
		each isNewCollection
			ifTrue:
				[ | permanentOid mc |
				permanentOid _ aMaCommitResult permanentOids at: each collectionOid ifAbsent: [ MagmaSoftwareError signal: 'expected to find permanent oid for new collection ', each collectionOid printString ].
				each collectionOid: permanentOid.
				mc _ 
					self session 
						objectWithOid: each collectionOid 
						ifAbsent:
							[ MagmaSoftwareError signal: 'expected to know about MagmaCollection ', each collectionOid, ' here.' ].
				(self ensureManagerCreatedFor: mc) 	createNewLargeCollection ].
		each hasAddedIndexes 
			ifTrue:
				[ (self collectionManagerFor: each collectionOid) createSupplementaryIndexes ] ].
	aMaCommitPackage allLargeCollectionChangesDo:
		[ :each |
		(self collectionManagerFor: each collectionOid)
			process: each
			recordResultsIn: aMaCommitResult ]! !

!MaObjectRepository methodsFor: 'accessing' stamp: 'cmm 4/21/2003 16:02'!
version

	^filer version! !

!MaObjectRepository methodsFor: 'accessing' stamp: 'cmm 7/4/2006 16:20'!
wantsToClose
	^ wantsToClose ! !

!MaObjectRepository methodsFor: 'notifications' stamp: 'cmm 11/2/2008 09:39'!
warnDba: aString 
	"Signal DBA notification.."
	self halt: 'Warn the DBA of potential trouble'! !

!MaObjectRepository methodsFor: 'actions' stamp: 'cmm 10/28/2008 11:17'!
write: aMaCommitLogEntry 
	self commitCritical: 
		[ | trimmedEntry |
		aMaCommitLogEntry
			registerUsing: self ;
			link ;
			record.
		self 
			write: aMaCommitLogEntry
			logged: true
			commitLogRecord: nil ]! !

!MaObjectRepository methodsFor: 'actions' stamp: 'cmm 3/2/2009 21:57'!
write: aMaCommitLogEntry logged: shouldLog commitLogRecord: aMaCommitLogRecord 
	"This is the main 'write' logic for Magma.  It is used for 1) regular commits, 2) recovering from records in the commits.log file and, 3) updating warm backups.  Case 1 will pass a non-trimmed aMaCommitLogEntry and NO aMaCommitLogRecord (nil).  Cases 2 and 3 will pass a trimmed MaCommitLogEntry and the MaCommitLogRecord it was materialized from."
	| trimmedEntry commitLogRecord |
	aMaCommitLogRecord ifNotNil: 
		[ "I could be a warm backup receiving previously-written commit-log entries.  Be absolutely positive it is the correct one!!"
		aMaCommitLogEntry result commitNumber = (self commitNumber + 1) ifFalse: 
			[ MagmaWrongCommitNumber new
				expectedCommitNumber: self commitNumber + 1 ;
				signal ].
		aMaCommitLogRecord branchCode ~= filer branchCode ifTrue: [ MagmaDifferentBranch signal ] ].
	recoveryManager nextCommitNumber.
	self challengeOthersWith: aMaCommitLogEntry.
	aMaCommitLogEntry didCommitSpecialOid ifTrue: [ self initializeClassDefinitionsSpecialOids ].
	"If aMaCommitLogRecord is specified (notNil), then we know aMaCommitLogEntry is already a trimmedCopy, otherwise aMaCommitLogEntry is for a new commit."
	trimmedEntry := aMaCommitLogRecord 
		ifNil: 
			[ aMaCommitLogEntry commitNumber: recoveryManager commitNumber.
			aMaCommitLogEntry trimmedCopy ]
		ifNotNil: [ aMaCommitLogEntry ].
	commitLogRecord := aMaCommitLogRecord ifNil: [ recoveryManager newCommitLogRecordFor: trimmedEntry ].
	"Don't log when we're *reading* from the log file to recover!!  Nor should we send anything off to warm-backups.  Those will be caught up afterward."
	shouldLog ifTrue: 
		[ "To test rollback-recovery, we cannot log the commits because, even though the files are rolled back, they would then be forward-recovered to include the commitDisaster commit!!  We cannot allow this because we compare the file hashes after the rollback."
		SimulateOutage ifFalse: 
			[ self restoreMode ifFalse: 
				[ "The recovery in an on-line system will be from the warm-backups.  Therefore send off to the warmBackups first so they have the best chance of being up to date."
				self sendToWarmBackups: commitLogRecord ].
			recoveryManager 
				log: commitLogRecord
				flush: repositoryController flushAggressively ] ].
	self
		applyToCache: trimmedEntry ;
		initializeNextOid ;
		flushCacheSoon	"needed after recovery, doesn't hurt regular commits"! !

MaObject subclass: #MaQueryExecutor
	instanceVariableNames: 'expression collectionManager trunk trunkPosition'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Magma server-private'!

!MaQueryExecutor class methodsFor: 'create' stamp: 'cmm 6/11/2006 20:06'!
collectionManager: aMagmaCollectionManager expression: aMaExpression 
	^ self new 
		setCollectionManager: aMagmaCollectionManager
		expression: aMaExpression! !

!MaQueryExecutor methodsFor: 'accessing' stamp: 'cmm 7/26/2006 17:53'!
fractionComplete
	^ trunkPosition 
		ifNil: [ 0 ]
		ifNotNil: 
			[ | trunkSize |
			trunkSize _ self trunk trunkSize.
			trunkSize = 0 
				ifTrue: [ 1 ]
				ifFalse: [ trunkPosition / self trunk trunkSize ] ]! !

!MaQueryExecutor methodsFor: 'private' stamp: 'cmm 2/19/2008 22:27'!
load: aMagmaCollectionSegment upTo: anInteger withObjects: aBoolean except: removedSet 
	"Load my segment with up to anInteger oids.  Since we load segments in succession, only capture the answer oids for the page that surrounds targetIndexInteger."
	| trunkEndPosition hitCount highestKey endIndex lowestKey |
	lowestKey := 1 bitShift: 4096.
	highestKey := hitCount := 0.
	trunkEndPosition := self 
		trunkFrom: aMagmaCollectionSegment trunkStart
		do: 
			[ : eachKey : eachOid | 
			(expression isClause or: 
				[ expression 
					shouldInclude: eachOid
					using: collectionManager ]) ifTrue: 
				[ highestKey := eachKey max: highestKey.
				lowestKey := eachKey min: lowestKey.
				(removedSet includes: eachOid) ifFalse: 
					[ hitCount := hitCount + 1.
					aBoolean ifTrue: [ aMagmaCollectionSegment addObject: eachKey -> eachOid ] ] ] ]
		until: [ hitCount = anInteger ].
	endIndex := aMagmaCollectionSegment startIndex + hitCount - 1.
	aMagmaCollectionSegment
		trunk: self trunk ;
		lowestKey: lowestKey ;
		highestKey: highestKey ;
		trunkEnd: trunkEndPosition.
	aMagmaCollectionSegment endIndex: endIndex.
	expression isClause 
		ifTrue: 
			[ aMagmaCollectionSegment lastKnownSize: trunk trunkSize ]
		ifFalse: 
			[ self 
				recordLastKnownSizeOn: aMagmaCollectionSegment
				to: endIndex
				ifBeyond: trunkEndPosition ]! !

!MaQueryExecutor methodsFor: 'private' stamp: 'cmm 7/28/2006 14:57'!
recordLastKnownSizeOn: aMagmaCollectionReaderSegment to: endIndex ifBeyond: trunkEndPosition 
	trunk trunkSize = 0 
		ifTrue: 
			[ "empty, no possible results"
			aMagmaCollectionReaderSegment lastKnownSize: 0 ]
		ifFalse: 
			[ trunkEndPosition ifNotNil: 
				[ aMagmaCollectionReaderSegment lastKnownSize: (trunkEndPosition >= trunk trunkSize ifTrue: [ endIndex ]) ] ]! !

!MaQueryExecutor methodsFor: 'initializing' stamp: 'cmm 6/11/2006 20:06'!
setCollectionManager: aMagmaCollectionManager expression: aMaExpression 
	collectionManager _ aMagmaCollectionManager.
	expression _ aMaExpression! !

!MaQueryExecutor methodsFor: 'private' stamp: 'cmm 6/13/2006 13:27'!
trunk
	^ trunk ifNil: 
		[ trunk _ MaQueryTrunk new.
		expression 
			addMinimumEnumerationsTo: trunk
			using: collectionManager.
		trunk ]! !

!MaQueryExecutor methodsFor: 'private' stamp: 'cmm 7/26/2006 17:36'!
trunkFrom: startPosition do: twoArgBlock until:  conditionBlock
	"Enumerate my query-trunk from a particular starting point."
	| trunkSearchPosition |
	trunkSearchPosition _ 1.
	trunkPosition _ nil.
	self trunk trunkDo: 
		[ : eachTrunkClause | 
		| hdx rangeSize |
		hdx _ collectionManager oidsIndexForAttribute: eachTrunkClause attribute.
		rangeSize _ hdx 
			numberOfEntriesFrom: eachTrunkClause lowKey
			to: eachTrunkClause highKey.
		eachTrunkClause lastKnownSize: rangeSize.
		"is this NOT the trunk-clause which will contain trunk-position startPosition?"
		trunkSearchPosition + rangeSize - 1 < startPosition 
			ifTrue: [ trunkSearchPosition _ trunkSearchPosition + rangeSize ]
			ifFalse: 
				[ "trunkSearchPosition is only set to position of the beginning of this clause of the trunk, trunkPosition keeps the actual trunkPosition."
				trunkPosition ifNil: [ trunkPosition _ startPosition - 1 "start one-back since we pre-increment (because post-increment leaves it advanced one too far)" ].
				conditionBlock value ifFalse:
					[ hdx 
						keysAndValuesFrom: eachTrunkClause lowKey
						startingAt: (startPosition - trunkSearchPosition + 1 max: 1)
						do: 
							[ : eachKey : eachOid | 
							trunkPosition _ trunkPosition + 1.
							twoArgBlock value: eachKey value: eachOid  ]
						until: 
							[ : eachKey : eachOid | 
							conditionBlock value 
							or: [ eachKey > eachTrunkClause highKey ] ].
					trunkSearchPosition _ trunkPosition ] ]. 
].
	^ trunkPosition! !

MaObject subclass: #MaRecoveryManager
	instanceVariableNames: 'applyFile serializer preMadeRecords groupId commitNumber filesToDelete commitLogFile targetRecoveryTime logStartTime branchCode location repository logArchiveFrequency'
	classVariableNames: 'DefaultRecordSize'
	poolDictionaries: ''
	category: 'Magma server-Recovery'!

!MaRecoveryManager commentStamp: 'cmm 6/11/2007 22:40' prior: 0!
I am responsible for recovery behaviors.

Commit-packages received into the server are prepared, serialized and flushed, wrapped in MaRecoveryRecord, to a commits file.  A separate process runs repeatedly, Delaying five seconds and then forking an atomic write of the extents.

The atomic write enumerates all of the MaTransactionalFileWriters and captures before-images of the portions of the files about to be overwritten, flushes them to the before-images file.  Then flushes all of the MaTransactionalFileWriters to their respective files.  Finally, an endApply record is flushed to the before-images file.

Reads can be interleaved with the atomic write but not commits because the atomic write does not represent a change in the state of the repository.

Recovery

If an outage occurs, we have, in the worst case:

	1) some data partially applied
	2) some data written to commitPackages, but not yet applied

Recovery addresses these in order.

Read the apply-file and look for an incomplete group, a set of RecoveryRecords that have no EndApplyRecord.  If this condition is discovered, the server was in the middle of an atomic-apply when failure occurred.  A partial apply occurred and we must write all of the before-images back to the various files.  Do that.

Now we know we do not have corrupt files, however we next need to apply all of the commitPackages that were written to the commit.log file but not yet applied.

We just apply the commitPackages whose commitNumber is greater than the commitNumber of the Filer.
!

!MaRecoveryManager class methodsFor: 'accessing' stamp: 'cmm 6/9/2005 15:55'!
applyFilename
	^ 'applied.images'! !

!MaRecoveryManager class methodsFor: 'accessing' stamp: 'cmm 7/1/2007 13:52'!
commitLogDirectoryName
	^ 'commits'! !

!MaRecoveryManager class methodsFor: 'accessing' stamp: 'cmm 6/6/2007 23:22'!
commitLogExtension
	^ 'log'! !

!MaRecoveryManager class methodsFor: 'accessing' stamp: 'cmm 6/6/2007 23:22'!
commitLogFilename
	^ 'commits'! !

!MaRecoveryManager class methodsFor: 'class initialization' stamp: 'cmm 5/23/2007 23:27'!
defaultRecordSize: anInteger 
	DefaultRecordSize _ anInteger! !

!MaRecoveryManager class methodsFor: 'class initialization' stamp: 'cmm 5/23/2007 23:27'!
initialize
	super initialize.
	self defaultRecordSize: 300! !

!MaRecoveryManager class methodsFor: 'create' stamp: 'cmm 7/8/2007 17:33'!
repository: aMaObjectRepository 
	^ self new
		setRepository: aMaObjectRepository ;
		yourself! !

!MaRecoveryManager methodsFor: 'file' stamp: 'cmm 6/13/2007 20:53'!
activeCommitLogFilename
	^ 'commits.log'! !

!MaRecoveryManager methodsFor: 'debugging' stamp: 'cmm 7/16/2005 17:58'!
applyRecords

	| answer |
	answer _ OrderedCollection new.
	[ self applyRecordsDo:
		[ : each |
		answer add: each independentCopy ] ]

		on: MaUnreadableRecoveryRecord
		do: [ : err | answer add: err messageText ].
	^ answer! !

!MaRecoveryManager methodsFor: 'recover' stamp: 'cmm 7/15/2005 16:25'!
applyRecordsDo: aBlock
	"Enumerate the block of RecoveryRecords for a single groupId in my applyFile."

	| currentPosition selectedRecord |
	currentPosition _ 0.
	[ selectedRecord _ 
		self 
			readRecoveryFile: applyFile
			at: currentPosition.
	selectedRecord isBeginApply
		ifTrue: [ groupId _ selectedRecord groupId ]
		ifFalse: 
			[ selectedRecord groupId = groupId
				ifFalse:
					[ MaUnreadableRecoveryRecord signal: 'expected groupId ', groupId printString, ' but encountered record with ', selectedRecord groupId printString. ] ].
	aBlock value: selectedRecord.
	selectedRecord isEndApply ]
		whileFalse:
			[ currentPosition _ currentPosition + selectedRecord physicalSize ]! !

!MaRecoveryManager methodsFor: 'file' stamp: 'cmm 8/1/2007 22:16'!
archiveCommitLog
	repository commitCritical: 
		[ commitLogFile ifNotNil: 
			[ commitLogFile close.
			commitLogFile _ nil.
			self renameActiveCommitLog ] ]! !

!MaRecoveryManager methodsFor: 'file' stamp: 'cmm 6/13/2007 20:55'!
archiveCommitLogFilename
	"Answer the proper name for the archive commits file to become a recovery file.  The filename is commits.123.log, where 123 is commitNumber of the last CommitRecord in the file."
	^ String streamContents: 
		[ : stream | 
		stream
			maPrint: self class commitLogFilename ;
			maPrint: $. ;
			maPrint: commitNumber ;
			maPrint: $. ;
			maPrint: self class commitLogExtension ]! !

!MaRecoveryManager methodsFor: 'record' stamp: 'cmm 7/12/2005 14:28'!
beginApply

	self preMadeBeginApplyRecord
		groupId: self getNextGroupId ;
		writeTo: applyFile  "no need to flush until the first ApplyRecord is written."! !

!MaRecoveryManager methodsFor: 'accessing' stamp: 'cmm 6/19/2007 23:21'!
branchCode
	^branchCode! !

!MaRecoveryManager methodsFor: 'accessing' stamp: 'cmm 6/19/2007 23:21'!
branchCode: anInteger
	branchCode := anInteger! !

!MaRecoveryManager methodsFor: 'file' stamp: 'cmm 8/3/2007 12:56'!
close
	applyFile close.
	applyFile _ nil.
	self archiveCommitLog! !

!MaRecoveryManager methodsFor: 'accessing' stamp: 'cmm 7/1/2007 11:47'!
commitLogDirectory
	^ location commitLogDirectory! !

!MaRecoveryManager methodsFor: 'accessing' stamp: 'cmm 6/11/2007 22:29'!
commitLogFile
	^ commitLogFile ifNil: [ self openCommitLog ]! !

!MaRecoveryManager methodsFor: 'private' stamp: 'cmm 8/15/2008 12:56'!
commitLogFileNamesFrom: startNumber do: oneArgBlock 
	"Create a SortedCollection of the entries by the commitLogNumber.  Attach the original names so we don't have to reconstruct them after the sort."
	| selectedEntries sortedEntries |
	selectedEntries := OrderedCollection new.
	self commitLogDirectory in: 
		[ : dir | 
		dir entries do: 
			[ : each | 
			| nameParts |
			nameParts := dir splitNameVersionExtensionFor: each name.
			(nameParts at: 2) >= startNumber ifTrue: [ selectedEntries add: {  (each name). nameParts  } ] ] ].
	sortedEntries := selectedEntries asSortedCollection: [ : a : b | a second second < b second second ].
	sortedEntries do: [ : each | oneArgBlock value: (self commitLogDirectory fullNameFor: each first) ].
	self hasActiveCommitLog ifTrue: 
		[ oneArgBlock value: (self directory fullNameFor: self activeCommitLogFilename) ]! !

!MaRecoveryManager methodsFor: 'file' stamp: 'cmm 11/12/2008 19:36'!
commitLogRecordsFrom: startCommitNumber do: aBlock until: untilBlock 
	"Enumerate my MaCommitLogRecords beyond my commitNumber.  Value aBlock with the first-class instance of MaCommitLogRecord."
	repository flushCritical: 
		[ self 
			commitLogFileNamesFrom: startCommitNumber
			do: 
				[ : each | 
				(FileStream readOnlyFileNamed: each) in: 
					[ : stream | 
					
					[ self 
						commitLogRecordsFrom: startCommitNumber
						in: stream
						do: aBlock
						until: untilBlock ] ensure: [ stream close ] ] ] ]! !

!MaRecoveryManager methodsFor: 'private' stamp: 'cmm 8/12/2008 17:21'!
commitLogRecordsFrom: anInteger in: aStandardFileStream do: oneArgBlock until: untilBlock 
	| currentPosition selectedRecord untilReached |
	untilReached := false.
	currentPosition := 0.
	[ aStandardFileStream atEnd or: [ untilReached ] ] whileFalse: 
		[ selectedRecord := self 
			readRecoveryFile: aStandardFileStream
			at: currentPosition.
		(selectedRecord commitNumber >= anInteger and: [ (untilReached := untilBlock value: selectedRecord) not ]) ifTrue: [ oneArgBlock value: selectedRecord ].
		currentPosition := currentPosition + selectedRecord physicalSize ]! !

!MaRecoveryManager methodsFor: 'accessing' stamp: 'cmm 5/24/2007 22:37'!
commitNumber
	^ commitNumber! !

!MaRecoveryManager methodsFor: 'initialize' stamp: 'cmm 5/24/2007 22:36'!
commitNumber: anInteger
	commitNumber _ anInteger! !

!MaRecoveryManager methodsFor: 'file' stamp: 'cmm 5/23/2007 23:27'!
defaultRecordSize
	^ DefaultRecordSize! !

!MaRecoveryManager methodsFor: 'private' stamp: 'cmm 8/6/2007 21:19'!
deleteApplyFile
	applyFile ifNotNil: [ applyFile close ].
	self directory deleteFileNamed: self class applyFilename! !

!MaRecoveryManager methodsFor: 'file' stamp: 'cmm 6/1/2007 14:16'!
deleteFilesQueuedForDelete
	filesToDelete do: 
		[ : eachAtomicFileStream | 
		| fd |
		eachAtomicFileStream deletePhysicalFiles ].
	self initializeFilesToDelete! !

!MaRecoveryManager methodsFor: 'file' stamp: 'cmm 8/2/2007 22:14'!
deleteUnnecessaryCommitLogFiles
	"Value oneArgBlock for each commit-log file in my commitsDirectory that is no longer needed for recovery of my repository."
	self unnecessaryCommitLogFileNamesDo: 
		[ : each | 
		self commitLogDirectory deleteFileNamed: each ]! !

!MaRecoveryManager methodsFor: 'accessing' stamp: 'cmm 7/1/2007 14:00'!
directory
	^location directory ! !

!MaRecoveryManager methodsFor: 'record' stamp: 'cmm 5/15/2007 00:12'!
endApply
	self preMadeEndApplyRecord
		groupId: groupId ;
		writeTo: applyFile.
	applyFile flush.
	self deleteFilesQueuedForDelete.
	"Ok, this group is a wrap.  We can now start back at the beginning."
	applyFile position: 0.
	serializer resetOidManager! !

!MaRecoveryManager methodsFor: 'file' stamp: 'cmm 5/31/2005 17:36'!
ensureByteArray: aByteArray canHold: anInteger

	| newAnswer |
	^aByteArray size < anInteger
		ifTrue:
			[ newAnswer _ ByteArray new: anInteger.
			newAnswer
				replaceFrom: 1
				to: aByteArray size
				with: aByteArray
				startingAt: 1.
			newAnswer ]
		ifFalse: [ aByteArray ]! !

!MaRecoveryManager methodsFor: 'file' stamp: 'cmm 6/7/2005 22:58'!
flush
	applyFile flush! !

!MaRecoveryManager methodsFor: 'record' stamp: 'cmm 6/5/2005 21:40'!
getNextGroupId 

	^ groupId _ groupId + 1! !

!MaRecoveryManager methodsFor: 'recover' stamp: 'cmm 7/1/2007 11:44'!
hasActiveCommitLog
	^ self directory fileExists: self activeCommitLogFilename! !

!MaRecoveryManager methodsFor: 'initialize' stamp: 'cmm 7/8/2007 17:41'!
initialize
	super initialize.
	groupId _ 0.
	self 
		initializePreMadeRecords ;
		initializeFilesToDelete ;
		initializeSerializer! !

!MaRecoveryManager methodsFor: 'initialize' stamp: 'cmm 6/6/2005 12:37'!
initializeFilesToDelete

	filesToDelete _ OrderedCollection new! !

!MaRecoveryManager methodsFor: 'initialize' stamp: 'cmm 6/11/2007 19:54'!
initializePreMadeRecords
	preMadeRecords _ { 
		(MaBeginApplyRecord new).
		(MaBeforeImageRecord new).
		(MaFileGrowthRecord new).
		(MaNewFileRecord new).
		(MaEndApplyRecord new).
		(MaCommitLogRecord new)
	 }! !

!MaRecoveryManager methodsFor: 'initialize' stamp: 'cmm 8/15/2008 12:20'!
initializeSerializer
	serializer := MagmaSession newCommitLogSerializer! !

!MaRecoveryManager methodsFor: 'recover' stamp: 'cmm 7/15/2005 16:25'!
isRollbackNeeded

	| hasRecords hasValidEndApplyIndicator |
	hasRecords _ false.
	hasValidEndApplyIndicator _ false.
	[ self applyRecordsDo:
		[ : each |
		each isBeginApply ifTrue: [ hasRecords _ true ].
		each isEndApply
			ifTrue:
				[ hasValidEndApplyIndicator _ true ] ] ]
		on: MaUnreadableRecoveryRecord
		do: 
			[ : error |
			"An incomplete set of records for a single transactionId was found."
			MagmaNotification signal: error messageText.
			error return ].
	^ hasRecords and: [ hasValidEndApplyIndicator not ]! !

!MaRecoveryManager methodsFor: 'commit-packages' stamp: 'cmm 10/24/2008 14:28'!
log: aMaCommitLogRecord flush: aBoolean 
	aMaCommitLogRecord writeTo: self commitLogFile.
	aBoolean ifTrue: [ self commitLogFile flush ].
	self timeToStartNewCommitLog ifTrue: [ self archiveCommitLog ]! !

!MaRecoveryManager methodsFor: 'accessing' stamp: 'cmm 12/19/2008 14:11'!
logArchiveFrequency
	^ logArchiveFrequency ifNil: [ logArchiveFrequency := 5 minutes asMilliSeconds ]! !

!MaRecoveryManager methodsFor: 'accessing' stamp: 'cmm 12/19/2008 14:42'!
logArchiveFrequency: aDuration 
	"Set the frequency Magma should create new commit.n.log files.  "
	logArchiveFrequency := aDuration asMilliSeconds! !

!MaRecoveryManager methodsFor: 'recover' stamp: 'cmm 10/28/2008 11:26'!
materializeCommitLogEntryIn: aMaCommitLogRecord 
	^ serializer resetOidManager materializeGraph: aMaCommitLogRecord record! !

!MaRecoveryManager methodsFor: 'commit-packages' stamp: 'cmm 10/24/2008 17:03'!
newCommitLogRecordFor: aMaCommitLogEntry 
	| commitPackageRecord |
	commitPackageRecord := self preMadeCommitPackageRecord.
	commitPackageRecord
		record: (serializer resetOidManager serializeGraph: aMaCommitLogEntry) trimmedByteArray ;
		commitNumber: commitNumber ;
		timestamp: DateAndTime now ;
		branchCode: branchCode.
	^ commitPackageRecord! !

!MaRecoveryManager methodsFor: 'commit-packages' stamp: 'cmm 5/23/2007 22:40'!
nextCommitNumber
	^ commitNumber _ commitNumber + 1! !

!MaRecoveryManager methodsFor: 'file' stamp: 'cmm 7/1/2007 13:48'!
open: aMagmaLocalLocation 
	location _ aMagmaLocalLocation.
	self openApplyFile.
	self commitLogDirectory assureExistence! !

!MaRecoveryManager methodsFor: 'file' stamp: 'cmm 8/3/2007 13:07'!
openApplyFile
	applyFile ifNil: [ applyFile _ self directory fileNamed: self class applyFilename ].
	applyFile
		binary ;
		position: 0! !

!MaRecoveryManager methodsFor: 'file' stamp: 'cmm 12/19/2008 14:12'!
openCommitLog
	logStartTime := Time millisecondClockValue.
	^ commitLogFile := (self directory fileNamed: self activeCommitLogFilename)
		binary ;
		setToEnd ;
		yourself	"just to be safe"! !

!MaRecoveryManager methodsFor: 'preMadeRecords' stamp: 'cmm 6/5/2005 16:40'!
preMadeBeforeImageRecord
	^ preMadeRecords at: MaBeforeImageRecord assignedRecordType! !

!MaRecoveryManager methodsFor: 'preMadeRecords' stamp: 'cmm 6/5/2005 16:41'!
preMadeBeginApplyRecord 
	^ preMadeRecords at: MaBeginApplyRecord assignedRecordType! !

!MaRecoveryManager methodsFor: 'preMadeRecords' stamp: 'cmm 6/11/2007 19:55'!
preMadeCommitPackageRecord 
	^ preMadeRecords at: MaCommitLogRecord assignedRecordType! !

!MaRecoveryManager methodsFor: 'preMadeRecords' stamp: 'cmm 6/5/2005 16:41'!
preMadeEndApplyRecord 
	^ preMadeRecords at: MaEndApplyRecord assignedRecordType! !

!MaRecoveryManager methodsFor: 'preMadeRecords' stamp: 'cmm 6/5/2005 16:41'!
preMadeFileGrowthRecord
	^ preMadeRecords at: MaFileGrowthRecord assignedRecordType! !

!MaRecoveryManager methodsFor: 'preMadeRecords' stamp: 'cmm 6/5/2005 16:41'!
preMadeNewFileRecord
	^ preMadeRecords at: MaNewFileRecord assignedRecordType! !

!MaRecoveryManager methodsFor: 'recover' stamp: 'cmm 6/11/2007 22:19'!
primitiveRecoverUsing: aMaObjectRepository 
	"This recovery only happens if there was an outage exactly at the moment where the system was writing its own definition.  In that case, it would not even be able to open properly to perform a normal recovery; therefore we do this initial recovyer which should then *allow* it to materialize its definition so that it can then perform a normal recovery which will get the large-collections.  Whew!!"
	MagmaRecoveryNotification signal: 'System recovery needed!!'.
	[ self applyRecordsDo: 
		[ : eachRecoveryRecord | 
		(eachRecoveryRecord isApplyToFile and: 
			[ eachRecoveryRecord isForObjectsFile or: [ eachRecoveryRecord isForFilePositionsFile ] ]) ifTrue: [ aMaObjectRepository unapply: eachRecoveryRecord ] ] ] 
		on: MaUnreadableRecoveryRecord
		do: 
			[ : error | 
			MagmaRecoveryNotification signal: error messageText , '.  System recovery complete.' ].
	^ true! !

!MaRecoveryManager methodsFor: 'file' stamp: 'cmm 6/1/2007 14:13'!
queueForDelete: aMaAtomicFileStream 
	filesToDelete add: aMaAtomicFileStream! !

!MaRecoveryManager methodsFor: 'file' stamp: 'cmm 7/16/2008 20:04'!
readRecoveryFile: aStandardFileStream at: filePosition 
	| messagePrefix bytesRead inputRecord |
	messagePrefix := 'Ungraceful end of commit-group:  '.
	inputRecord := MaRecoveryRecord new: self defaultRecordSize.
	bytesRead := aStandardFileStream 
		maRead: inputRecord bufferSize
		bytesFromPosition: 1
		of: inputRecord byteArray
		atFilePosition: filePosition.
	(bytesRead 
		between: 1
		and: MaRecoveryRecord sizeOfPhysicalSizeField) ifTrue: 
		[ MaUnreadableRecoveryRecord signal: messagePrefix , 'no forward-recovery records found' ].
	inputRecord physicalSize = 0 ifTrue: [ MaUnreadableRecoveryRecord signal: 'End of recovery group' ].
	inputRecord physicalSize > self unreasonablePhysicalSize ifTrue: 
		[ MaUnreadableRecoveryRecord signal: messagePrefix , 'unreasonable physicalSize found' ].
	"check physicalSize to see if we've read enough."
	inputRecord physicalSize > bytesRead ifTrue: 
		[ "We haven't read enough, read the rest, ensuring we get as much as we should."
		inputRecord byteArray: (self 
				ensureByteArray: inputRecord byteArray
				canHold: inputRecord physicalSize).
		bytesRead := aStandardFileStream 
			maRead: inputRecord physicalSize - bytesRead
			bytesFromPosition: bytesRead + 1
			of: inputRecord byteArray
			atFilePosition: filePosition + bytesRead.
		bytesRead = (inputRecord physicalSize - self defaultRecordSize) ifFalse: [ MaUnreadableRecoveryRecord signal: messagePrefix , 'end of file' ] ].
	self validateRecord: inputRecord.
	^ self selectPreMadeRecord: inputRecord! !

!MaRecoveryManager methodsFor: 'recover' stamp: 'cmm 3/2/2009 22:41'!
reapplyCommitPackagesUsing: aMaObjectRepository 
	| count continue |
	count := 0.
	continue := true.
	
	[ self 
		commitLogRecordsFrom: commitNumber + 1
		do: 
			[ : each | 
			continue == false ifFalse: 
				[ "Apply ONLY if we have the exact next commit."
				| eachCommitLogEntry |
				(branchCode = each branchCode and: [ each commitNumber = (commitNumber + 1) ]) 
					ifTrue: 
						[ count = 0 ifTrue: 
							[ continue := MagmaForwardRecoveryNeeded signal: 'A forward recovery from ' , each timestamp printString , ' is necessary.  If you wish to recover to a point-in-time, press Debug and save the desired target DateAndTime in my ''targetRecoveryTime''.'.
							self validateTargetRecoveryTime ].
						count := count + 1.
						eachCommitLogEntry := self materializeCommitLogEntryIn: each.
						aMaObjectRepository commitCritical: 
							[ aMaObjectRepository 
								write: eachCommitLogEntry
								logged: false
								commitLogRecord: each.
							repository session advanceWith: eachCommitLogEntry ] ]
					ifFalse: 
						[ each branchCode = branchCode 
							ifTrue: 
								[ each commitNumber > (commitNumber + 1) ifTrue: 
									[ MagmaMissingCommitLogWarning signal: 'Looking for commit record ' , (commitNumber + 1) asString , ' but the oldest commit record found was ' , each commitNumber printString.
									^ self ] ]
							ifFalse: [ MagmaDifferentBranch signal ] ] ] ]
		until: 
			[ : each | 
			continue and: [ targetRecoveryTime notNil and: [ each timestamp > targetRecoveryTime ] ] ] ] 
		on: MaUnreadableRecoveryRecord
		do: 
			[ : error | 
			MagmaWarning signal: count printString , ' reapply''s completed before encountering an error in the commit-log:  ' , error messageText , '.  ' ]! !

!MaRecoveryManager methodsFor: 'file' stamp: 'cmm 8/2/2007 00:33'!
renameActiveCommitLog
	"To support one or more warm backups, a staging directory should be used in conjunction with Banyan (or any external copy script) to copy the files to the remote directories.  To make this easier, the current file is written in the repository directory and, when later closed, moved to the separate staging directory (commits) when it is safe to be copied to the warm backup's commitLogs.
	If the image crashes, the active commit.log file is still in the repository directory."
	commitLogFile ifNotNil: 
		[ MagmaSoftwareError signal: 'expected the active-commits file to be closed.' ].
	self hasActiveCommitLog ifTrue: 
		[ self directory 
			rename: self activeCommitLogFilename
			toBe: (self commitLogDirectory fullNameFor: self archiveCommitLogFilename) ]! !

!MaRecoveryManager methodsFor: 'file' stamp: 'cmm 8/6/2007 21:29'!
resetApplyFile
	"Done with a recovery, now we should truncate the applyFile so future sessions have no chance to encounter the data think recovery operations are needed."
	applyFile
		truncate ;
		close ;
		open! !

!MaRecoveryManager methodsFor: 'recover' stamp: 'cmm 8/6/2007 21:08'!
rollbackUsing: aMaObjectRepository 
	"If a beginApply record with a valid checkSum is found, then we must find a corresponding endApply record also with a valid checkSum.  If we don't, then we must reverse the entire transaction by starting back at the beginning and applying all of the beforeImage records with valid checkSums back to their files."
	MagmaServerPreferences debugRecovery ifTrue: [ self halt: 'about to rollback' ].
	[ self applyRecordsDo: 
		[ : each | 
		each isApplyToFile ifTrue: [ aMaObjectRepository unapply: each ] ] ] 
		on: MaUnreadableRecoveryRecord
		do: 
			[ : error | 
			MagmaRecoveryNotification signal: error messageText , '.  Rollback complete.' ].
	aMaObjectRepository repositoryController close.	"because class-definitions or largeCollections may have even needed rolled-back; refreshView will not refresh them."
	self deleteApplyFile.	"so we don't try to recover again on the next open.."
	aMaObjectRepository repositoryController open! !

!MaRecoveryManager methodsFor: 'file' stamp: 'cmm 7/10/2007 20:46'!
selectPreMadeRecord: aMaRecoveryRecord 
	^ (preMadeRecords 
		at: aMaRecoveryRecord recordType
		ifAbsent: 
			[ MaUnreadableRecoveryRecord signal: 'Ungraceful end of commit-group:  invalid recordType found (' , aMaRecoveryRecord recordType printString , ').' ])
		byteArray: aMaRecoveryRecord byteArray ;
		yourself! !

!MaRecoveryManager methodsFor: 'accessing' stamp: 'cmm 8/15/2008 14:43'!
serializer
	^ serializer! !

!MaRecoveryManager methodsFor: 'initialize' stamp: 'cmm 7/8/2007 17:34'!
setRepository: aMaObjectRepository
	repository _ aMaObjectRepository! !

!MaRecoveryManager methodsFor: 'commit-packages' stamp: 'cmm 12/19/2008 14:12'!
timeToStartNewCommitLog
	^ (Time millisecondsSince: logStartTime) > self logArchiveFrequency! !

!MaRecoveryManager methodsFor: 'file' stamp: 'cmm 8/2/2007 21:46'!
unnecessaryCommitLogFileNamesDo: oneArgBlock 
	"Value oneArgBlock for each commit-log file in my commitsDirectory that is no longer needed for recovery of my repository."
	self commitLogDirectory in: 
		[ : dir | 
		dir entries do: 
			[ : each | 
			| nameParts |
			nameParts _ dir splitNameVersionExtensionFor: each name.
			(nameParts at: 2) <= commitNumber ifTrue: 
				[ oneArgBlock value: (self commitLogDirectory fullNameFor: each localName) ] ] ]! !

!MaRecoveryManager methodsFor: 'error handling' stamp: 'cmm 8/12/2007 19:51'!
unreasonablePhysicalSize
	"No commit record should ever be more than 10-meg.  Due to the nature of disaster-recovery, it is possible we are reading a partially-written file.  It could potentially have even shot-craps in the middle of writing the physicalSize portion of the record.  Therefore, we cannot trust even that not to have bytes in it that cause the computer to attempt to allocate gigs of resources.."
	^ MaObjectBuffer unreasonablePhysicalSize! !

!MaRecoveryManager methodsFor: 'private' stamp: 'cmm 7/16/2008 20:04'!
validateRecord: inputRecord 
	inputRecord hasValidCheckSum ifFalse: 
		[ MaUnreadableRecoveryRecord signal: 'record could not validate its checkSum.' ]! !

!MaRecoveryManager methodsFor: 'private' stamp: 'cmm 6/11/2007 21:22'!
validateTargetRecoveryTime
	targetRecoveryTime ifNotNil: 
		[ targetRecoveryTime class name = DateAndTime ifFalse: 
			[ MagmaUserError signal: 'targetRecoveryTime should be a kind of DateAndTime.' ] ]! !

!MaRecoveryManager methodsFor: 'record' stamp: 'cmm 8/13/2007 22:23'!
writeBeforeImageRecordsFrom: aMaAtomicFileStream fileId: fileId 
	| hasWrittenFileGrowth |
	hasWrittenFileGrowth _ false.
	aMaAtomicFileStream writersDo: 
		[ : each | 
		each willCauseGrowth 
			ifTrue: 
				[ "Only need to write one of these."
				hasWrittenFileGrowth ifFalse: 
					[ self preMadeFileGrowthRecord
						originalSize: aMaAtomicFileStream committedSize ;
						fileId: fileId ;
						groupId: groupId ;
						writeTo: applyFile.
					hasWrittenFileGrowth _ true ] ]
			ifFalse: 
				[ self preMadeBeforeImageRecord
					location: each filePos ;
					record: each currentRecord ;
					fileId: fileId ;
					groupId: groupId ;
					writeTo: applyFile ] ]! !

!MaRecoveryManager methodsFor: 'record' stamp: 'cmm 5/24/2006 22:21'!
writeNewFileRecordFrom: aMaTransactionalFileStream fileId: fileId

	self preMadeNewFileRecord
		filename: aMaTransactionalFileStream filename ;
		fileId: fileId ;
		groupId: groupId ;
		writeTo: applyFile! !

!MaRecoveryManager methodsFor: 'record' stamp: 'cmm 8/13/2007 22:23'!
writeRollbackRecordsFor: aMaAtomicFileStream fileId: fileId 
	aMaAtomicFileStream isNew 
		ifTrue: 
			[ self 
				writeNewFileRecordFrom: aMaAtomicFileStream
				fileId: fileId ]
		ifFalse: 
			[ self 
				writeBeforeImageRecordsFrom: aMaAtomicFileStream
				fileId: fileId ]! !

MaObject subclass: #MaTransactionLog
	instanceVariableNames: 'entries specialOids'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Magma server-private'!

!MaTransactionLog commentStamp: '<historical>' prior: 0!
This class resides on the server.
!

!MaTransactionLog methodsFor: 'actions' stamp: 'cmm 1/8/2009 18:52'!
abortTransactionEntryFor: sessionIdInteger repository: aMaObjectRepository 
	^ (self 
		renewEntryFor: sessionIdInteger
		withNew: MaTransactionLogEntry
		for: aMaObjectRepository) result! !

!MaTransactionLog methodsFor: 'private' stamp: 'cmm 11/7/2008 14:20'!
challengeEntriesWith: aMaCommitLogEntry 
	"We record all objects that were committed to all the entries in the receiver.  Some of the entries are in transaction, some are not.  The ones that are will use aMaCommitLogEntry to verify their commit has no conflicts with this entry.  The ones that are not will use the result of the receiver to know which objects need refreshed."
	| entriesToRemove |
	entriesToRemove := nil.
	self entriesDo: 
		[ : each | 
		(each isForSameSessionAs: aMaCommitLogEntry) ifFalse: 
			[ (each hasExcessiveChallengers and: [ each isSystem not ]) 
				ifTrue: 
					[ entriesToRemove ifNil: [ entriesToRemove := OrderedCollection new ].
					entriesToRemove add: each ]
				ifFalse: 
					[ each
						addChallengingEntry: aMaCommitLogEntry ;
						recordChangesToSpecialObjectsIn: aMaCommitLogEntry
							using: self ] ] ].
	"Now check for memory utilisation, see if we're running low on memory.  If we are, zap the oldest session since it will have the most challengingEntries."
	SmalltalkImage current maIsLowOnMemory ifTrue: 
		[ entries 
			maDetect: [ : each | each isSystem not ]
			ifFound: [ : foundEntry | entriesToRemove add: foundEntry ]
			ifNone: [ MagmaLowOnMemory signal: 'Low on memory!!' ] ].
	^ entriesToRemove! !

!MaTransactionLog methodsFor: 'actions' stamp: 'cmm 7/11/2005 15:07'!
entriesDo: aBlock

	entries do: aBlock! !

!MaTransactionLog methodsFor: 'accessing' stamp: 'cmm 1/8/2009 18:43'!
entryFor: sessionId 
	^ self 
		entryFor: sessionId
		ifAbsent: [ MagmaSessionLost signal ]! !

!MaTransactionLog methodsFor: 'accessing' stamp: 'cmm 1/8/2009 18:42'!
entryFor: sessionId ifAbsent: aBlock 
	^ entries 
		detect: [ : each | each sessionId = sessionId ]
		ifNone: aBlock! !

!MaTransactionLog methodsFor: 'testing' stamp: 'cmm 1/8/2009 18:44'!
hasEntryFor: sessionUuid 
	self 
		entryFor: sessionUuid
		ifAbsent: [ ^ false ].
	^ true! !

!MaTransactionLog methodsFor: 'private' stamp: 'cmm 3/9/2009 17:33'!
initialize
	super initialize.
	"An OrderedCollection of MaCommitLogEntrys and MaTransactionLogEntrys.."
	entries := OrderedCollection new.
	specialOids := Dictionary new! !

!MaTransactionLog methodsFor: 'testing' stamp: 'cmm 12/31/2002 12:31'!
isFirstEntryFor: sessionId

	^entries notEmpty and: [ entries first connection sessionId = sessionId ]
! !

!MaTransactionLog methodsFor: 'building' stamp: 'cmm 3/6/2009 16:27'!
newSessionFor: userIdString sessionId: aUuid numberOfChallengers: anInteger clientConnection: aMaClientConnection 
	| connection |
	connection := MagmaClientConnection 
		userId: userIdString
		sessionId: aUuid
		maximumNumberOfChallengers: anInteger
		clientConnection: aMaClientConnection
		isSystem: entries isEmpty.
	entries add: (MaTransactionLogEntry 
			id: aUuid
			connection: connection).
	aMaClientConnection ifNotNil: [ aMaClientConnection info: userIdString , ' - ' , aUuid asString ].
	^ connection! !

!MaTransactionLog methodsFor: 'building' stamp: 'cmm 1/8/2009 18:53'!
newTransactionFor: sessionUuid repository: aMaObjectRepository 
	^ (self 
		renewEntryFor: sessionUuid
		withNew: MaCommitLogEntry
		for: aMaObjectRepository) result! !

!MaTransactionLog methodsFor: 'special oids' stamp: 'cmm 5/19/2004 22:24'!
onChangeOf: anInteger clientsShould: aSymbol

	specialOids
		at: anInteger
		put: aSymbol! !

!MaTransactionLog methodsFor: 'building' stamp: 'cmm 6/12/2006 23:50'!
removeEntryForSessionId: sessionId ifAbsent: aBlock 
	^ entries 
		maDetect: [ : each | each sessionId = sessionId ]
		ifFound: [ : foundTransactionLogEntry | entries remove: foundTransactionLogEntry ]
		ifNone: aBlock! !

!MaTransactionLog methodsFor: 'actions' stamp: 'cmm 1/8/2009 18:51'!
renewEntryFor: sessionUuid withNew: entryClass for: aMaObjectRepository 
	"Gives the session identified by sessionUuid a new TranslactionLogEntry, where its challengers will be accumulated.  The entry prior to replacement is answered."
	| entry |
	entry := self 
		renewEntryFor: sessionUuid
		withNew: entryClass
		ifAbsent: [ MagmaSessionLost signal ].
	aMaObjectRepository 
		renewServerNotificationsIn: entry result
		from: entry.
	entry result
		commitNumber: aMaObjectRepository commitNumber ;
		oidCount: aMaObjectRepository oidCount.
	^ entry! !

!MaTransactionLog methodsFor: 'private' stamp: 'cmm 11/5/2008 15:55'!
renewEntryFor: sessionUuid withNew: aClass ifAbsent: aBlock
	"aClass should be either MaTransactionLogEntry or MaCommitLogEntry"

	| entry |
	entry _
		(self
			removeEntryForSessionId: sessionUuid
			ifAbsent: [ ^ aBlock value ]).
	entry recordAllChangedObjects.
	entries add: 
		(aClass
			id: sessionUuid
			connection: entry result connection).
	^ entry! !

!MaTransactionLog methodsFor: 'actions' stamp: 'cmm 1/8/2009 18:57'!
renewTransactionEntryFor: sessionUuid repository: aMaObjectRepository 
	^ (self 
		renewEntryFor: sessionUuid
		withNew: (self entryFor: sessionUuid) class
		for: aMaObjectRepository) result! !

!MaTransactionLog methodsFor: 'special oids' stamp: 'cmm 5/19/2004 00:10'!
specialOidsDo: twoArgBlock

	specialOids keysAndValuesDo: twoArgBlock! !

MaObject subclass: #MagmaCompressor
	instanceVariableNames: 'inputController referencedCollections referencedClasses sourceLocation outputFiler'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Magma server-Utilities'!

!MagmaCompressor commentStamp: 'cmm 8/29/2002 22:35' prior: 0!
I garbage collect a Magma repository and output a smaller file by eliminating buffers no longer referenced from the root buffer.

I always create a new object file, but if you use compressAndRename, I'll rename the source files to 'old...' so you can get with using the compressed repository right away.

If you don't want me to rename, just use #compressTo:.

Oh, and I'm not meant to work while the files are on-line.  For that, use:

	yourRepositoryController garbageCollect

(Note:  a potential strategy for on-line GC would be to set a boolean flag indicating to the filer not to reuse buffers made available through grown objects, but rather to always add to the end.  During this time, a background thread can start reading from the root and build a new copy of the 'memory' map (the MaIntervalCollection).  Once it has accounted for every byte in the real memory map, we know the real one can be replaced with this new one.  Do a quick write of the new memory map and you're done!!)!

!MagmaCompressor class methodsFor: 'creation' stamp: 'cmm 9/24/2004 16:49'!
inputController: aMagmaRepositoryController
	"NOT YET SUPPORTED.  The premise of this is that you can compress a running repository, but it is not complete.  Do not use!!"
	^self new inputController: aMagmaRepositoryController! !

!MagmaCompressor class methodsFor: 'creation' stamp: 'cmm 1/18/2007 23:58'!
source: pathStringOrMagmaLocalLocation
	^ self new sourceLocation: pathStringOrMagmaLocalLocation asMagmaLocalLocation! !

!MagmaCompressor methodsFor: 'compress' stamp: 'cmm 5/27/2007 16:15'!
compressTo: outputLocation 
	| shouldCloseInput location |
	location _ outputLocation asMagmaLocalLocation.
	shouldCloseInput _ inputController isNil.
	inputController ifNil: [ self openInput ].
	self
		createOutput: location ;
		copyPrimitiveAttributes ;
		writeGraphs ;
		ensureNoOidReuse ;
		copyMagmaCollections.
	outputFiler close.
	shouldCloseInput ifTrue: [ inputController close ]! !

!MagmaCompressor methodsFor: 'private' stamp: 'cmm 4/29/2007 21:48'!
copyMagmaCollections
	referencedCollections do: 
		[ : eachOid | 
		| sourceManager |
		sourceManager _ inputController repository collectionManagerFor: eachOid.
		sourceManager filenames do: 
			[ : eachMcFilename | 
			(MaAtomicFileStream fileNamed: (inputController directory fullNameFor: eachMcFilename)) physicalFilenamesDo: 
				[ : eachPhysicalName | 
				inputController directory 
					maCopyFileNamed: eachPhysicalName
					toDirectory: outputFiler directory ] ] ]! !

!MagmaCompressor methodsFor: 'private' stamp: 'cmm 3/4/2009 12:45'!
copyPrimitiveAttributes
	| inputFiler |
	inputFiler := inputController repository filer.
	inputController requestCritical: 
		[ outputFiler primitiveAttributeAddressesMap keysAndValuesDo: 
			[ : eachKey : eachValue | 
			eachKey isSymbol ifTrue: 
				[ outputFiler 
					perform: (eachKey , ':') asSymbol
					with: (inputFiler perform: eachKey) ] ] ]! !

!MagmaCompressor methodsFor: 'private' stamp: 'cmm 3/29/2007 16:44'!
createOutput: aMagmaLocalLocation
	outputFiler _ MaObjectFiler create: aMagmaLocalLocation directory ! !

!MagmaCompressor methodsFor: 'private' stamp: 'cmm 5/27/2007 16:36'!
ensureNoOidReuse
	"We can never allow an oid to be reused.  Once an object is created and persisted, its oid must be forever consumed.  Even if that object is later dereferenced, its oid will never be reused.  This is necessary to properly support universal-identity which is essential for MagmaForwardingProxys."
	"determine the next oid to be used in our inputController."
	| nextOid |
	nextOid _ inputController repository filer getNextOid.
	"Force our output controller to allocate up to the current oid."
	outputFiler getNextOid < nextOid ifTrue: 
		[ outputFiler 
			index: nextOid - 1
			at: 0 ]! !

!MagmaCompressor methodsFor: 'initialize-release' stamp: 'cmm 4/2/2007 17:17'!
initialize
	super initialize.
	referencedClasses _ Set new.
	referencedCollections _ Set new! !

!MagmaCompressor methodsFor: 'initialize-release' stamp: 'cmm 8/17/2002 15:25'!
inputController: aMagmaRepositoryController

	inputController _ aMagmaRepositoryController! !

!MagmaCompressor methodsFor: 'private' stamp: 'cmm 3/29/2007 16:30'!
markClassReferenced: aMaObjectBuffer 
	| classIdManager |
	classIdManager := inputController session serializer classIdManager.
	(classIdManager 
		definitionForClassId: aMaObjectBuffer classId
		version: aMaObjectBuffer classVersion) 
		withAllSuperclassDefinitionsDo: 
			[ : eachClassDefinition | 
			referencedClasses add: eachClassDefinition id ].
	"If a class-reference, be sure to get the referenced class itself."
	(aMaObjectBuffer classId = (classIdManager idForClass: MaClassReference))
		ifTrue:
			[ | indexOfId classReferenceDefinition |
			classReferenceDefinition _
				classIdManager
					definitionForClassId: aMaObjectBuffer classId
					version: aMaObjectBuffer classVersion.
			"Get the index of the 'id' variable."
			indexOfId _ classReferenceDefinition allInstVarNames indexOf: 'id'.
			indexOfId = 0 ifTrue: [ MagmaSoftwareError signal: 'Expected ''id'' variable.' ].
			(classIdManager 
				definitionForClassId:
					(inputController session
						objectWithOid: (aMaObjectBuffer maInstVarAt: indexOfId)
						ifAbsent: [ MagmaSoftwareError signal: 'Expected an to find an id' ])
				version: 1) 

				withAllSuperclassDefinitionsDo: 
					[ : eachClassDefinition | 
					referencedClasses add: eachClassDefinition id ] ]! !

!MagmaCompressor methodsFor: 'private' stamp: 'cmm 8/22/2002 21:09'!
markCollectionReferenced: oidInteger

	referencedCollections add: oidInteger! !

!MagmaCompressor methodsFor: 'private' stamp: 'kph 10/5/2006 06:37'!
openInput

	inputController _ MagmaRepositoryController open: sourceLocation ! !

!MagmaCompressor methodsFor: 'initialize-release' stamp: 'kph 10/5/2006 06:36'!
sourceLocation: location

	sourceLocation _ location! !

!MagmaCompressor methodsFor: 'private' stamp: 'cmm 3/29/2007 16:51'!
writeGraphs
	| inputTraverser |
	inputTraverser _ MagmaFileTraverser repositoryController: inputController.
	inputTraverser anchorGraphDo: 
		[ : eachBuffer | 
		self markClassReferenced: eachBuffer.
		(inputController isClassIdForLargeCollection: eachBuffer classId) ifTrue: [ self markCollectionReferenced: eachBuffer oid ].
		(MaOidCalculator isOidForUserObject: eachBuffer oid) ifTrue: 
			[ outputFiler
				write: eachBuffer
					index: true ;
				filesDo: [ : eachFile : eachFileId | eachFile maCommit ] ] ]! !

MaObject subclass: #MagmaFileTraverser
	instanceVariableNames: 'repositoryController oidsVisited graphBuffer directory'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Magma server-Utilities'!

!MagmaFileTraverser commentStamp: 'cmm 8/5/2002 21:45' prior: 0!
I walk the object buffers of an open Magma file and execute a Block for each buffer.

I'm useful for the Magma garbage collection function, the compression function and for general purpose information gathering about what's in the repository.!

!MagmaFileTraverser class methodsFor: 'creation' stamp: 'cmm 1/18/2007 21:37'!
for: pathStringOrMagmaLocalLocation
	^ self new 
		directory: (pathStringOrMagmaLocalLocation asMagmaLocalLocation pathName)
		yourself! !

!MagmaFileTraverser class methodsFor: 'creation' stamp: 'cmm 8/15/2002 22:52'!
repositoryController: aMagmaRepositoryController

	^self new repositoryController: aMagmaRepositoryController! !

!MagmaFileTraverser methodsFor: 'actions' stamp: 'cmm 5/2/2005 10:37'!
anchorGraphDo: aBlock

	self
		graphFrom: (repositoryController requestCritical: [ self repository filer anchorOid ])
		do: aBlock! !

!MagmaFileTraverser methodsFor: 'private' stamp: 'cmm 8/21/2002 00:24'!
close

	repositoryController ifNotNil:
		[ repositoryController requestCritical: [ repositoryController close ] ]! !

!MagmaFileTraverser methodsFor: 'private' stamp: 'cmm 9/24/2004 16:15'!
deleteOidsVisited

	oidsVisited ifNotNil: [ oidsVisited close ].
	directory deleteFileNamed: self tempOidsVisitedFilename! !

!MagmaFileTraverser methodsFor: 'private' stamp: 'cmm 3/15/2009 21:12'!
graphBuffer
	^ graphBuffer! !

!MagmaFileTraverser methodsFor: 'actions' stamp: 'cmm 5/2/2005 10:37'!
graphFrom: oid do: aBlock
	"Value aBlock against every buffer in the receivers file."

	| readStrategy shouldClose |
	shouldClose _ false.
	repositoryController ifNil: [ shouldClose _ true. self open ].
	readStrategy _ MaBasicReadStrategy minimumDepth: 0.
	MagmaNotification signal: 'Beginning traversal of ', oid printString, ' of ', directory pathName.
	graphBuffer _ repositoryController session serializer graphBuffer copy resetBufferSize; yourself.  "sorry Demeter"
	[ self initializeOidsVisited.
	self
		graphFrom: oid
		do: aBlock
		using: readStrategy ]

		ensure:
			[ shouldClose ifTrue: [ self close ].
			self deleteOidsVisited ].
	MagmaNotification signal: 'Traversal of ', oid printString, ' of ', directory pathName, ' complete.'! !

!MagmaFileTraverser methodsFor: 'private' stamp: 'cmm 3/8/2007 12:59'!
graphFrom: oidInteger do: oneArgBlock using: aMaReadStrategy 
	| buffer indexPosition |
	(MaOidCalculator isOidForUserObject: oidInteger) ifFalse: [ ^ self ].
	indexPosition _ MaOidCalculator indexPositionForOid: oidInteger.
	(oidsVisited 
		at: indexPosition
		ifAbsent: [ 0 ]) = 0 ifFalse: [ ^ self ].
	oidsVisited 
		at: indexPosition
		put: oidInteger.
	repositoryController requestCritical: 
		[ buffer _ self repository filer 
			appendObject: oidInteger
			into: graphBuffer ].
	oneArgBlock value: buffer.
	buffer 
		graphDo: oneArgBlock
		using: aMaReadStrategy
		for: self.
	graphBuffer decreaseBufferSizeBy: buffer physicalSize! !

!MagmaFileTraverser methodsFor: 'private' stamp: 'cmm 10/26/2006 14:43'!
graphFromLargeCollection: oidInteger do: oneArgBlock using: aMaReadStrategy
	| fileIndex |
	fileIndex _ (repositoryController 
		requestCritical: 
			[ repositoryController repository collectionManagerFor: oidInteger ]) memberIndex.
	MagmaNotification signal: 'beginning enumeration of ', fileIndex size printString, ' objects in MagmaCollection ', oidInteger printString.
	fileIndex
		do: oneArgBlock
		for: self
		using: aMaReadStrategy.
	MagmaNotification signal: 'MagmaCollection enumeration complete:  ', oidInteger printString! !

!MagmaFileTraverser methodsFor: 'private' stamp: 'cmm 5/2/2005 12:04'!
graphFromPointers: aMaObjectBuffer do: oneArgBlock using: aMaReadStrategy
	"aMaObjectBuffer isPointers must be true to get here."

	aMaObjectBuffer instVarsDoWithIndex:
		[ :eachOid :index |
		self
			graphFrom: eachOid
			do: oneArgBlock
			using: aMaReadStrategy ]! !

!MagmaFileTraverser methodsFor: 'private' stamp: 'cmm 7/12/2005 23:43'!
initializeOidsVisited

	oidsVisited ifNotNil: [ self deleteOidsVisited ].
	oidsVisited _ 
		MaLargeArrayOfNumbers
			createNamed: self tempOidsVisitedFilename
			bitSize: MaObjectFiler maxDbSize! !

!MagmaFileTraverser methodsFor: 'testing' stamp: 'cmm 10/26/2006 18:40'!
isForLargeCollection: aMaFixedObjectBuffer
	^ repositoryController isClassIdForLargeCollection: aMaFixedObjectBuffer classId! !

!MagmaFileTraverser methodsFor: 'private' stamp: 'kph 10/5/2006 06:31'!
open
	"Use if you created with just a path, or if you know the MagmaRepositoryController needs opened.."

	repositoryController _ MagmaRepositoryController open: (MagmaLocalLocation path: directory pathName)! !

!MagmaFileTraverser methodsFor: 'private' stamp: 'cmm 8/13/2002 23:29'!
repository

	^repositoryController repository! !

!MagmaFileTraverser methodsFor: 'accessing' stamp: 'cmm 1/19/2005 21:36'!
repositoryController

	^ repositoryController! !

!MagmaFileTraverser methodsFor: 'private' stamp: 'cmm 9/24/2004 15:54'!
repositoryController: aMagmaRepositoryController

	repositoryController _ aMagmaRepositoryController.
	directory _ aMagmaRepositoryController directory! !

!MagmaFileTraverser methodsFor: 'private' stamp: 'cmm 9/26/2004 21:30'!
tempOidsVisitedFilename

	^ repositoryController directory fullNameFor: 'oidsVisited.tmp'! !

MaObject subclass: #MagmaRepositoryController
	instanceVariableNames: 'repository session serverSerializer requestInterruptGuard localLocation preferences serverStatistics backupProcess'
	classVariableNames: 'OpenControllers ControllersGuard'
	poolDictionaries: ''
	category: 'Magma server-API'!

!MagmaRepositoryController commentStamp: 'cmm 7/29/2007 17:46' prior: 0!
A MagmaRepositoryController is the main user interface to a Magma repository.  With this class you create, open and delete repositories.  

Instance Variables
	localLocation:  a MagmaLocalLocation specifying where the repository is opened on.
	preferences:  A MagmaServerPreferences specifying various server-level options.
	repository:  The center of the Magma server universe, manages server operations.  Private from general public use.
	requestInterruptGuard:  Monitor to allow only one request at a time to execute.
	serverSerializer:  The serializer used to send certain whole graph-buffers back to the client.  (My sessions serializer cannot be used because it is optimized for a MagmaSession; i.e., only serializing each object one level deep, etc.).
	session:  My standard session used to perform standard operations on the model, mostly for LargeCollection support.
!

!MagmaRepositoryController class methodsFor: 'utilities' stamp: 'cmm 2/24/2009 14:24'!
branch: pathStringOrMagmaLocalLocation 
	"Break the repository at pathStringOrMagmaLocalLocation from its original source (if it was a fullBackup).  The repository will get an entirely new MagmaId and be the only copy of its newly associated MagmaNode, so therefore it won't try to join its old MagmaNode when it starts up.  This allows a 'production copy' to be made and launched in a test environment safely.
	Warning:  Once this is performed, the copy of the repository is like an entirely separate repository and will never again be able to join its source."
	| location controller |
	location := pathStringOrMagmaLocalLocation asMagmaLocalLocation.
	
	[ controller := self open: location.
	controller branch ] ensure: [ controller ifNotNil: [ controller close ] ]! !

!MagmaRepositoryController class methodsFor: 'utilities' stamp: 'cmm 1/18/2007 21:29'!
compress: sourceLocation to: targetLocation
	"sourceLocation and targetLocation may be either a String (relative or fully-qualified path) or a MagmaLocalLocation."
	(MagmaCompressor source: sourceLocation) compressTo: targetLocation! !

!MagmaRepositoryController class methodsFor: 'accessing' stamp: 'cmm 7/15/2008 14:27'!
controllerOpenedOn: aMagmaLocalLocation 
	"Answer the RepositoryController already opened on aMagmaLocalLocation.  If no existing controller is already open, answer nil."
	^ self controllersGuard: 
		[ OpenControllers 
			maDetect: [ : each | each localLocation = aMagmaLocalLocation ]
			ifFound: [ : foundController | foundController ensureOpen ]
			ifNone: [ nil ] ]! !

!MagmaRepositoryController class methodsFor: 'private' stamp: 'cmm 8/8/2007 23:58'!
controllersGuard: aBlock 
	^ (ControllersGuard ifNil: [ self initializeControllersGuard ]) critical: aBlock! !

!MagmaRepositoryController class methodsFor: 'utilities' stamp: 'cmm 3/10/2009 20:12'!
create: pathStringOrMagmaLocalLocation 
	"Creates a repository with an initial root of nil.  I #assureExistence of the location."
	| location |
	location := pathStringOrMagmaLocalLocation asMagmaLocalLocation.
	self validateCanCreateIn: location.
	self new createRepositoryAt: location.
	(self openedOn: location) in: 
		[ : rp | 
		
		[ rp fullBackupTo: location seedDirectory.
		[ rp isFullBackupRunning ] whileTrue: [ (Delay forSeconds: 1) wait ] ] ensure: [ rp close ] ]! !

!MagmaRepositoryController class methodsFor: 'utilities' stamp: 'cmm 1/18/2007 21:21'!
create: pathStringOrMagmaLocalLocation root: anObject
	| session location |
	location _ pathStringOrMagmaLocalLocation asMagmaLocalLocation.
	self create: location.
	session _ MagmaSession openLocal: location.
	session
		connect: (MagmaUser id: 'system create:root:') ;
		commit: [ session root: anObject ] ;
		disconnect ;
		closeRepository! !

!MagmaRepositoryController class methodsFor: 'utilities' stamp: 'cmm 7/1/2007 12:32'!
delete: pathStringOrMagmaLocalLocation 
	"Delete the just the files in fullyQualifiedPathString that are part of the Magma repository that resides there."
	pathStringOrMagmaLocalLocation asMagmaLocalLocation delete! !

!MagmaRepositoryController class methodsFor: 'initializing' stamp: 'cmm 3/12/2009 22:22'!
initialize
	OpenControllers ifNotNil: [ OpenControllers do: [ : each | each close ] ].
	OpenControllers := Set new.
	self initializeControllersGuard.
	Smalltalk
		addToStartUpList: self ;
		addToShutDownList: self.
	Smalltalk 
		addToStartUpList: MaServerSocket
		after: self! !

!MagmaRepositoryController class methodsFor: 'initializing' stamp: 'cmm 9/23/2008 14:26'!
initializeControllersGuard
	^ ControllersGuard := Mutex new! !

!MagmaRepositoryController class methodsFor: 'private' stamp: 'cmm 3/5/2009 10:05'!
open: aMagmaLocalLocation
	^ self new open: aMagmaLocalLocation asMagmaLocalLocation! !

!MagmaRepositoryController class methodsFor: 'accessing' stamp: 'cmm 6/25/2008 10:57'!
openControllers
	"The collection of repository's that are opened by this image, not necessarily serving, however."
	^ OpenControllers! !

!MagmaRepositoryController class methodsFor: 'accessing' stamp: 'cmm 7/15/2008 14:25'!
openedOn: aMagmaLocalLocation 
	"Answer the RepositoryController already opened on aMagmaLocalLocation.  If no existing controller is already open, open a new one and answer it."
	| existingController |
	^ self controllersGuard: 
		[ (existingController := self controllerOpenedOn: aMagmaLocalLocation) 
			ifNil: [ self open: aMagmaLocalLocation ]
			ifNotNil: [ existingController ] ]! !

!MagmaRepositoryController class methodsFor: 'initializing' stamp: 'cmm 12/27/2004 23:16'!
shutDown: aboutToQuit

	super shutDown: aboutToQuit.
	MagmaSession shutDown: aboutToQuit.
	OpenControllers
		do:
			[ :each | 
			[ MagmaNotification signal: 'closing Magma repository ', each pathName asString, '.  Will reopen on next startup.' ]
				on: MagmaNotification
				do: [ : noti | Transcript cr; show: noti messageText. noti pass ].
			each primClose ]! !

!MagmaRepositoryController class methodsFor: 'initializing' stamp: 'cmm 3/11/2005 14:58'!
startUp: amResuming

	super startUp: amResuming.
	OpenControllers copy do: 
		[ :each | 
		[ [ each ensureOpen ]
			on: MagmaNotification
			do: [ : noti | Transcript cr; show: noti messageText. noti pass ] ]
				on: MagmaEnvironmentError
				do: 
					[ : error | "not present anymore, remove from OpenControllers and resignal."
					each close.
					error pass ] ].
	"Now that we're open, ensure any server-sockets that may need to restart can be."
	MaServerSocket startUp: amResuming! !

!MagmaRepositoryController class methodsFor: 'private' stamp: 'cmm 5/1/2003 00:24'!
systemSessionUserId

	^'__system'! !

!MagmaRepositoryController class methodsFor: 'validation' stamp: 'cmm 4/29/2007 21:26'!
validateCanCreateIn: aMagmaLocalLocation
	"Magma databases do not share the same directory.  Not only would it be hard to identify which files go with which repository, some of the filenames are based on oids, which would not always be unique across repositories."
	(aMagmaLocalLocation isRepository)
		ifTrue:
			[ MagmaUserError signal: aMagmaLocalLocation pathName, ' is already a Magma repository.  A directory may contain only one Magma repository.' ]! !

!MagmaRepositoryController methodsFor: 'backup / restore' stamp: 'cmm 8/1/2007 22:16'!
archiveCommitLog
	repository archiveCommitLog! !

!MagmaRepositoryController methodsFor: 'backup / restore' stamp: 'cmm 3/4/2009 15:00'!
backupProcess
	^ backupProcess! !

!MagmaRepositoryController methodsFor: 'unsupported' stamp: 'cmm 6/6/2005 13:35'!
beReadOnly

	self requestCritical: [ preferences beReadOnly ]! !

!MagmaRepositoryController methodsFor: 'unsupported' stamp: 'cmm 6/6/2005 13:35'!
beReadWrite

	self requestCritical: [ preferences beReadWrite ]! !

!MagmaRepositoryController methodsFor: 'private' stamp: 'cmm 2/24/2009 15:58'!
branch
	repository initializeBranchCode.
	self session commit: 
		[ | oldMagmaId |
		oldMagmaId := self definition magmaId.
		self definition initializeMagmaId.
		self definition magmaId
			name: oldMagmaId name ;
			description: oldMagmaId description ]! !

!MagmaRepositoryController methodsFor: 'accessing' stamp: 'cmm 11/18/2008 12:54'!
branchCode
	"The branchCode is a randomly-generated number used to prevent accidentally applying the wrong commit-logs to a repository.  When performing a forward recovery, each CommitLogRecord's #branchCode must match mine.  Every time a commitRestore is performed, this branchCode is updated to a new number."
	^ repository branchCode! !

!MagmaRepositoryController methodsFor: 'private' stamp: 'cmm 10/25/2008 15:05'!
branchCode: anInteger 
	"Private - Do not update the branchCode from outside!!"
	repository branchCode: anInteger! !

!MagmaRepositoryController methodsFor: 'preferences' stamp: 'cmm 7/5/2005 20:16'!
cacheFlushFrequency

	^ preferences cacheFlushFrequency! !

!MagmaRepositoryController methodsFor: 'preferences' stamp: 'cmm 9/12/2007 21:52'!
cacheFlushFrequency: numberOfSeconds 
	^ self requestCritical: [ preferences cacheFlushFrequency: numberOfSeconds ]! !

!MagmaRepositoryController methodsFor: 'high-availability' stamp: 'cmm 10/23/2008 10:50'!
catch: thisLocation upTo: aMagmaRemoteLocation 
	"I am open on a full-backup of the repository at aMagmaRemoteLocation.  Download commit-log records until I am caught up and then I will begin receiving real-time updates from the server at aMagmaRemoteLocation."
	repository 
		catch: thisLocation
		upTo: aMagmaRemoteLocation! !

!MagmaRepositoryController methodsFor: 'accessing' stamp: 'cmm 6/2/2008 22:08'!
close
	"requestCritical: in case an extractProcess managed to queue one last request in at the last second.."
	self requestCritical: 
		[ self class controllersGuard: 
			[ OpenControllers 
				remove: self
				ifAbsent: 
					[ "do nothing"
					 ] ].
		self primClose ]! !

!MagmaRepositoryController methodsFor: 'private' stamp: 'cmm 10/26/2004 21:54'!
codeBaseByteArray

	^Array
		with: repository classDefinitionsByteArray
		with: (serverSerializer serializeGraph: self definition codeBase) trimmedByteArray! !

!MagmaRepositoryController methodsFor: 'backup / restore' stamp: 'cmm 7/1/2007 11:23'!
commitLogDirectory
	^ self localLocation commitLogDirectory! !

!MagmaRepositoryController methodsFor: 'accessing' stamp: 'cmm 7/15/2008 14:52'!
commitNumber
	"Answer the number of commits ever applied to this repository.  This is useful for determining which commit.n.log files are no longer needed (all which are lower than this number are no longer needed)."
	^ repository commitNumber! !

!MagmaRepositoryController methodsFor: 'high-availability' stamp: 'cmm 7/9/2007 09:27'!
commitRestore
	"Bring the repository out of restore mode and update my branchCode.  No more commit logs with the old branchCode may be applied."
	repository commitRestore! !

!MagmaRepositoryController methodsFor: 'initializing' stamp: 'cmm 1/16/2005 21:54'!
connect

	session connect: (MagmaUser id: self class systemSessionUserId).
	repository initializeSpecialOidsList! !

!MagmaRepositoryController methodsFor: 'unsupported' stamp: 'cmm 6/23/2002 10:48'!
connections
	"Used for monitoring server status via some console gui."

	self maMarked: 'dev'.! !

!MagmaRepositoryController methodsFor: 'private' stamp: 'cmm 9/16/2008 22:21'!
createRepositoryAt: aMagmaLocalLocation 
	| anchor |
	self initializeSession.
	self session initializeDefinition.
	self session definition addMinimumClassesForOdbmsUsing: self serverSerializer.
	localLocation := aMagmaLocalLocation.
	self
		repository: (MaObjectRepository 
				create: localLocation directory
				controller: self) ;
		connect ;
		forceWrite: (anchor := MaRootAnchor new
				definition: self definition ;
				yourself).
	repository filer
		anchorOid: (self session oidFor: anchor) ;
		classDefinitionsOid: (self session oidFor: self definition classDefinitions) ;
		definitionOid: (self session oidFor: self definition).
	self close! !

!MagmaRepositoryController methodsFor: 'accessing' stamp: 'cmm 5/2/2004 20:40'!
definition
	"Answers a MagmaRepositoryDefinition"

	^ session definition! !

!MagmaRepositoryController methodsFor: 'backup / restore' stamp: 'cmm 8/3/2007 11:02'!
deleteUnnecessaryCommitLogs
	"Delete all commit log files from my 'commits' subdirectory that have already been written to my repository files and, therefore, are no longer needed by this repository.  This is not performed automatically because the files might be needed for one or more warm backups."
	repository deleteUnnecessaryCommitLogs! !

!MagmaRepositoryController methodsFor: 'accessing' stamp: 'cmm 5/11/2007 12:37'!
directory
	"Answer the directory in which my files reside."
	"Note, this is the same FileDirectory that MaObjectFiler has.  It was necessary to duplicate this here to support the auto-reopen feature upon image restart."
	^ localLocation directory! !

!MagmaRepositoryController methodsFor: 'high-availability' stamp: 'cmm 11/18/2008 12:44'!
ensureInRestoreMode
	repository restoreMode ifFalse: [ repository beginRestore ].
	^ nil! !

!MagmaRepositoryController methodsFor: 'private' stamp: 'cmm 2/20/2005 16:31'!
ensureOpen

	self isOpen ifFalse: [ self open ]! !

!MagmaRepositoryController methodsFor: 'private' stamp: 'cmm 3/5/2009 23:39'!
executeSwapWith: aMagmaRemoteLocation commitNumber: commitNumber branchCode: branchCode 
	"I am the primary.  aMagmaRemoteLocation wants to be the primary now."
	| commitLogRecords currentCommitNumber |
	self branchCode = branchCode ifFalse: [ MagmaDifferentBranch signal: 'Branch codes don''t match?!!' ].
	currentCommitNumber := self commitNumber.
	commitLogRecords := repository commitRecordsFrom: commitNumber.
	commitLogRecords notEmpty and: 
		[ commitLogRecords last commitNumber = currentCommitNumber ifFalse: 
			[ MagmaEnvironmentError signal: 'Warm backup is too far behind at this time!!' ] ].
	"We will just manually update the node rather than use MagmaNodeUpdate so its simpler and more explicit.  Using MagmaNodeUpdate *may* work, I'm not sure.."
	self session commit: [ self node swapWithPrimary: aMagmaRemoteLocation ].
	repository beginRestore.
	"Now, capture that last commit!!"
	commitLogRecords := commitLogRecords , (repository commitRecordsFrom: currentCommitNumber + 1).
	^ commitLogRecords! !

!MagmaRepositoryController methodsFor: 'preferences' stamp: 'cmm 10/22/2006 21:38'!
filePoolSize: anInteger 
	"Set the maximum number of simultaneously-open files that Magma will have."
	repository filePoolSize: anInteger! !

!MagmaRepositoryController methodsFor: 'accessing' stamp: 'kph 4/27/2007 02:53'!
filenames
	^ localLocation fileNames! !

!MagmaRepositoryController methodsFor: 'preferences' stamp: 'cmm 7/7/2005 23:21'!
flushAggressively

	^ preferences flushAggressively! !

!MagmaRepositoryController methodsFor: 'preferences' stamp: 'cmm 7/5/2005 12:51'!
flushAggressively: aBoolean

	^ self requestCritical: [ preferences flushAggressively: aBoolean ]! !

!MagmaRepositoryController methodsFor: 'private' stamp: 'cmm 2/10/2009 14:04'!
forceWrite: anObject 
	"This is somewhat of a hacking method to enable us to write objects needed
	for the repository to operate (i.e., commit transactions) during initial creation."
	| commitPackage answer |
	commitPackage := MaCommitPackage new.
	anObject maIsLargeCollection ifTrue: [ commitPackage addLargeCollectionChanges: anObject changes ].
	session serializer 
		serializeGraph: anObject
		do: [ : eachObject | commitPackage addObject: eachObject ].
	"Prime the serializer."
	session begin.
	commitPackage copy serializeObjectsUsing: session.
	commitPackage serializeObjectsUsing: session.
	answer := self forceWritePackage: commitPackage.
	session 
		refreshViewUsing: answer
		includingLocal: false.
	^ answer! !

!MagmaRepositoryController methodsFor: 'private' stamp: 'cmm 2/10/2009 14:04'!
forceWritePackage: aMaCommitPackage 
	"This is somewhat of a hacking method to enable us to write objects needed
	for the repository to operate (i.e., commit transactions) during initial creation."
	^ 
	[ repository 
		submitAll: aMaCommitPackage
		for: session id
		beginAnother: false ] ensure: [ session decrementTransactionLevel ]! !

!MagmaRepositoryController methodsFor: 'high-availability' stamp: 'cmm 12/29/2008 19:45'!
forkCatch: thisLocation upTo: aMagmaRemoteLocation 
	"Kick off a background process to catch up to the server running at aMagmaRemoteLocation."
	[ self 
		catch: thisLocation
		upTo: aMagmaRemoteLocation ] newProcess
		name: 'Catching up to ' , aMagmaRemoteLocation printString ;
		resume! !

!MagmaRepositoryController methodsFor: 'backup / restore' stamp: 'cmm 3/8/2009 12:15'!
fullBackupTo: aFileDirectory 
	"Backup the repository to aFileDirectory while the repository is open.  Files in the aFileDirectory will be overwritten."
	self isFullBackupRunning ifTrue: [ MagmaUserError signal: 'Full backup is already running.' ].
	aFileDirectory fullName = self directory fullName ifTrue: 
		[ MagmaUserError signal: 'Must not back up to running server''s directory!!' ].
	aFileDirectory isRemoteDirectory 
		ifTrue: 
			[ (aFileDirectory fileExists: MaObjectFiler objectsPhysicalFilename) ifTrue: 
				[ MagmaDeleteWarning signal: 'The repository at ' , aFileDirectory printString , ' will be deleted'.
				aFileDirectory deleteLocalFiles ] ]
		ifFalse: [ (MagmaLocalLocation path: aFileDirectory fullName) delete ].
	(aFileDirectory exists and: [ aFileDirectory entries notEmpty ]) ifTrue: 
		[ MagmaEnvironmentError signal: 'Could not delete old backup.  In use?' ].
	aFileDirectory assureExistence.
	self requestCritical: 
		[ "start new commit log that begins with the very next transaction after this backup."
		| backupBlock |
		self archiveCommitLog.
		backupProcess := MaClientProcess doBlock: 
			[ | tempDir filenames |
			
			[ tempDir := FileDirectory on: (repository directory fullNameFor: 'temp').
			tempDir exists ifTrue: [ tempDir recursiveDelete ].
			tempDir assureExistence.
			repository flushCritical: 
				[ "Flush now so we minimize the amount kept in RAM, and to ensure the backup is at the most current state."
				repository flushCache.
				filenames := self filenames.
				backupProcess progressor
					taskSize: filenames size + 2 ;
					unitsVerbPhrase: 'files copied' ;
					description: 'copying ' , MaObjectFiler objectsPhysicalFilename , ' to temp'.
				backupProcess shouldStop ifFalse: 
					[ repository directory 
						maCopyFileNamed: MaObjectFiler objectsPhysicalFilename
						toDirectory: tempDir ].
				"Let's not wait for resume because the server is eating memory while waiting!!"
				"				backupProcess waitForResume."
				backupProcess progressor advance.
				filenames do: 
					[ : each | 
					({ 
						(MaRecoveryManager applyFilename).
						(MaObjectFiler objectsPhysicalFilename)
					 } includes: each) ifFalse: 
						[ backupProcess shouldStop ifFalse: 
							[ backupProcess progressor description: 'Copying ' , each.
							repository directory 
								maCopyFileNamed: each
								toDirectory: aFileDirectory.
							"							backupProcess waitForResume."
							backupProcess progressor advance ] ] ] ].
			(tempDir notNil and: [ backupProcess shouldStop not ]) ifTrue: 
				[ (MaObjectFiler open: tempDir)
					restoreMode: true ;
					close.
				tempDir 
					maCopyFileNamed: MaObjectFiler objectsPhysicalFilename
					toDirectory: aFileDirectory.
				backupProcess progressor advance ] ] ensure: 
				[ backupProcess stopped 
					ifTrue: [ backupProcess signalWaiterOnStop ]
					ifFalse: [ backupProcess progressor description: 'Backup complete.' ].
				tempDir ifNotNil: [ tempDir exists ifTrue: [ tempDir recursiveDelete ] ] ] ].
		backupProcess
			name: 'backing up to ' , aFileDirectory fullName ;
			start ]! !

!MagmaRepositoryController methodsFor: 'backup / restore' stamp: 'cmm 3/7/2009 14:22'!
fullBackupToServerPath: pathString 
	"Backup the repository to the directory identified by pathString while the repository is open.  If there is a Magma repository at pathString, all files will be deleted ahead of time, otherwise none will."
	| didDelete |
	didDelete := false.
	[ self fullBackupTo: (MagmaLocalLocation path: pathString) directory ] 
		on: MagmaDeleteWarning
		do: 
			[ : warn | 
			didDelete := true.
			warn resume ].
	^ didDelete 
		ifTrue: [ 'Backup at ' , pathString , ' is being replaced with a new backup.' ]
		ifFalse: [ 'A new backup is being created at ' , pathString ]! !

!MagmaRepositoryController methodsFor: 'initializing' stamp: 'cmm 12/17/2008 17:14'!
initialize
	super initialize.
	requestInterruptGuard := Mutex new.
	preferences := MagmaServerPreferences new.
	self serverStatistics: MagmaServerStatistics new! !

!MagmaRepositoryController methodsFor: 'initializing' stamp: 'cmm 3/27/2002 22:00'!
initializeServerSerializer
	"Why is this separate, 'serverSerializer' needed?  Because the session serializer requires a different traversalStrategy.  The serverSerializer needs to be able to fully serialize, to the ends of the graphs, the 'response' objects of various requests."

	serverSerializer _ self session serializer copyWithNewBuffer initializeTraversalStrategy
! !

!MagmaRepositoryController methodsFor: 'initializing' stamp: 'cmm 5/28/2007 22:17'!
initializeSession
	session _ MagmaSession repositoryController: self.
	session readStrategy: self serverReadStrategy.
	self initializeServerSerializer! !

!MagmaRepositoryController methodsFor: 'private' stamp: 'cmm 9/1/2003 19:40'!
initializeSystemReadStrategy

	repository initializeSystemReadStrategyUsing: self serverSerializer classIdManager! !

!MagmaRepositoryController methodsFor: 'accessing' stamp: 'cmm 10/26/2006 14:55'!
isClassIdForLargeCollection: anInteger 
	session serializer classIdManager largeCollectionIdsDo: [ : each | each = anInteger ifTrue: [ ^ true ] ].
	^ false! !

!MagmaRepositoryController methodsFor: 'backup / restore' stamp: 'cmm 3/5/2009 16:41'!
isFullBackupRunning
	^ self requestCritical: 
		[ backupProcess notNil and: [ backupProcess everStarted and: [ backupProcess isRunning ] ] ]! !

!MagmaRepositoryController methodsFor: 'accessing' stamp: 'cmm 6/2/2008 13:31'!
isOpen
	^ repository notNil and: [ repository isOpen ]! !

!MagmaRepositoryController methodsFor: 'preferences' stamp: 'cmm 6/6/2005 13:38'!
isReadOnly

	^ preferences isReadOnly! !

!MagmaRepositoryController methodsFor: 'accessing' stamp: 'kph 4/27/2007 02:56'!
localLocation
	"Answer the localLocation where my files reside."
	 
	^ localLocation! !

!MagmaRepositoryController methodsFor: 'backup / restore' stamp: 'cmm 8/2/2007 00:16'!
logArchiveFrequency
	"Answer the frequency which commit.log files are closed and a new one started.  Commit log files are used to apply to a backup to bring it up to date."
	^ repository logArchiveFrequency! !

!MagmaRepositoryController methodsFor: 'backup / restore' stamp: 'cmm 8/2/2007 00:16'!
logArchiveFrequency: aDuration 
	"Override the default frequency which commit.log files are closed and a new one started.  Commit log files are used to apply to a backup to bring it up to date.

	Example:  myRepositoryController logArchiveFrequency: 15 minutes."
	repository logArchiveFrequency: aDuration! !

!MagmaRepositoryController methodsFor: 'accessing' stamp: 'cmm 8/30/2008 10:14'!
magmaId
	^ self definition ifNotNilDo: [ : definition | definition magmaId ]! !

!MagmaRepositoryController methodsFor: 'private' stamp: 'cmm 6/10/2005 10:59'!
materializeClassDefinitions
	"This is needed because, when opening the repository, the default classes do not include MagmaRepositoryDefinition, which is delivered as part of any session-connection.  When the system-session attempts to connect, it tries to read the repository-definition but can't because it can't get a proper read-strategy for it."

	[ session loadClassDefinitionsFrom: repository classDefinitionsByteArray ]
		on: Error
		do:
			[ : err | MagmaCorruptionError signal: err messageText ]! !

!MagmaRepositoryController methodsFor: 'unsupported' stamp: 'cmm 5/2/2004 20:40'!
maximumNumberOfChallengers: anInteger 
	self session 
		commit: [ self definition maximumNumberOfChallengers: anInteger ]! !

!MagmaRepositoryController methodsFor: 'private' stamp: 'cmm 11/7/2008 14:32'!
newSessionFor: aMaRepositoryConnectionRequest sessionId: aUuid 
	^ self repository 
		newSessionFor: aMaRepositoryConnectionRequest userId
		sessionId: aUuid
		numberOfChallengers: MagmaRepositoryDefinition defaultMaximumNumberOfChallengers
		clientConnection: aMaRepositoryConnectionRequest clientConnection! !

!MagmaRepositoryController methodsFor: 'private' stamp: 'cmm 10/22/2008 16:53'!
newTransactionFor: sessionId 
	preferences isReadOnly ifTrue: 
		[ MagmaUserError signal: 'This repository is set to read-only, commits are prohibited.' ].
	self restoreMode ifTrue: 
		[ MagmaUserError signal: 'Transactions are prohibited because this repository is in restore mode.' ].
	^ repository newTransactionFor: sessionId! !

!MagmaRepositoryController methodsFor: 'high-availability' stamp: 'cmm 10/30/2008 16:33'!
node
	^ self definition node! !

!MagmaRepositoryController methodsFor: 'statistics' stamp: 'cmm 5/2/2003 14:56'!
numberOfConnectedClients

	^repository sessions size - 1! !

!MagmaRepositoryController methodsFor: 'statistics' stamp: 'cmm 12/22/2008 13:16'!
oidCount
	"The number of oids allocated by this repository.  The total number of objects ever committed, even if some have since been garbage-collected."
	^ repository oidCount! !

!MagmaRepositoryController methodsFor: 'private' stamp: 'cmm 3/2/2009 16:45'!
open
	"This is called when restarting controllers during image restart."
	| didRetry |
	didRetry := false.
	[ self privateOpen: localLocation ] 
		on: MagmaCorruptionError
		do: 
			[ : err | 
			didRetry 
				ifTrue: [ err pass ]
				ifFalse: 
					[ MagmaRollbackRecoveryNeeded signal: 'Outage occurred while writing system-definitions!!  Will now attempt to repair.'.
					repository primitiveRecover.
					err retry ] ].
	"Do NOT cascade the following messages.  rollbackIfNecessary may replace my repository with a new one."
	repository rollbackIfNecessary.
	repository rollforwardIfPossible! !

!MagmaRepositoryController methodsFor: 'private' stamp: 'cmm 7/8/2008 19:19'!
open: aMagmaLocalLocation 
	localLocation := aMagmaLocalLocation.
	self class controllersGuard: [ self open ]! !

!MagmaRepositoryController methodsFor: 'accessing' stamp: 'cmm 5/11/2007 12:36'!
pathName
	^ localLocation pathName! !

!MagmaRepositoryController methodsFor: 'private' stamp: 'cmm 11/24/2008 12:45'!
primClose
	self requestCritical: 
		[ session ifNotNil: 
			[ session disconnect.
			session linksDo: [ : each | each localRequestServer shutdown ] ].
		self serverStatistics release.
		repository close ]! !

!MagmaRepositoryController methodsFor: 'private' stamp: 'cmm 7/8/2008 19:19'!
privateOpen: aMagmaLocalLocation 
	self isOpen ifTrue: 
		[ MagmaUserError signal: 'Repository ' , aMagmaLocalLocation asString , ' is already open.' ].
	MagmaNotification signal: 'opening Magma repository ' , self pathName.
	self
		initializeSession ;
		repository: (MaObjectRepository 
				open: aMagmaLocalLocation directory
				controller: self) ;
		initializeSystemReadStrategy ;
		materializeClassDefinitions ;
		connect ;
		refreshView	"so more classes will be pulled from unusedDepths of the SystemReadStrategy".
	OpenControllers add: self! !

!MagmaRepositoryController methodsFor: 'client/server required methods' stamp: 'cmm 3/6/2009 10:48'!
processRequest: aMagmaRepositoryRequest 
	aMagmaRepositoryRequest repositoryController: self.
	^ [ self requestCritical: [ aMagmaRepositoryRequest process ] ] 
		on: Error
		do: 
			[ : anError | 
			self maMarked: 'dev'.	"This error routine must be improved.  It needs to log and notify for unhandled errors!!"
			(anError maOriginalClass = MagmaLowOnMemory and: [ SmalltalkImage current maIsHeadless ]) ifTrue: 
				[ self 
					serverSave: false
					andExit: true	"saving the image requires a lot of memory" ].
			(MagmaPreferences debug and: 
				[ ({ 
					MagmaCommitConflictError.
					MagmaSessionLost.
					MagmaWrongCommitNumber.
					MagmaWrongServerError
				 } includes: anError maOriginalClass) not and: 
					[ anError isMaUserError not and: [ SmalltalkImage current maIsHeadless not ] ] ]) ifTrue: [ self halt: anError messageText ].
			anError ]! !

!MagmaRepositoryController methodsFor: 'private' stamp: 'cmm 12/19/2008 14:02'!
refreshView
	self session refresh.
	repository systemReadStrategy makeReadyForUseUsing: self session serializer classIdManager! !

!MagmaRepositoryController methodsFor: 'backup / restore' stamp: 'cmm 3/2/2009 15:49'!
replayCommitLogs
	"Search my 'commits' subdirectory for the proper commit log file to begin replaying from.  Replay it and any subsequent commit log files as far forward possible."
	repository rollforwardIfPossible! !

!MagmaRepositoryController methodsFor: 'private' stamp: ''!
repository

	^repository
! !

!MagmaRepositoryController methodsFor: 'private' stamp: 'cmm 12/31/2002 12:54'!
repository: aMaObjectRepository

	repository _ aMaObjectRepository.
	aMaObjectRepository repositoryController: self! !

!MagmaRepositoryController methodsFor: 'private' stamp: 'cmm 12/27/2002 15:55'!
requestCritical: aBlock
	"Coordinate access to all server elements."

	^requestInterruptGuard critical: aBlock! !

!MagmaRepositoryController methodsFor: 'backup / restore' stamp: 'cmm 10/22/2008 15:29'!
restoreMode
	"Answer whether the receiver is operating in restoreMode.  Warm backups operate in restoreMode, taking commits broadcasted to them from the 'primary' and applying them."
	^ repository restoreMode! !

!MagmaRepositoryController methodsFor: 'private' stamp: 'cmm 7/14/2008 16:21'!
serverReadStrategy
	"Make sure the indexes stays a proxy because the server will not necessarily have the full MagmaCollectionIndex hierarchy."
	^ (MaReadStrategy minimumDepth: 1)
		forVariableNamed: 'changes'
			onAny: MagmaCollection
			readToDepth: 99999 ;
		forVariableNamed: 'indexes'
			onAny: MagmaCollection
			readToDepth: -1 ;
		forVariableNamed: 'descriptions'
			onAny: MagmaCollection
			readToDepth: 99999 ;
		"For MagmaIdRequest"
		onAny: MagmaId
			readToDepth: 99999 ;
		yourself
! !

!MagmaRepositoryController methodsFor: 'private' stamp: 'cmm 12/21/2008 15:37'!
serverSave: saveOption andExit: exitOption 
	"Wait one second to give time for the response to be sent out so the client doesn't get a NetworkError."
	[ (Delay forSeconds: 1) wait.
	SmalltalkImage current 
		snapshot: saveOption
		andQuit: exitOption ] fork! !

!MagmaRepositoryController methodsFor: 'private' stamp: 'cmm 11/7/2008 14:23'!
serverSaveAndExit
	"Wait one second to give time for the response to be sent out so the client doesn't get a NetworkError."
	
	[ (Delay forSeconds: 1) wait.
	SmalltalkImage current 
		snapshot: true
		andQuit: true ] fork! !

!MagmaRepositoryController methodsFor: 'private' stamp: ''!
serverSerializer
	"The serverSerializer is used for processing client requests.  It is also
	used during administration and possibly needed during initial creation of a
	repository."

	^serverSerializer
! !

!MagmaRepositoryController methodsFor: 'accessing' stamp: 'cmm 5/7/2008 17:23'!
serverStatistics
	^ serverStatistics! !

!MagmaRepositoryController methodsFor: 'initializing' stamp: 'cmm 9/29/2008 23:09'!
serverStatistics: aMagmaServerStatistics 
	serverStatistics := aMagmaServerStatistics! !

!MagmaRepositoryController methodsFor: 'private' stamp: 'cmm 10/30/2008 16:53'!
session
	^ session! !

!MagmaRepositoryController methodsFor: 'accessing' stamp: 'cmm 4/3/2003 22:59'!
sessionId

	^session id
! !

!MagmaRepositoryController methodsFor: 'private' stamp: 'cmm 3/25/2009 11:30'!
softBranch
	"A softBranch is one in which only the branchCode changes, not the MagmaId.  When the user specifies #branch, it is a 'hard branch', and can never rejoin the original."
	repository initializeBranchCode! !

!MagmaRepositoryController methodsFor: 'high-availability' stamp: 'cmm 11/26/2008 08:50'!
validateLastBranchCodeUsing: primarySession 
	"Validate that the branchCode of my last commit-record is the same as the branchCode of the primary repository I am becoming a warm-backup for."
	"There is a slim chance that a warm-backup will be exactly one-commit behind if the network cable to the primary server is cut after it logged one last commit, but before the bits made it out of the nic to the warm backup.  So the last commit warm backup (B) received was 172, but the crashed primary (A) actually got 173 written to its log.  Then a different commit 173 has since occurred to the new primary (B).  Now (A) is starting back up intending to catch back up to being a warm backup for (B), but this must be prevented because it has a different 173.

What follows are just my initial notes about this use-case:

What to do on restart of the primary for each scenario:

	#(''backup +1'' ''primary +0'') - Primary can catch up
	#(''backup +0'' ''primary +1'') 
		- Network cable cut in middle of commit 173, just before sending to warm backup.  So warm backup is still at 172.
		- Note committing client is now waiting for response which will timeout.  Therefore, no expectations that commit 173 succeeded.
		- ConnectionClosed signaled to other clients causes failover.  A totally different commit 173 happens on the new primary!!
		- Now, let's say the log:flush: happens on the original primary in which the network cable was cut.
		- Now, on restarting, that commit 173 (the +1 commit) is different than the 173 of the warm backup.
			A:  When restarting the crashed server, compare the branchCode of the last commitRecord (173) between the two.  If they are different, the repositories are branched and the crashed server must be reestablished, it cannot be immediately failed back over."
	| myBranchCode correspondingCommitRecordInPrimary primaryBranchCode |
myBranchCode := self branchCode .
	correspondingCommitRecordInPrimary := (primarySession commitLogRecordsFrom: self commitNumber) first.
	"The moment of truth.."
	primaryBranchCode := correspondingCommitRecordInPrimary branchCode.
correspondingCommitRecordInPrimary commitNumber = self commitNumber ifFalse: [ MagmaUnavailableCommitRecord signal: 'Commit record unavailable, won''t be able to roll forward.'].
	primaryBranchCode = myBranchCode ifFalse: 
		[ "Oh noo!!  This (or the other) copy of the repository will have to be deleted and re-established via #fullBackup from the new primary."
		MagmaDifferentBranch signal: 'Commit ' , self commitNumber asString , ' is different between the primary and this repository, they are branched.  Neither can be a warm backup for the other.' ]! !

!MagmaRepositoryController methodsFor: 'client/server required methods' stamp: 'cmm 9/19/2004 22:08'!
value: aMagmaRepositoryRequest

	^ self processRequest: aMagmaRepositoryRequest! !

!MagmaRepositoryController methodsFor: 'statistics' stamp: 'cmm 4/21/2003 16:02'!
version

	^repository version! !

MaObject subclass: #MagmaServerPreferences
	instanceVariableNames: 'attributes'
	classVariableNames: 'DebugRecovery'
	poolDictionaries: ''
	category: 'Magma server-Preferences'!

!MagmaServerPreferences class methodsFor: 'debugging' stamp: 'cmm 6/9/2005 21:42'!
debugRecovery

	^ DebugRecovery! !

!MagmaServerPreferences class methodsFor: 'debugging' stamp: 'cmm 6/9/2005 21:42'!
debugRecovery: aBoolean

	DebugRecovery _ aBoolean! !

!MagmaServerPreferences class methodsFor: 'debugging' stamp: 'cmm 6/9/2005 22:06'!
initialize

	self debugRecovery: false! !

!MagmaServerPreferences methodsFor: 'security' stamp: 'cmm 6/6/2005 13:35'!
beReadOnly

	self 
		setAttribute: #isReadOnly
		to: true! !

!MagmaServerPreferences methodsFor: 'security' stamp: 'cmm 6/6/2005 13:35'!
beReadWrite

	self 
		setAttribute: #isReadOnly
		to: false! !

!MagmaServerPreferences methodsFor: 'recovery' stamp: 'cmm 7/18/2005 12:48'!
cacheFlushFrequency

	^ self
		valueOfAttribute: #cacheFlushFrequency
		ifAbsent: [ 5 "seconds" ]! !

!MagmaServerPreferences methodsFor: 'recovery' stamp: 'cmm 7/5/2005 20:18'!
cacheFlushFrequency: anInteger

	self
		setAttribute: #cacheFlushFrequency
		to: anInteger! !

!MagmaServerPreferences methodsFor: 'recovery' stamp: 'cmm 9/12/2007 21:49'!
flushAggressively
	^ self 
		valueOfAttribute: #flushAggressively
		ifAbsent: [ true ]! !

!MagmaServerPreferences methodsFor: 'recovery' stamp: 'cmm 7/12/2005 13:10'!
flushAggressively: aBoolean

	self
		setAttribute: #flushAggressively
		to: aBoolean! !

!MagmaServerPreferences methodsFor: 'initialize' stamp: 'cmm 6/6/2005 13:05'!
initialize
	super initialize.
	attributes _ Dictionary new! !

!MagmaServerPreferences methodsFor: 'security' stamp: 'cmm 6/6/2005 13:44'!
isReadOnly

	^ self 
		valueOfAttribute: #isReadOnly
		ifAbsent: [ false ]! !

!MagmaServerPreferences methodsFor: 'private' stamp: 'cmm 6/6/2005 13:07'!
setAttribute: attributeSymbol to: value

	attributes
		at: attributeSymbol
		put: value! !

!MagmaServerPreferences methodsFor: 'private' stamp: 'cmm 6/6/2005 13:09'!
valueOfAttribute: attributeSymbol ifAbsent: aBlock
	^ attributes 
		at: attributeSymbol
		ifAbsent: aBlock! !

!MagmaBeWarmBackupRequest methodsFor: '*magma server' stamp: 'cmm 2/27/2009 15:35'!
nodeCheck
	self shouldBeSecondary ! !

!MagmaBeWarmBackupRequest methodsFor: '*magma server' stamp: 'cmm 2/27/2009 15:12'!
process
	console ifNil: 
		[ MagmaEnvironmentError signal: 'MagmaServerConsole required for ' , self class name ].
	console beWarmBackupFor: primaryLocation! !

!MagmaPathNameRequest methodsFor: '*magma server' stamp: 'cmm 9/24/2004 15:13'!
process

	^ repositoryController pathName! !

!MaTerm methodsFor: '*magma server-private' stamp: 'cmm 8/12/2007 20:39'!
addMinimumEnumerationsTo: aMaQueryTrunk using: aMagmaLargeCollectionManager 
	self subclassResponsibility! !

!MaTerm methodsFor: '*magma server-private' stamp: 'cmm 8/12/2007 20:40'!
minimumSearchSpaceSizeUsing: aMagmaCollectionManager
	self subclassResponsibility! !

!MaTerm methodsFor: '*magma server-private' stamp: 'cmm 8/12/2007 20:41'!
shouldInclude: oid using: aMagmaCollectionManager 
	self subclassResponsibility ! !

!MagmaCollection methodsFor: '*magma server' stamp: 'cmm 7/12/2005 12:52'!
implementationClass

	^ MaHashIndex! !

!MagmaCollection methodsFor: '*magma server' stamp: 'cmm 3/17/2005 21:31'!
newManagerUsing: aMaObjectRepository

	^ (MagmaCollectionManager
		collection: self
		repository: aMaObjectRepository)
			oidOfIndexesCollection: (aMaObjectRepository session oidFor: self indexes) ;
			yourself! !

!MaCommitRequest methodsFor: '*magma server' stamp: 'cmm 11/16/2008 19:12'!
nodeCheck
	self shouldBePrimary ! !

!MaCommitRequest methodsFor: '*magma server' stamp: 'cmm 11/16/2008 19:22'!
process
	| nodeUpdates answer |
	answer := 
	[ self repository 
		submitAll: package
		for: sessionId
		beginAnother: beginAnother ] 
		on: MagmaNodeUpdate
		do: 
			[ : update | 
			nodeUpdates := nodeUpdates ifNil: [ OrderedCollection new ].
			nodeUpdates add: update.
			update resume ].
	nodeUpdates ifNotNil: 
		[ "this commit refreshes my sessions view"
		self session commit: [ nodeUpdates do: [ : each | each applyTo: repositoryController node ] ] ].
	^ answer! !

!MaKeyIntervalInfoRequest methodsFor: '*magma server' stamp: 'cmm 5/14/2005 18:31'!
process

	^ self repository
		numberOfEntriesFrom: lowKey
		to: highKey
		inLargeCollection: collectionOid
		on: attribute! !

!MaAnchorRequest methodsFor: '*magma server' stamp: 'cmm 5/14/2005 18:09'!
process

	^ self repository
		readAnchorFor: sessionId
		using: readStrategy! !

!MaRawBufferRequest methodsFor: '*magma server' stamp: 'cmm 3/15/2009 19:10'!
process
	"Answer an Array, the first element is the MaObjectBuffer, the second is the MaCommitLogRecord, if available."
	| buffer answer |
	answer := OrderedCollection new.
	buffer := answer add: (self repository filer bufferFor: oid).
	"Also get the referenced buffers, it will be the first thing asked for from this client."
	buffer isPointers ifTrue: 
		[ buffer instVarsDoWithIndex: 
			[ : eachOid : index | 
			(MaOidCalculator isOidForUserObject: eachOid) ifTrue: [ answer add: (self repository filer bufferFor: eachOid) ] ] ].
	^ answer asArray! !

!MagmaEnsureCorrectNodeConfiguration methodsFor: '*magma server' stamp: 'cmm 11/14/2008 14:38'!
process
	console ensureCorrectNodeConfiguration.
	^ nil! !

!MaLargeArrayOfRecords methodsFor: '*magma server' stamp: 'cmm 10/22/2006 15:07'!
filePool: aMaFileStreamPool
	broker filePool: aMaFileStreamPool! !

!MaLargeArrayOfRecords methodsFor: '*magma server-recovery' stamp: 'cmm 6/5/2005 18:24'!
unapplyUsing: aMaApplyRecord

	broker unapplyUsing: aMaApplyRecord! !

MaRequestServerConsole subclass: #MagmaServerConsole
	instanceVariableNames: 'server controller'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Magma server-API'!

!MagmaServerConsole methodsFor: 'high-availability' stamp: 'cmm 3/14/2009 12:49'!
beWarmBackupFor: primaryLocation 
	"If I am currently started on backup files created from my primary having had sent #fullBackupTo:, which means I should be in restoreMode, then establish a session to primaryLocation and bring my copy of the repository up to date.  Once up to date, the primary server will then broadcast further updates to me in real-time, and instruct clients that I am a warm backup for that primary.  If they detect the primary server is down they will start sending requests to me, at which point I will verify the server availability myself and will, if the connection is refused, #takeOverAsPrimary."
	| primarySession |
	self isRunning ifFalse: [ MagmaUserError signal: 'Must be running to be a warm backup.' ].
	primarySession := primaryLocation newAdminSession.
	
	[ primarySession remoteMagmaId = controller magmaId ifFalse: 
		[ MagmaUserError signal: 'Cannot be a warm backup of ' , primaryLocation printString , ' because I am a totally different repository!!' ].
	primarySession inRestoreMode ifTrue: 
		[ MagmaUserError signal: 'Don''t want to be a warm backup of another warm backup.' ].
	primarySession remoteCommitNumber < controller commitNumber ifTrue: 
		[ MagmaUserError signal: 'Cannot be a warm backup of ' , primaryLocation printString , ' because it is behind this repository!!' ].
	controller
		validateLastBranchCodeUsing: primarySession ;
		ensureInRestoreMode.
	primarySession disconnect.
	controller 
		forkCatch: self thisLocation
		upTo: primaryLocation ] 
		on: NetworkError
		do: 
			[ : err | 
			"Don't do anything.  This server couldn't connect to the primary, but it will sit and wait patiently for the primary to come up.  When it does, it will seek out this server and link up.  If, before then, a client requests this to be the primary, I will become so."
			 ].
	^ true! !

!MagmaServerConsole methodsFor: 'private' stamp: 'cmm 2/24/2009 15:05'!
checkIfAmPrimaryOvertaken: portInteger 
	"Check whether I was a primary overtaken by one of my backups and put myself safely into restoreMode if I was.  Even if I was only a secondary, still check the other secondary's to see if any of them is now the primary and note it if they have."
	| thisLocation |
	thisLocation := MagmaRemoteLocation 
		host: NetNameResolver localHostName
		port: portInteger.
	self node secondaryLocationsDo: 
		[ : each | 
		| sess |
		each = thisLocation ifFalse: 
			[ 
			[ sess := (self node sessionFor: each) ensureConnected.
			(sess magmaId = controller session magmaId and: [ sess inRestoreMode not ]) ifTrue: 
				[ controller ensureInRestoreMode.
				"Update the node in memory so #ensureCorrectNodeConfiguration will work correctly.."
				controller magmaId node: sess magmaId node veryDeepCopy	"(so as not to share the same object between sessions)" ] ] 
				on: NetworkError
				do: [ : err | MagmaNodeUpdate signalRemove: each ] ] ]! !

!MagmaServerConsole methodsFor: 'accessing' stamp: 'cmm 7/15/2008 22:14'!
compressThreshold
	"Compressing large responses can make a significant performance improvement on slow networks.  Compression makes things worse with small responses or on really fast networks.  Answer the number of bytes a response needs for compression to occur."
	^ server compressThreshold! !

!MagmaServerConsole methodsFor: 'accessing' stamp: 'cmm 7/15/2008 22:14'!
compressThreshold: anInteger
	"Compressing large responses can make a significant performance improvement on slow networks.  Compression makes things worse with small responses or on really fast networks.  Answer the number of bytes a response needs for compression to occur."
	server compressThreshold: anInteger! !

!MagmaServerConsole methodsFor: 'accessing' stamp: 'cmm 12/18/2002 21:52'!
controller

	^controller! !

!MagmaServerConsole methodsFor: 'accessing' stamp: 'cmm 10/14/2008 14:16'!
definition
	^ controller definition! !

!MagmaServerConsole methodsFor: 'private' stamp: 'cmm 3/16/2009 12:59'!
ensureCorrectNodeConfiguration
	"Ensure I am performing the duty my MagmaNode says I should be performing, either as the primary or one of the warmBackups.  I am read-only on the node, I do not modify it.  The intended usage of this method is, you update the node to what you want it to be, then I 'make it so'."
	self isPrimary 
		ifTrue: 
			[ "To be extra conservative, configure all other nodes to restoreMode (before coming out of restoreMode myself)."
			self sessionsForOtherLocationsDo: [ : each | each ensureInRestoreMode ].
			self restoreMode ifTrue: 
				[ "I need to become the primary, make it so."
				controller commitRestore ].
			"Now the others can become warm backups."
			self sessionsForOtherLocationsDo: [ : each | each ensureCorrectNodeConfiguration ] ]
		ifFalse: [ self node primaryLocation ifNotNilDo: [ : pl | self beWarmBackupFor: pl ] ]! !

!MagmaServerConsole methodsFor: 'accessing' stamp: 'cmm 10/14/2008 14:31'!
hostname
	^ NetNameResolver localHostName ! !

!MagmaServerConsole methodsFor: 'initialize-release' stamp: 'cmm 9/29/2008 16:11'!
initializeStatistics
	statistics := MagmaServerStatistics new! !

!MagmaServerConsole methodsFor: 'high-availability' stamp: 'cmm 3/16/2009 12:47'!
isPrimary
	^ self node primaryLocation 
		ifNil: [ false ]
		ifNotNilDo: [ : loc | loc = self thisLocation ]! !

!MagmaServerConsole methodsFor: 'start / stop' stamp: 'cmm 7/16/2008 10:20'!
isRunning
	^ server notNil and: [ server isRunning ]! !

!MagmaServerConsole methodsFor: 'high-availability' stamp: 'cmm 3/4/2009 15:54'!
isSecondary
	^ self restoreMode and: 
		[ self node notNil and: [ self node secondaryLocations includes: self thisLocation ] ]! !

!MagmaServerConsole methodsFor: 'accessing' stamp: 'cmm 10/30/2008 13:55'!
node
	^ self definition ifNotNilDo: [ : def | def node ]! !

!MagmaServerConsole methodsFor: 'private' stamp: 'cmm 12/17/2008 17:35'!
noteAboutToStart
	super noteAboutToStart.
	self preIgnitionSequence: self port! !

!MagmaServerConsole methodsFor: 'private' stamp: 'cmm 12/17/2008 17:35'!
noteJustStarted
	super noteJustStarted.
	self postIgnitionSequence ! !

!MagmaServerConsole methodsFor: 'notifications' stamp: 'cmm 9/29/2008 16:17'!
noteServerShutdown
	super noteServerShutdown.
	controller close.
	self note: controller printString , ' closed.'! !

!MagmaServerConsole methodsFor: 'start / stop' stamp: 'cmm 9/29/2008 23:12'!
open: pathOrMagmaLocalLocation 
	"As indicated, pathOrMagmaLocalLocation may be either the String path (relative or fully-qualified) or a MagmaLocalLocation."
	controller 
		ifNil: 
			[ controller := (MagmaRepositoryController open: pathOrMagmaLocalLocation asMagmaLocalLocation)
				serverStatistics: statistics ;
				yourself ]
		ifNotNil: 
			[ "Reopen existing controller rather than replacing it with a new construction so prior preferences and statistics are preserved."
			controller open: pathOrMagmaLocalLocation asMagmaLocalLocation ]! !

!MagmaServerConsole methodsFor: 'accessing' stamp: 'cmm 10/14/2008 14:31'!
port
	^ server ifNotNil: [ server port ]! !

!MagmaServerConsole methodsFor: 'private' stamp: 'cmm 12/17/2008 17:34'!
postIgnitionSequence
	self ensureCorrectNodeConfiguration! !

!MagmaServerConsole methodsFor: 'private' stamp: 'cmm 12/30/2008 18:45'!
preIgnitionSequence: portInteger 
	| nodeUpdates |
	nodeUpdates := OrderedCollection new.
	[ self checkIfAmPrimaryOvertaken: portInteger ] 
		on: MagmaNodeUpdate
		do: 
			[ : update | 
			nodeUpdates add: update.
			update resume ].
	nodeUpdates ifNotEmpty: 
		[ self serverSession commit: [ nodeUpdates do: [ : each | each applyTo: self node ] ] ].
	self updatePrimaryLocation: portInteger.
	self statistics transition! !

!MagmaServerConsole methodsFor: 'start / stop' stamp: 'cmm 1/1/2009 17:22'!
processOn: portInteger 
	"Start the Magma server listening on port portInteger."
	(controller isNil or: [ controller isOpen not ]) ifTrue: 
		[ MagmaUserError signal: 'The repository must be opened before the server can be started.' ].
	server := (MaNetworkRequestServer protocol: MagmaSession protocol) console: self.
	server 
		processOn: portInteger
		using: self! !

!MagmaServerConsole methodsFor: 'high-availability' stamp: 'cmm 3/4/2009 15:45'!
removeSecondaryLocation: aMagmaRemoteLocation 
	MagmaNodeUpdate new
		remove: aMagmaRemoteLocation ;
		processUsing: self! !

!MagmaServerConsole methodsFor: 'high-availability' stamp: 'cmm 10/27/2008 09:52'!
restoreMode
	"Answer whether I am running in restoreMode.  restoreMode only allows commits from a primary to be applied in sequence."
	^ controller restoreMode! !

!MagmaServerConsole methodsFor: 'accessing' stamp: 'cmm 3/3/2005 10:10'!
server

	^ server! !

!MagmaServerConsole methodsFor: 'accessing' stamp: 'cmm 11/15/2008 16:03'!
serverSession
	^ controller session ! !

!MagmaServerConsole methodsFor: 'private' stamp: 'cmm 11/13/2008 20:57'!
sessionsForOtherLocationsDo: oneArgBlock 
	"For the other locations in the Node other than thisLocation, value oneArgBlock."
	self node locationsDo: 
		[ : each | 
		each = self thisLocation ifFalse: [ oneArgBlock value: (self node sessionFor: each) ] ]! !

!MagmaServerConsole methodsFor: 'start / stop' stamp: 'cmm 3/4/2009 15:53'!
shutdown
	self isSecondary ifTrue: 
		[ self node primarySession ifNotNilDo: [ : sess | sess removeWarmBackup: self thisLocation ] ].
	server ifNotNil: [ server shutdown ].
	self node ifNotNilDo: [ : node | node sessions do: [ : each | each disconnect ] ].
	self statistics ifNotNilDo: [ : stat | stat release ]! !

!MagmaServerConsole methodsFor: 'high-availability' stamp: 'cmm 10/14/2008 17:35'!
thisLocation
	"Answer the MagmaRemoteLocation in the node that represents this running server."
	^ MagmaRemoteLocation 
		host: NetNameResolver localHostName
		port: self port! !

!MagmaServerConsole methodsFor: 'private' stamp: 'cmm 3/12/2009 15:43'!
updatePrimaryLocation: portInteger 
	"When I'm started up, update the primaryLocation to this location, if it has changed."
	self restoreMode ifFalse: 
		[ "We can't simply use #thisLocation because we aren't running yet (so no port)!!"
		| primaryLoc thisLoc |
		thisLoc := MagmaRemoteLocation 
			host: NetNameResolver localHostName
			port: portInteger.
		primaryLoc := self node primaryLocation.
		(primaryLoc isNil or: [ primaryLoc isLocalHost not or: [ self node primaryLocation ~= thisLoc ] ]) ifTrue: [ controller session commit: [ self node primaryLocation: thisLoc ] ] ]! !

!MagmaServerConsole methodsFor: 'accessing' stamp: 'cmm 11/14/2008 10:31'!
useCompression: aBoolean 
	self compressThreshold: 
		(aBoolean 
			ifTrue: [ 500 "bytes" ]
			ifFalse: [ MaObjectBuffer unreasonablePhysicalSize*2 ])! !

!MagmaServerConsole methodsFor: 'private' stamp: 'cmm 3/7/2009 11:14'!
value: aMagmaRepositoryRequest 
	"Make the console available too for network-serving repositories."
	^ 
	[ aMagmaRepositoryRequest
		console: self ;
		nodeCheck.
	controller value: aMagmaRepositoryRequest ] 
		on: Error
		do: 
			[ : anError | 
			self maMarked: 'dev'.	"This error routine must be improved.  It needs to log and notify for unhandled errors!!"
			(anError maOriginalClass = MagmaLowOnMemory and: [ SmalltalkImage current maIsHeadless ]) ifTrue: 
				[ controller
					serverSave: false
					andExit: true ].
			(MagmaPreferences debug and: 
				[ ({ 
					MagmaCommitConflictError.
					MagmaSessionLost.
					MagmaWrongServerError.
					MagmaWrongCommitNumber
				 } includes: anError maOriginalClass) not and: 
					[ anError isMaUserError not and: [ SmalltalkImage current maIsHeadless not ] ] ]) ifTrue: [ self halt: anError messageText ].
			anError ]! !

!MaDownloadCodeRequest methodsFor: '*magma server' stamp: 'cmm 11/16/2008 19:24'!
nodeCheck
	self shouldBeSecondary! !

!MaDownloadCodeRequest methodsFor: '*magma server' stamp: 'cmm 10/26/2004 21:54'!
process

	^ repositoryController codeBaseByteArray! !

!MaRepositoryConnectionRequest methodsFor: '*magma server' stamp: 'cmm 12/18/2008 17:16'!
process
	MagmaSession magmaVersion = magmaVersion ifFalse: 
		[ MagmaUserError signal: 'This server is running Magma version ' , MagmaSession magmaVersion printString , '.  You may not connect with version ' , magmaVersion printString , ' of the client code.  You may be able to install the code using browseCode.' ].
	uuid ifNotNil: 
		[ repositoryController magmaId uuid = uuid ifFalse: 
			[ MagmaUserError signal: 'Cannot connect because the repository has been replaced.  Please connect a new session.' ] ].
	commitNumber > 0 ifTrue: 
		[ repositoryController commitNumber < commitNumber ifTrue: 
			[ MagmaWrongCommitNumber signal: 'Cannot connect this session because the repository has been restored to an earlier state.  Please connect a new session.' ] ].
	^ repositoryController 
		newSessionFor: self
		sessionId: sessionId! !

!MaRepositoryConnectionRequest methodsFor: '*magma server' stamp: 'cmm 9/21/2008 15:24'!
userId
	^ userId! !

!MaOidCalculator class methodsFor: '*magma server' stamp: 'cmm 6/8/2005 17:07'!
indexPositionForOid: anInteger

	^ anInteger - self firstUserObjectOid + 1! !

!MaOidCalculator class methodsFor: '*magma server' stamp: 'cmm 8/13/2002 23:57'!
isOidForOptimizedObject: oidInteger
	"Answer whether oidInteger is an oid for anObject which will never require its own buffer.  These include optimized Integers, nil, true, false, and all Characters."

	^(self isOidForNewObject: oidInteger) not
		and: [ (self isOidForUserObject: oidInteger) not ]! !

!MaOidCalculator class methodsFor: '*magma server' stamp: 'cmm 11/13/2002 23:05'!
isOidForUserObject: anInteger
	"A 'user' object is a persistent object stored in a Magma database that does not have a special oid, such as SmallIntegers, Floats, Booleans, etc. have."

	^anInteger
		between: self firstUserObjectOid
		and: self lastUserObjectOid! !

!MaOidCalculator class methodsFor: '*magma server' stamp: 'cmm 4/21/2005 15:41'!
lastUserObjectOid

	^ 281468533259712   "self oidForSmallestFloat - 1"! !

!MaOidCalculator class methodsFor: '*magma server' stamp: 'cmm 12/2/2004 13:32'!
oidForFilePositionIndexPosition: anInteger

	^ self firstUserObjectOid + anInteger! !

!MaFileRecordBroker methodsFor: '*magma server' stamp: 'cmm 10/22/2006 15:02'!
filePool: aMaFileStreamPool 
	file maIsTransactional ifTrue: [ file pool: aMaFileStreamPool ]! !

!MaFileRecordBroker methodsFor: '*magma server-recovery' stamp: 'cmm 7/6/2005 22:52'!
filesDo: oneArgBlock

	oneArgBlock value: file! !

!MaFileRecordBroker methodsFor: '*magma server-recovery' stamp: 'cmm 4/10/2007 22:35'!
unapplyUsing: aMaApplyRecord 
	aMaApplyRecord unapplySelf: file! !

!MaWriteRequest methodsFor: '*magma server' stamp: 'cmm 3/4/2009 09:17'!
process
	| commitLogEntry |
	self restoreMode ifFalse: 
		[ " MagmaEnvironmentError signal: 'Only repositories in restoreMode are updated in this way.' "
		"Not signaling an error because, after a #takeOverAsPrimary, there will be unavoidable extranneous WriteRequests left-over in the queue.  It's easier just to ignore them for the test cases.  Also, WriteRequests are never checked for a response anyway, the sender of a WriteREquest would never know there was an error."
		^ nil ].
	commitLogEntry := self recoveryManager materializeCommitLogEntryIn: commitLogRecord.
	
	[ self repository 
		write: commitLogEntry
		logged: true
		commitLogRecord: commitLogRecord.
	self session advanceWith: commitLogEntry ] 
		on: MagmaWrongCommitNumber
		do: 
			[ : err | 
			"console ensureCorrectNodeConfiguration"
			"<-- probably overkill"
			err expectedCommitNumber < commitLogRecord commitNumber 
				ifTrue: 
					[ "Gap fill"
					repositoryController 
						catch: console thisLocation
						upTo: repositoryController node primaryLocation ]
				ifFalse: 
					[ MagmaNotification signal: ('Received commitNumber ' , commitLogRecord commitNumber asString , ' but need ' , err expectedCommitNumber asString) ] ]! !

!MaFixedObjectBuffer methodsFor: '*magma server' stamp: 'cmm 7/20/2007 11:16'!
canConflict
	^ self classId ~= MagmaClassIdManager magmaCounterClassId! !

!MaFixedObjectBuffer methodsFor: '*magma server' stamp: 'cmm 3/8/2007 12:58'!
graphDo: oneArgBlock using: aMaReadStrategy for: aMagmaFileTraverser
	aMagmaFileTraverser
		graphFromPointers: self
		do: oneArgBlock
		using: aMaReadStrategy.
	(aMagmaFileTraverser isForLargeCollection: self)
		ifTrue:
			[ aMagmaFileTraverser
				graphFromLargeCollection: self oid
				do: oneArgBlock
				using: aMaReadStrategy ]! !

!MaReadStrategy methodsFor: '*magma server' stamp: 'cmm 1/26/2005 12:22'!
depthOfAttribute: attributeIndex onClassWithId: classId version: versionNumber
	"Answer the depth that the attribute at attributeIndex should be traversed into *from the current depth*."

	^depths
		maAt: { classId. versionNumber }
		ifPresent:
			[ :depth |
			depth isInteger
				ifTrue: [ depth ]
				ifFalse: [ depth at: attributeIndex ifAbsent: [ 0 ] ] ]
		ifAbsent: [ 0 ]! !

!MagmaCollectionChanges methodsFor: '*magma server' stamp: 'cmm 11/6/2006 22:39'!
requiresServerRefresh
	^ self hasAddedIndexes or: [ self hasRemovedIndexes ]! !

!MagmaLargeCollectionSegment methodsFor: '*magma server-private' stamp: 'cmm 2/18/2008 21:33'!
collectBuffers: sessionId from: aMaObjectRepository using: aMaReadStrategy 
	| readStrategy |
	readStrategy := aMaObjectRepository 
		ensureReadStrategy: aMaReadStrategy
		forSession: sessionId.
	self objects withIndexDo: 
		[ : eachHashAndOid : n | 
		eachHashAndOid isVariableBinding 
			ifTrue: 
				[ eachHashAndOid value: (aMaObjectRepository 
						byteArrayAt: eachHashAndOid value
						using: readStrategy) ]
			ifFalse: 
				[ "For member index accesses, there are no extra hash keys, so don't waste the bandwidth."
				self objects 
					at: n
					put: 
						(aMaObjectRepository 
							byteArrayAt: eachHashAndOid "really just the oid"
							using: readStrategy) ] ]! !

!MagmaServerStatisticsRequest methodsFor: '*magma server' stamp: 'cmm 5/7/2008 17:22'!
process
	^ repositoryController serverStatistics! !

!MaBasicReadStrategy methodsFor: '*magma server' stamp: 'cmm 1/26/2005 17:28'!
depthOfAttribute: attributeIndex onClassWithId: classId version: versionNumber
	"Answer the delta-depth from the current depth to read.  This is distinct from my minimumDepth, which was used at the start of this graph-read."

	^ 0! !

!MaReaderTrunkRequest methodsFor: '*magma server' stamp: 'cmm 11/17/2008 09:32'!
nodeCheck
	self shouldBeSecondary! !

!MaReaderTrunkRequest methodsFor: '*magma server' stamp: 'cmm 8/6/2006 18:29'!
process
	^ self repository 
		getTrunkFor: collectionOid
		expression: expression! !

!MaPrimitiveAttributesRequest methodsFor: '*magma server' stamp: 'cmm 12/22/2008 13:36'!
process
	^ self repository primitiveAttributes
		at: #oidCount
			put: repositoryController oidCount ;
		yourself! !

!MaRecordBroker methodsFor: '*magma server' stamp: 'cmm 10/22/2006 14:58'!
filePool: aMaFileStreamPool
	"only applicable to MaFileRecordBrokers"! !

!MaRecordBroker methodsFor: '*magma server-recovery' stamp: 'cmm 7/5/2005 22:29'!
filesDo: aBlock
	"nothing to do by default."! !

!MaLoadFromReaderRequest methodsFor: '*magma server' stamp: 'cmm 11/16/2008 23:31'!
nodeCheck
	self shouldBePrimary! !

!MaLoadFromReaderRequest methodsFor: '*magma server' stamp: 'cmm 10/16/2006 21:01'!
process
	^ self repository 
		registerAndLoad: commitPackage 
		using: expression
		from: sourceCollectionOid
		forSession: sessionId
		distinct: distinct! !

!MaCommitLogRecordRequest methodsFor: '*magma server' stamp: 'cmm 3/15/2009 19:37'!
process
	^ endingCommitNumber 
		ifNil: [ self repository commitRecordsFrom: startingCommitNumber ]
		ifNotNil: 
			[ self repository 
				commitRecordsFrom: startingCommitNumber
				to: endingCommitNumber ]! !

!MagmaLocalLocation methodsFor: '*magma server' stamp: 'cmm 7/1/2007 13:52'!
commitLogDirectory
	^ FileDirectory on: (self directory fullNameFor: MaRecoveryManager commitLogDirectoryName)! !

!MagmaLocalLocation methodsFor: '*magma server' stamp: 'cmm 3/6/2009 14:39'!
delete
	"Delete the files at my location."
	self isRepository 
		ifTrue: 
			[ MagmaDeleteWarning signal: 'The Magma database at ' , self pathName , ' is about to be deleted.'.
			self directory recursiveDelete ]
		ifFalse: 
			[ MagmaNotification signal: 'There is no Magma repository at ' , self pathName ]! !

!MagmaLocalLocation methodsFor: '*magma server' stamp: 'cmm 4/20/2007 16:31'!
isRepository
	^ self directory exists and: [ self directory fileExists: MaObjectFiler objectsPhysicalFilename ]! !

!MagmaLocalLocation methodsFor: '*magma server' stamp: 'cmm 11/5/2008 21:36'!
newServerLink
	^ MagmaSession newLocalLinkFor: (MagmaRepositoryController openedOn: self)! !

!MagmaLocalLocation methodsFor: '*magma server' stamp: 'cmm 3/10/2009 19:59'!
seedDirectory
	"The initial copy of the repository, to which commit.n.log files may be loaded to advance it forward."
	^ FileDirectory on: (self directory fullNameFor: 'seed')! !

!MaSaveAndExitImageRequest methodsFor: '*magma server' stamp: 'cmm 3/12/2009 22:03'!
process
	"Just like selecting it from the World menu.  shutdown process has responsibility to do it cleanly."
	repositoryController 
		serverSave: save
		andExit: exit.
	^ true! !

!MaReadRequest methodsFor: '*magma server' stamp: 'cmm 5/14/2005 18:11'!
process

	^ self repository
		read: oid
		for: sessionId
		using: readStrategy! !

MaRecoveryRecord subclass: #MaApplyRecord
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Magma server-Recovery'!

!MaApplyRecord commentStamp: 'cmm 5/27/2007 20:23' prior: 0!
I am an abstract class for the types of records that, on recovery, require rollback of some particular file.  My fileId identifies which file it is.

	... see superclass ...
	4 - groupId (3 bytes), a number identifying a group of writes
	... subclasses may add more fields ...
!

!MaApplyRecord class methodsFor: 'as yet unclassified' stamp: 'cmm 6/3/2005 17:16'!
filePositionsFileId
	^ 2! !

!MaApplyRecord class methodsFor: 'as yet unclassified' stamp: 'cmm 6/3/2005 17:16'!
objectsFileId
	^ 1! !

!MaApplyRecord methodsFor: 'accessing' stamp: 'cmm 5/27/2007 20:12'!
bufferPositionOfGroupId
	"this fields is 3-bytes."
	^ 4! !

!MaApplyRecord methodsFor: 'accessing' stamp: 'cmm 5/27/2007 20:14'!
groupId
	^ byteArray 
		maUint: 24
		at: self bufferPositionOfGroupId! !

!MaApplyRecord methodsFor: 'accessing' stamp: 'cmm 5/27/2007 20:14'!
groupId: anInteger 
	byteArray 
		maUint: 24
		at: self bufferPositionOfGroupId
		put: anInteger! !

!MaApplyRecord methodsFor: 'accessing' stamp: 'cmm 5/27/2007 20:24'!
headerSize
	"super headerSize + size of groupId field"
	^ 7! !

!MaApplyRecord methodsFor: 'printing' stamp: 'cmm 5/27/2007 20:32'!
maPrintAbbreviatedOn: aStream
	super maPrintAbbreviatedOn: aStream.
	self
		maPrintAttributes: #(groupId)
		on: aStream! !

MaApplyRecord subclass: #MaApplyToFileRecord
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Magma server-Recovery'!

!MaApplyToFileRecord commentStamp: 'cmm 5/27/2007 20:23' prior: 0!
I am an abstract class for the types of records that, on recovery, require rollback of some particular file.  My fileId identifies which file it is.

	... see superclass ...
	7 - file-id (8-bytes), identifies which file this record describes
	... subclasses may add more fields ...
!

!MaApplyToFileRecord methodsFor: 'accessing' stamp: 'cmm 5/27/2007 20:23'!
bufferPositionOfFileId
	"This field is 8-bytes to make it easy for Magma.  6-bytes for the oid, 2-bytes for the index #."
	^ 7! !

!MaApplyToFileRecord methodsFor: 'accessing' stamp: 'cmm 8/10/2007 00:23'!
collectionOid
	"for Magma, the fileId is stored in two parts, the high-order 6 bytes represent the oid of the collection.  The low-order 2 bytes represent the index # for that collection or 0 if it is the memberIndex.  This method answers the oid."

	^ self fileId bitShift: -16! !

!MaApplyToFileRecord methodsFor: 'accessing' stamp: 'cmm 5/27/2007 20:23'!
fileId

	^ byteArray
		maUint: 64
		at: self bufferPositionOfFileId! !

!MaApplyToFileRecord methodsFor: 'accessing' stamp: 'cmm 5/27/2007 20:23'!
fileId: anInteger

	byteArray
		maUint: 64
		at: self bufferPositionOfFileId 
		put: anInteger! !

!MaApplyToFileRecord methodsFor: 'accessing' stamp: 'cmm 5/27/2007 20:25'!
headerSize
	"super headerSize + size of fileId field"
	^ 15! !

!MaApplyToFileRecord methodsFor: 'as yet unclassified' stamp: 'cmm 5/27/2007 20:54'!
isApplyToFile
	^ true! !

!MaApplyToFileRecord methodsFor: 'testing' stamp: 'cmm 8/10/2007 00:23'!
isForFilePositionsFile
	^ self fileId = self class filePositionsFileId! !

!MaApplyToFileRecord methodsFor: 'testing' stamp: 'cmm 8/10/2007 00:23'!
isForObjectsFile
	^ self fileId = self class objectsFileId! !

!MaApplyToFileRecord methodsFor: 'printing' stamp: 'cmm 5/27/2007 20:32'!
maPrintAbbreviatedOn: aStream
	super maPrintAbbreviatedOn: aStream.
	self
		maPrintAttributes: #(fileId)
		on: aStream! !

!MaApplyToFileRecord methodsFor: 'as yet unclassified' stamp: 'cmm 8/14/2007 22:31'!
unapplySelf: aStandardFileStream 
	self subclassResponsibility! !

MaApplyToFileRecord subclass: #MaBeforeImageRecord
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Magma server-Recovery'!

!MaBeforeImageRecord commentStamp: 'cmm 5/27/2007 20:16' prior: 0!
This type of RecoveryRecord is used to restore any bytes in any file to their state prior to the beginning of the interrupted commit.

My #fileId identifies the file that was updated.

My #location provides information necessary to know *where* in the file I came from.

	... see superclass ...
	15 - file-position
	21 - the record
	... see superclass ...

I should not be created for updates to new files or for writes beyond the end of the file.!

!MaBeforeImageRecord class methodsFor: 'accessing' stamp: 'cmm 6/5/2005 16:38'!
assignedRecordType
	^ 2! !

!MaBeforeImageRecord methodsFor: 'debugging' stamp: 'cmm 6/2/2005 13:05'!
asMaObjectBuffer

	self fileId = 1 ifFalse: [ self error: 'not a before-image of a MaObjectBuffer.' ].
	^ (MaObjectBuffer
		byteArray: (ByteArray new: MaObjectBuffer oidSize/8), self record
		startPos: 0) asSentientBuffer! !

!MaBeforeImageRecord methodsFor: 'initializing' stamp: 'cmm 6/3/2005 16:18'!
bodySize
	^ 1000  "variable length"! !

!MaBeforeImageRecord methodsFor: 'private' stamp: 'cmm 5/27/2007 20:16'!
bufferPositionOfLocation
	"this field is 6 bytes"
	^ 15! !

!MaBeforeImageRecord methodsFor: 'private' stamp: 'cmm 5/27/2007 20:26'!
headerSize
	"super headerSize + size of location field."
	^ 21! !

!MaBeforeImageRecord methodsFor: 'testing' stamp: 'cmm 6/3/2005 15:53'!
isFixed
	^ false! !

!MaBeforeImageRecord methodsFor: 'accessing' stamp: 'cmm 6/1/2005 13:18'!
location

	^ byteArray
		maUint: 48
		at: self bufferPositionOfLocation! !

!MaBeforeImageRecord methodsFor: 'initializing' stamp: 'cmm 6/1/2005 13:18'!
location: anInteger

	byteArray
		maUint: 48
		at: self bufferPositionOfLocation
		put: anInteger! !

!MaBeforeImageRecord methodsFor: 'printing' stamp: 'cmm 6/5/2005 17:09'!
maPrintAbbreviatedOn: aStream
	super maPrintAbbreviatedOn: aStream.
	self
		maPrintAttributes: #(location recordSize)
		on: aStream! !

!MaBeforeImageRecord methodsFor: 'recovery' stamp: 'cmm 4/10/2007 22:36'!
unapplySelf: aMaTransactionalFileStream 
	aMaTransactionalFileStream
		maWrite: self recordSize
			bytesFromPosition: 1
			of: self record
			atFilePosition: self location ;
		maCommit! !

MaApplyToFileRecord subclass: #MaFileGrowthRecord
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Magma server-Recovery'!

!MaFileGrowthRecord commentStamp: 'cmm 5/25/2007 00:20' prior: 0!
I am used to track files that grow in size while applying a commit.  I record the #originalSize such that, during recovery, the file can be truncated: back to its original size.

I should not be created for new files.

	... see superclass ...
	10 - originalSize (6 bytes)
	... see superclass ...!

!MaFileGrowthRecord class methodsFor: 'accessing' stamp: 'cmm 6/5/2005 16:39'!
assignedRecordType
	^ 3! !

!MaFileGrowthRecord methodsFor: 'private' stamp: 'cmm 6/3/2005 16:00'!
bodySize  "in bytes"
	^ self sizeOfOriginalSizeField / 8! !

!MaFileGrowthRecord methodsFor: 'printing' stamp: 'cmm 6/5/2005 17:09'!
maPrintAbbreviatedOn: aStream
	super maPrintAbbreviatedOn: aStream.
	self
		maPrintAttributes: #(originalSize)
		on: aStream! !

!MaFileGrowthRecord methodsFor: 'access' stamp: 'cmm 6/3/2005 16:01'!
originalSize
	^ byteArray
		maUint: self sizeOfOriginalSizeField 
		at: self headerSize! !

!MaFileGrowthRecord methodsFor: 'access' stamp: 'cmm 6/3/2005 16:01'!
originalSize: anInteger
	^ byteArray
		maUint: self sizeOfOriginalSizeField 
		at: self headerSize
		put: anInteger! !

!MaFileGrowthRecord methodsFor: 'private' stamp: 'cmm 6/3/2005 15:59'!
sizeOfOriginalSizeField

	^ 48 "bits"! !

!MaFileGrowthRecord methodsFor: 'recovery' stamp: 'cmm 4/28/2007 14:30'!
unapplySelf: aMaTransactionalFileStream 
	aMaTransactionalFileStream truncate: self originalSize! !

MaApplyToFileRecord subclass: #MaNewFileRecord
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Magma server-Recovery'!

!MaNewFileRecord commentStamp: 'cmm 5/25/2007 00:20' prior: 0!
Sometimes new files are created during a commit.  If a disaster happens before that commit finishes then, upon recovery, those files should be deleted.

I record the #filename of the file that should be deleted.

	... see superclass ...
	10 - filename (variable length)
	... see superclass ...!

!MaNewFileRecord class methodsFor: 'accessing' stamp: 'cmm 6/5/2005 16:39'!
assignedRecordType
	^ 4! !

!MaNewFileRecord methodsFor: 'accessing' stamp: 'cmm 6/3/2005 16:18'!
bodySize

	^ 300  "variable length"! !

!MaNewFileRecord methodsFor: 'accessing' stamp: 'cmm 9/21/2005 09:54'!
filename
	^ (byteArray
		copyFrom: self headerSize+1
		to: self checkSumPosition-1) asString! !

!MaNewFileRecord methodsFor: 'accessing' stamp: 'cmm 7/4/2005 22:54'!
filename: aString

	self record: aString! !

!MaNewFileRecord methodsFor: 'accessing' stamp: 'cmm 6/3/2005 15:53'!
isFixed
	^ false! !

!MaNewFileRecord methodsFor: 'printing' stamp: 'cmm 6/5/2005 17:10'!
maPrintAbbreviatedOn: aStream
	super maPrintAbbreviatedOn: aStream.
	self
		maPrintAttributes: #(filename)
		on: aStream! !

!MaNewFileRecord methodsFor: 'recovery' stamp: 'cmm 6/4/2007 08:51'!
unapplySelf: anUndefinedObject 
	(MaAtomicFileStream fileNamed: self filename) deletePhysicalFiles! !

MaApplyRecord subclass: #MaBeginApplyRecord
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Magma server-Recovery'!

!MaBeginApplyRecord commentStamp: 'cmm 6/3/2005 15:15' prior: 0!
I am the first record of a group of RecoveryRecords for the same transactionId.  If a disaster happens in the middle of a commit and, during recovery, a valid BeginApply record is found in the apply file *without* a valid EndApply record, then we know the commit did not finish and all of the records must be unapplied.!

!MaBeginApplyRecord class methodsFor: 'accessing' stamp: 'cmm 6/5/2005 16:39'!
assignedRecordType
	^ 1! !

!MaBeginApplyRecord methodsFor: 'testing' stamp: 'cmm 6/3/2005 16:57'!
isBeginApply
	^ true! !

!MaBeginApplyRecord methodsFor: 'printing' stamp: 'cmm 6/5/2005 17:10'!
maPrintAbbreviatedOn: aStream
	super maPrintAbbreviatedOn: aStream.
	self
		maPrintAttributes: #(isBeginApply)
		on: aStream! !

MaApplyRecord subclass: #MaEndApplyRecord
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Magma server-Recovery'!

!MaEndApplyRecord commentStamp: 'cmm 3/25/2007 12:36' prior: 0!
I am the last record of a group of RecoveryRecords for the same transactionId.  If a disaster happens then, during recovery, I am looked for in the applyFile with the same transactionId as my BeginApply counterpart.  If I'm found, then no were no commits happening at the time of the failure and, thus, no backward recovery is necessary.!

!MaEndApplyRecord class methodsFor: 'accessing' stamp: 'cmm 6/5/2005 16:39'!
assignedRecordType
	^ 5! !

!MaEndApplyRecord methodsFor: 'testing' stamp: 'cmm 6/3/2005 16:57'!
isEndApply
	^ true! !

!MaEndApplyRecord methodsFor: 'printing' stamp: 'cmm 6/5/2005 17:10'!
maPrintAbbreviatedOn: aStream
	super maPrintAbbreviatedOn: aStream.
	self
		maPrintAttributes: #(isEndApply)
		on: aStream! !

MagmaWarning subclass: #MagmaDeleteWarning
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Magma server-Exceptions'!

!MagmaDeleteWarning commentStamp: 'cmm 7/16/2007 22:07' prior: 0!
A MagmaDeleteWarning is signaled when a repository is about to be deleted.!

MagmaWarning subclass: #MagmaForwardRecoveryNeeded
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Magma server-Recovery'!

!MagmaForwardRecoveryNeeded methodsFor: 'as yet unclassified' stamp: 'cmm 3/2/2009 19:53'!
defaultAction
	^ self resume: true! !

MagmaWarning subclass: #MagmaMissingCommitLogWarning
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Magma server-Exceptions'!

MagmaWarning subclass: #MagmaRollbackRecoveryNeeded
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Magma server-Recovery'!

!MagmaSession class methodsFor: '*magma server-private' stamp: 'cmm 2/8/2005 21:44'!
newLocalLinkFor: aMagmaRepositoryController

	^ MaLocalServerLink localRequestServer: (MaLocalRequestServer processor: aMagmaRepositoryController)! !

!MagmaSession class methodsFor: '*magma server' stamp: 'cmm 3/25/2009 11:30'!
openLocal: pathStringOrMagmaLocalLocation 
	^ self repositoryController: (MagmaRepositoryController openedOn: pathStringOrMagmaLocalLocation asMagmaLocalLocation) softBranch! !

!MagmaSession class methodsFor: '*magma server' stamp: 'cmm 12/26/2004 23:21'!
repositoryController: aMagmaRepositoryController
	"Use this constructor only for *additional* local connections, i.e., two or more sessions connected to the same local repository.  Use the same MagmaRepositoryController you used to create the original."

	^ self link: (self newLocalLinkFor: aMagmaRepositoryController)! !

!MagmaSession methodsFor: '*magma server' stamp: 'cmm 2/10/2009 10:56'!
ensureOpen
	self repositoryController ifNil: [ MagmaEnvironmentError signal: 'local repository not open.' ].
	self isLocal ifFalse: [ MagmaUserError signal: 'Only available for local connections.' ].
	self isOpen ifFalse: [ self repositoryController ensureOpen ]! !

!MagmaSession methodsFor: '*magma server' stamp: 'cmm 1/16/2005 19:02'!
initializeDefinition

	definition _ MagmaRepositoryDefinition new! !

!MagmaSession methodsFor: '*magma server' stamp: 'cmm 1/13/2005 18:01'!
isOpen
	^ self isRemote 
	or: [ "local" self repositoryController isOpen ]! !

!MagmaSession methodsFor: '*magma server' stamp: 'cmm 11/5/2008 10:09'!
repositoryController
	^ primaryLink localRequestServer processor! !

!MaReadMultipleRequest methodsFor: '*magma server' stamp: 'cmm 2/11/2008 17:23'!
process
	^ self repository 
		refreshAll: oids
		for: sessionId! !

!MaLargeCollectionSizeRequest methodsFor: '*magma server' stamp: 'cmm 11/16/2008 23:30'!
nodeCheck
	self shouldBeSecondary! !

!MaLargeCollectionSizeRequest methodsFor: '*magma server' stamp: 'cmm 5/14/2005 18:29'!
process

	^ self repository sizeOfLargeCollection: collectionOid! !

!MaAbortTransactionRequest methodsFor: '*magma server' stamp: 'cmm 11/15/2008 14:07'!
nodeCheck
	self shouldBePrimary! !

!MaAbortTransactionRequest methodsFor: '*magma server' stamp: 'cmm 5/14/2005 18:07'!
process

	^ self repository abortTransactionFor: sessionId! !

!MaAbstractReadRequest methodsFor: '*magma server' stamp: 'cmm 3/8/2009 15:10'!
nodeCheck
	requiredProgress isInteger ifFalse: 
		[ MagmaSoftwareError signal: 'requiredProgress must be an Integer, not a ' , requiredProgress class name ].
	repositoryController commitNumber < requiredProgress ifTrue: [ MagmaReadServerTooFarBehind signal ].
	self shouldBeSecondary! !

!MagmaRepositoryRequest methodsFor: '*magma server' stamp: 'cmm 11/13/2008 20:38'!
console
	^ console! !

!MagmaRepositoryRequest methodsFor: '*magma server' stamp: 'cmm 11/14/2008 13:27'!
console: aMagmaServerConsole 
	console := aMagmaServerConsole.
	repositoryController := aMagmaServerConsole controller! !

!MagmaRepositoryRequest methodsFor: '*magma server' stamp: 'cmm 12/30/2008 12:27'!
isPrimary
	"Answer whether I thisLocation is the primary server of a multi-server node.  If a single-server, answer false."
	^ self restoreMode not and: [ self node isBackedUp ]! !

!MagmaRepositoryRequest methodsFor: '*magma server' stamp: 'cmm 3/4/2009 15:31'!
isSecondary
	^ console isSecondary! !

!MagmaRepositoryRequest methodsFor: '*magma server' stamp: 'cmm 11/14/2008 13:32'!
node
	^ repositoryController node! !

!MagmaRepositoryRequest methodsFor: '*magma server' stamp: 'cmm 11/14/2008 13:24'!
nodeCheck
	"If I am a read-request sent to a primary, or a commit-request sent to a secondary, it's a red-flag to check on node-availability.  Check the node it *should* be going to.  If it is truly unavailable then update my MagmaNode accordingly and then #ensureCorrectNodeConfiguration.  If it is still available, then perhaps it is a client-issue, signal a MagmaWrongServerError to cause the client to reset its own configuration."

	"Optional subclass override."
	! !

!MagmaRepositoryRequest methodsFor: '*magma server' stamp: ''!
process
	"Any persistent objects returned by this should be materialized only by the sessions
	serializer, not the repositoryLink's serializer because the sessions serializer is what
	works with the transaction to mark it read so changes can be detected."

	self subclassResponsibility
! !

!MagmaRepositoryRequest methodsFor: '*magma server' stamp: 'cmm 10/28/2008 11:28'!
recoveryManager
	^ self repository recoveryManager ! !

!MagmaRepositoryRequest methodsFor: '*magma server' stamp: 'cmm 5/14/2005 18:07'!
repository

	^ repositoryController repository! !

!MagmaRepositoryRequest methodsFor: '*magma server' stamp: 'cmm 7/5/2004 11:57'!
repositoryController: aMagmaRepositoryController

	repositoryController _ aMagmaRepositoryController
! !

!MagmaRepositoryRequest methodsFor: '*magma server' stamp: 'cmm 11/14/2008 13:33'!
restoreMode
	^ repositoryController restoreMode! !

!MagmaRepositoryRequest methodsFor: '*magma server' stamp: 'cmm 11/14/2008 13:30'!
session
	^ repositoryController session! !

!MagmaRepositoryRequest methodsFor: '*magma server' stamp: 'cmm 3/6/2009 15:48'!
shouldBePrimary
	"This request should only run on the primary server.  If I am in restoreMode (not the primary) then ping the primary to see whether it is still there.  If it isn't, take over, otherwise signal a MagmaWrongServerError back to the client."
	self isSecondary ifTrue: 
		[ 
		[ self node pingPrimary.
		MagmaWrongServerError new
			correctServer: self node primaryLocation ;
			signal ] 
			on: NetworkError
			do: 
				[ : netErr | 
				console ifNotNil: 
					[ "Wait briefly for any further MaWriteRequests to arrive over the network, THEY must be processed before the NodeUpdate if they are present."
					
					[ | writeRequests serverSocket |
					(Delay forMilliseconds: 1800) wait.
					writeRequests := OrderedCollection new.
					serverSocket := self serverSocket.
					serverSocket requestQueue flushAllSuchThat: 
						[ : eachLifecycle | 
						| requestObject |
						requestObject := console server materializeRequest: eachLifecycle requestByteArray.
						requestObject isWriteRequest 
							ifTrue: 
								[ MagmaNotification signal: 'Found a WriteRequest in teh queue!!'.
								eachLifecycle requestObject: requestObject.
								serverSocket dequeueRequest: eachLifecycle.
								writeRequests add: eachLifecycle.
								true ]
							ifFalse: [ false ] ].
					writeRequests := writeRequests asSortedCollection: 
						[ : a : b | 
						a commitLogRecord commitNumber < b commitLogRecord commitNumber ].
					writeRequests do: [ : each | serverSocket processRequest: each ] ] 
						on: Error
						do: 
							[ : err | 
							MaClientServerPreferences debug 
								ifTrue: [ err pass ]
								ifFalse: 
									[ "Ok, let's be extra aggressive in not letting server crash.  If one request generates an unexpected error, just log it and discard, but still process subsequent requests."
									Smalltalk 
										logError: err description
										inContext: err signalerContext
										to: 'SqueakDebug.log' ] ] ].
				MagmaNodeUpdate new
					remove: self node primaryLocation ;
					setPrimary: console thisLocation ;
					processUsing: console ] ]! !

!MagmaRepositoryRequest methodsFor: '*magma server' stamp: 'cmm 3/6/2009 10:05'!
shouldBeSecondary
	"This request should only run on the secondary server.  If I am not in restoreMode (the primary) then ping the primary to see whether it is still there.  If it isn't, take over, otherwise signal a MagmaWrongServerError back to the client."
	| session |
	self isPrimary ifTrue: 
		[ (failedLocation isNil or: [ (self node secondaryLocations includes: failedLocation) not ]) ifTrue: 
			[ MagmaWrongServerError new
				correctServer: self node anySecondaryLocation ;
				signal ].
		session := self node sessionFor: failedLocation.
		console removeSecondaryLocation: failedLocation.
		[ session beWarmBackupFor: self thisLocation ] 
			on: NetworkError
			do: 
				[ : err | 
				"It really is down, don't do anything."
				 ] ]! !

!MagmaRepositoryRequest methodsFor: '*magma server' stamp: 'cmm 11/14/2008 13:34'!
thisLocation
	^ console ifNotNil: [ console thisLocation ]! !

!MaClause methodsFor: '*magma server-private' stamp: 'cmm 7/5/2008 21:31'!
addMinimumEnumerationsTo: aMaQueryTrunk using: aMagmaLargeCollectionManager 
	aMaQueryTrunk addClause: (self asTrunkClause lastKnownSize: (self isAlwaysFalse 
				ifTrue: [ 0 ]
				ifFalse: 
					[ (aMagmaLargeCollectionManager oidsIndexForAttribute: attribute) 
						numberOfEntriesFrom: lowKey
						to: highKey ]))! !

!MaClause methodsFor: '*magma server-private' stamp: 'cmm 7/5/2008 21:31'!
minimumSearchSpaceSizeUsing: aMagmaCollectionManager 
	^ self isAlwaysFalse 
		ifTrue: [ 0 ]
		ifFalse: 
			[ (aMagmaCollectionManager oidsIndexForAttribute: attribute) 
				numberOfEntriesFrom: lowKey
				to: highKey ]! !

!MaClause methodsFor: '*magma server-private' stamp: 'cmm 7/5/2008 21:31'!
shouldInclude: oid using: aMagmaCollectionManager 
	"in the keys indexes, the oid is the key, the key is the value."
	self isAlwaysFalse ifTrue: [ ^ false ].
	(aMagmaCollectionManager keysIndexForAttribute: attribute) 
		entriesFrom: oid
		do: 
			[ : eachRecord : bax : level | 
			((eachRecord valueAtByteArrayIndex: bax) 
				between: lowKey
				and: highKey) ifTrue: 
				[ "key"
				^ true ] ]
		until: [ : eachRecord : bax : level | (eachRecord keyAtByteArrayIndex: bax) > oid ].
	^ false! !

!MagmaLargeCollection methodsFor: '*magma server' stamp: 'cmm 8/12/2007 20:43'!
implementationClass
	self subclassResponsibility! !

!MagmaLargeCollection methodsFor: '*magma server' stamp: 'cmm 3/16/2005 12:53'!
newManagerUsing: aMagmaSession

	self subclassResponsibility! !

!MaObjectBuffer methodsFor: '*magma server' stamp: 'cmm 1/25/2005 14:58'!
appendGraphNode: oid into: aMaSerializedGraphBuffer using: aMaReadStrategy currentDepth: anInteger minDepth: minDepth with: alreadyAppendedSet filer: aMaObjectFiler

	| buffer |

	(alreadyAppendedSet includes: oid) ifTrue: [ ^self ].
	alreadyAppendedSet add: oid.
	buffer _ aMaObjectFiler
		appendObject: oid
		into: aMaSerializedGraphBuffer.
	buffer
		appendGraphUsing: aMaReadStrategy
		into: aMaSerializedGraphBuffer
		currentDepth: anInteger + 1
		minDepth: minDepth
		with: alreadyAppendedSet
		filer: aMaObjectFiler! !

!MaObjectBuffer methodsFor: '*magma server' stamp: 'cmm 1/25/2005 14:01'!
appendGraphUsing: aMaReadStrategy into: aMaSerializedGraphBuffer currentDepth: currentDepth minDepth: minDepth with: alreadyAppendedSet filer: aMaObjectFiler

	1
		to: self objectInstSize
		do:
			[ :varIndex | | oid |
			(MaOidCalculator isOidForOptimizedObject: (oid _ self maInstVarAt: varIndex))
				ifFalse:
					[ | newMinDepth |
					currentDepth < (
						newMinDepth _
							(aMaReadStrategy
								depthOfAttribute: varIndex
								onClassWithId: self classId
								version: self classVersion) + currentDepth
									max: minDepth)
										ifTrue:
											[ self
												appendGraphNode: oid
												into: aMaSerializedGraphBuffer
												using: aMaReadStrategy
												currentDepth: currentDepth
												minDepth: newMinDepth
												with: alreadyAppendedSet
												filer: aMaObjectFiler ] ] ]
! !

!MaObjectBuffer methodsFor: '*magma server' stamp: 'cmm 3/26/2007 22:40'!
asSentientBuffer
	"Convert me to a buffer of the correct type."
	^ self species 
		createFromByteArray: byteArray
		startPos: startPos! !

!MaObjectBuffer methodsFor: '*magma server' stamp: 'cmm 5/24/2005 17:20'!
bufferPositionOfBytesToStore
	"Default 1-based position of the byte within each byteArray that we start storing to disk at."

	^ 7! !

!MaObjectBuffer methodsFor: '*magma server' stamp: 'cmm 7/20/2007 11:11'!
canConflict
	^ true! !

!MaObjectBuffer methodsFor: '*magma server' stamp: 'cmm 3/8/2007 12:58'!
graphDo: oneArgBlock using: aMaReadStrategy for: aMagmaFileTraverser

	self subclassResponsibility! !

!MaObjectBuffer methodsFor: '*magma server' stamp: 'cmm 11/24/2004 13:35'!
numberOfBytesToStore
	"Answer the number of bytes that should be stored for this buffer.  This is used by Magma, which doesn't need to store the oid since that is kept in the oid table."

	^self physicalSize - self bufferPositionOfBytesToStore + 1! !

!MaObjectBuffer methodsFor: '*magma server' stamp: 'cmm 7/23/2007 15:13'!
preWriteAdjustmentFrom: existingBuffer 
	"Called just before writing to disk.  Used to apply MagmaCounter deltas."
	self classId = MagmaClassIdManager magmaCounterClassId ifTrue: [ self updateSharedValueFrom: existingBuffer ]! !

!MaObjectBuffer methodsFor: '*magma server' stamp: 'cmm 7/24/2007 20:58'!
updateSharedValueFrom: existingObjectBuffer 
	"These inst-var indices match MagmaCounter; hard-coded for maximum performance."
	| shared delta |
	shared _ existingObjectBuffer 
		ifNil: [ "A brand new counter" 0 ]
		ifNotNil: 
			[ MaOidCalculator integerFromOid: (existingObjectBuffer maInstVarAt: 1) ].
	delta _ MaOidCalculator integerFromOid: (self maInstVarAt: 2).
	self 
		maInstVarAt: 1
		put: (MaOidCalculator oidForInteger: shared + delta).
	self 
		maInstVarAt: 2
		put: (MaOidCalculator oidForInteger: 0)! !

!MaCommitPackage methodsFor: '*magma server' stamp: 'cmm 4/5/2005 21:34'!
shouldServerRefresh
	"Anytime a largeCollection domain object changes, the server should
	refresh so we process all of the new large collections or their new
	indexes."

	^ allLargeCollectionChanges notNil
	and: [ allLargeCollectionChanges anySatisfy: [ :each | each requiresServerRefresh ] ]! !

!MaServerPerformRequest methodsFor: '*magma server' stamp: 'cmm 9/6/2007 21:33'!
process
	^ repositoryController 
		perform: selector
		withArguments: arguments! !

MagmaNotification subclass: #MagmaRecoveryNotification
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Magma server-Recovery'!

!MaExpression methodsFor: '*magma server-private' stamp: 'cmm 5/31/2006 20:23'!
addMinimumEnumerationsTo: aMaQueryTrunk using: aMagmaLargeCollectionManager 
	"For each OR'd condition, we must add a new range to aMaQueryTrunk.  Within each OR, we select the shortest ANDed condition to add to aMaQueryTrunk."
	self isConjunction 
		ifTrue: 
			[ | smallest smallestSize |
			smallest _ smallestSize _ nil.
			self terms do: 
				[ : each | 
				| eachSize |
				eachSize _ each minimumSearchSpaceSizeUsing: aMagmaLargeCollectionManager.
				(smallest isNil or: [ eachSize < smallestSize ]) ifTrue: 
					[ smallest _ each.
					smallestSize _ eachSize ] ].
			smallest 
				addMinimumEnumerationsTo: aMaQueryTrunk
				using: aMagmaLargeCollectionManager ]
		ifFalse: 
			[ self terms do: 
				[ : each | 
				each 
					addMinimumEnumerationsTo: aMaQueryTrunk
					using: aMagmaLargeCollectionManager ] ]! !

!MaExpression methodsFor: '*magma server-private' stamp: 'cmm 10/8/2006 21:04'!
minimumSearchSpaceSizeUsing: aMaLargeCollectionManager 
	| calculateSizeBlock |
	calculateSizeBlock _ 
	[ : each | 
	each minimumSearchSpaceSizeUsing: aMaLargeCollectionManager ].
	^ self isConjunction 
		ifTrue: [ (self terms collect: calculateSizeBlock) min ]
		ifFalse: [ self terms detectSum: calculateSizeBlock ]! !

!MaExpression methodsFor: '*magma server-private' stamp: 'cmm 4/12/2006 00:22'!
shouldInclude: oid using: aMagmaCollectionManager 
	^ self isConjunction 
		ifTrue: 
			[ self terms allSatisfy: 
				[ : each | 
				each 
					shouldInclude: oid
					using: aMagmaCollectionManager ] ]
		ifFalse: 
			[ self terms anySatisfy: 
				[ : each | 
				each 
					shouldInclude: oid
					using: aMagmaCollectionManager ] ]! !

!MaLargeArrayOfNumbers methodsFor: '*magma server' stamp: 'cmm 3/22/2005 18:07'!
do: aBlock for: aMagmaFileTraverser using: aMaReadStrategy

	1 to: self size do:
		[ : x |  | oid |
		(oid _ self at: x) ~= 0
			ifTrue:
				[ aMagmaFileTraverser
					graphFrom: oid
					do: aBlock
					using: aMaReadStrategy ] ]! !

!MaVariableBuffer methodsFor: '*magma server' stamp: 'cmm 3/8/2007 12:59'!
graphDo: oneArgBlock using: aMaReadStrategy for: aMagmaFileTraverser
	aMagmaFileTraverser
		graphFromPointers: self
		do: oneArgBlock
		using: aMaReadStrategy! !

!Object methodsFor: '*magma server' stamp: 'cmm 10/22/2006 15:01'!
maIsTransactional
	^ false! !

PackageInfo subclass: #MagmaServerPackageInfo
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Magma server-Info'!

!MagmaServerPackageInfo class methodsFor: 'overriding' stamp: 'cmm 12/27/2008 13:32'!
packageName
	^ 'Magma server'! !

!MagmaServerPackageInfo methodsFor: 'as yet unclassified' stamp: 'cmm 10/31/2004 23:42'!
changesText

	^'
1.0gamma8
	- removed "allowNewClasses" feature.  This will be handled with real security later.
	- all administration functions are now either remote requests or no longer needed.
	- support for first-class MagmaRepositoryCodeBase.

1.0gamma7
	- Compatible with Squeak 3.7.
	- Better support for Symbols (thanks to Brent Pinkney!!).
	- fix for storing Exceptions in the repository (to accommodate changes in Squeak 3.6).
	- New Date index type for MagmaCollections.
	- New Float index type for MagmaCollections.
	- MagmaSessions cleaned from MethodContext receivers, allowing full garbage collection (thanks to Giovanni!!).
	- Critical bugs with MagmaCollections fixed and now pass the new, extra-stringent tests in the Tester package.
	
1.0gamma6
	- works with Squeak 3.6.  Doesn''t work with versions before 3.6.
	- fixed issue with servers that were running when the image was saved (functionality implemented in 1.0gamma5).  Fixed similar bug for saving with open local repository connections.
	- disconnect now cleans up the proxies for that session.  Please be patient, this can take a while if you have a lot.  Also now cleaning dynamic largeCollection attributes when disconnecting a session.
	- fixed bug with creating repository using an unqualified file name.
	- fixed an unnoticed reference to platform-specific code (ByteArray>>unsignedByteAt: uses FFI primitives).
	- improved the internal system-read-strategy.


1.0gamma5
	- Servers and clients can now survive image restarts.
	- added a new session preference (see MagmaPreferences), "signalProxyMaterialization" so you can capture MagmaNotifications and fine-tune your read strategies.
	- Another performance improvement!!  ReadStrategies are now cached on the server, instead of in your MagmaSession, improving read-efficiency when using read strategies because the strategy-object no longer needs to be serialized and transported for every proxy materialization.  Additionally, ReadStrategies for classes that don''t exist in the repository are no longer dropped.  Instead, they are remembered until you put those classes into the repository.
	- commit conflicts no longer provide the entire MagmaClientConnection object who conflicted with your session, just the userId.  This is due to the aforementioned cached readStrategy, no need to be sending that down on a commit conflict.
	- Since we have eradicated the "root trap" we can make replacing the root consistent with other Magma changes.  A transaction is now required.
	- several bug fixes.
'! !

!MagmaServerPackageInfo methodsFor: 'accessing' stamp: 'cmm 3/21/2009 13:53'!
prerequisitePackageNames
	^ super prerequisitePackageNames , { 
			(MaExceptionHandlingPackageInfo packageName).
			(MaBaseAdditionsPackageInfo packageName).
			'Collections-BTree'.
			(MaObjectSerializationPackageInfo packageName).
			(MaSpecialCollectionsPackageInfo packageName).
			(MaContextualSearchPackageInfo packageName).
			(MaClientServerPackageInfo packageName).
			(MagmaClientPackageInfo packageName)
		 }! !

!MagmaServerPackageInfo methodsFor: 'as yet unclassified' stamp: 'cmm 2/4/2003 18:12'!
readmeText

	^'This package is used to host a Magma repository.  For more information visit http://minnow.cc.gatech.edu/squeak/2665.'! !

!MaTransactionRequest methodsFor: '*magma server' stamp: 'cmm 11/17/2008 09:33'!
nodeCheck
	self shouldBePrimary! !

!MaTransactionRequest methodsFor: '*magma server' stamp: 'cmm 12/17/2002 11:13'!
process
	^repositoryController newTransactionFor: sessionId! !

!MaVariableWordBuffer methodsFor: '*magma server' stamp: 'cmm 1/25/2005 14:07'!
appendGraphUsing: aMaReadStrategy into: aMaSerializedGraphBuffer currentDepth: currentDepth minDepth: minDepth with: alreadyAppendedSet filer: aMaObjectFiler

	"Nothing to do for a variable-word object."! !

!MaVariableWordBuffer methodsFor: '*magma server' stamp: 'cmm 3/8/2007 12:59'!
graphDo: oneArgBlock using: aMaReadStrategy for: aMagmaFileTraverser
	"nothing to do for a word-object buffer."! !

!MagmaFullBackupRequest methodsFor: '*magma server' stamp: 'cmm 3/7/2009 14:20'!
process
	^ repositoryController fullBackupToServerPath: serverPath! !

!MaRefreshSegmentRequest methodsFor: '*magma server' stamp: 'cmm 6/4/2006 20:12'!
process
	^ self repository refreshSegmentIn: self! !

!MaLoadProgressRequest methodsFor: '*magma server' stamp: 'cmm 11/16/2008 23:31'!
nodeCheck
	self shouldBePrimary ! !

!MaLoadProgressRequest methodsFor: '*magma server' stamp: 'cmm 11/16/2008 23:32'!
process
	^ (self repository collectionManagerFor: collectionOid) loadProgress! !

!MagmaNoteUnreachableServer methodsFor: '*magma server' stamp: 'cmm 11/22/2008 17:44'!
nodeCheck
	self shouldBePrimary! !

!MagmaNoteUnreachableServer methodsFor: '*magma server' stamp: 'cmm 3/1/2009 17:30'!
process
	"My #shouldBePrimary method has already ensured I'm the primary."
	^ nil! !

!ReadWriteStream methodsFor: '*magma server' stamp: 'cmm 5/31/2005 17:53'!
maRead: someNumber bytesFromPosition: bufferPosition of: aByteArray atFilePosition: filePosition

	^ self
		position: filePosition ;
		readInto: aByteArray
		startingAt: bufferPosition
		count: someNumber! !

!ReadWriteStream methodsFor: '*magma server' stamp: 'cmm 5/7/2005 22:55'!
maWrite: numberOfBytes bytesFromPosition: bufferPosition of: byteArray atFilePosition: filePosition

	self position: filePosition.
	self
		next: numberOfBytes
		putAll: byteArray
		startingAt: bufferPosition! !

!ReadWriteStream methodsFor: '*magma server' stamp: 'cmm 6/15/2005 15:15'!
readInto: aByteArray startingAt: startIndex count: numberOfBytes
	"Read n objects into the given collection.  Return the number of bytes read."

	| count |
	count _ 0.
	[ count < numberOfBytes ]
		whileTrue:
			[ | byte |
			byte _ self next.
			byte
				ifNil: [ ^ count ]
				ifNotNil:
					[ aByteArray
						at: startIndex+count
						put: byte ].
			count _ count + 1 ].
	^ count! !

!MaRefreshRequest methodsFor: '*magma server' stamp: 'cmm 1/8/2009 19:02'!
process
	^ self repository renewTransactionEntryFor: sessionId! !

!MaSerializedGraphBuffer methodsFor: '*magma server' stamp: 'cmm 8/13/2002 21:27'!
decreaseBufferSizeBy: anInteger

	self bufferSize: self bufferSize - anInteger
! !

!MaGetMcLockerRequest methodsFor: '*magma server' stamp: 'cmm 11/16/2008 19:33'!
nodeCheck
	self shouldBePrimary ! !

!MaGetMcLockerRequest methodsFor: '*magma server' stamp: 'cmm 11/16/2008 19:32'!
process
	^ self repository 
		collectionManagerFor: mcOid
		ifAbsent: [ MagmaEnvironmentError signal: 'Error, no largeCollectoinManager for that collection found.' ]! !
MaAtomicFileStream initialize!
MaObjectRepository initialize!
MaRecoveryManager initialize!
MagmaRepositoryController initialize!
MagmaServerPreferences initialize!
